/* dipolefield.c : draw vectorfield */

/* compile with :
pushd /my/prive/projects/Canvas/v3/doc
cc dipolefield.c -o dipolefield -I/my/c/include -lm -W -Wall
*/

#include "defines.c"
#include "errordef.c"
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <math.h>
#include "mem/ckalloc.c"

/*----------------- routines using math coords -------------------------------------------------*/

/* a pointcharge -1 is at position (-1,0) , and a pointcharge +1 is at position (+1,0) */

#define chargepx 1
#define chargepy 0
#define chargemx -1
#define chargemy 0
#define chargeradius 0.11
#define minrangex -3
#define maxrangex  3
#define minrangey -2
#define maxrangey  2
#define gridstep 0.15
int rowpoints ;
int colpoints ;

typedef struct { /* for returnvalue of ForceAtPoint() */
  double x	 ; /* position */
  double y	 ;
  double fx	 ; /* orthogonal force component */
  double fy	 ;
  } Fap ;

Fap* faplist ; /* really a 2D array, but without rowpointers */

Fap ForceAtPoint( double x, double y );

void MakeFaplist( void ){
  double x, y	 ;
  int	 ix, iy	 ;

  rowpoints = (maxrangex - minrangex + gridstep) / gridstep ; /* sets global */
  colpoints = (maxrangey - minrangey + gridstep) / gridstep ; /* sets global */
  faplist = (Fap*) ckalloc( rowpoints * colpoints * sizeof(Fap) );

  for( iy=0, y=minrangey ; y <= maxrangey ; y+=gridstep, iy++ ){
    for( ix=0, x=minrangex ; x <= maxrangex ; x+=gridstep, ix++ ){
      faplist[iy*rowpoints+ix] = ForceAtPoint( x, y );
      }
    }
  }

Fap ForceAtPoint( double x, double y ){
  Fap	 ret			 ;
  double l2, invl2, invl	 ;
  double dx, dy			 ;
  double fmx, fmy, fpx, fpy	 ;

  ret.x = x ; ret.y = y ; ret.fx = ret.fy = 0 ;

  /* component due to minus charge : */
  dx = x - chargemx ;
  dy = y - chargemy ;
  l2 = dx * dx + dy * dy ; if( l2 <= chargeradius * chargeradius ){ return(ret); }
  invl2 = 1.0 / l2 ;	 /* magnitude of force is proportional to inverse of square of distance */
  invl = sqrt( invl2 );
  fmx = dx * invl * invl2 ;
  fmy = dy * invl * invl2 ;

  /* component due to plus charge : */
  dx = x - chargepx ;
  dy = y - chargepy ;
  l2 = dx * dx + dy * dy ; if( l2 <= chargeradius * chargeradius ){ return(ret); }
  invl2 = 1.0 / l2 ;
  invl = sqrt( invl2 );
  fpx = dx * invl * invl2 ;
  fpy = dy * invl * invl2 ;

  ret.fx = fpx - fmx ;
  ret.fy = fpy - fmy ;
  return( ret );
  }

/*----------------- routines for sending messages to canvas ------------------------------------*/

#define MSGBUFSIZE 100000
uchar	 msgbuf[MSGBUFSIZE];
uchar*	 msgbufnext	 = msgbuf ;
#define REPLYBUFSIZE 1000
uchar	 replybuf[REPLYBUFSIZE] ;

void Send( void ){
  int			 len		 ;
  int			 sockfd		 ;
  uchar*		 sockfilename	 = (uchar*)"/tmp/canvas.sock" ;
  struct sockaddr_un	 sockstructun	 ;
  int			 sunlen		 ;
  int			 ck		 ;

  len = msgbufnext - msgbuf ; if( len <= 0 ){ return ; }
t("len=%d",len);
  sockfd = socket( PF_UNIX, SOCK_STREAM, 0 ); if( sockfd < 0 ){ esx("cant create socket."); }
  sockstructun.sun_family = AF_UNIX ;
  strcpy( sockstructun.sun_path , (char*)sockfilename );
  sunlen = SUN_LEN( &sockstructun );
  if( 0 > connect( sockfd, (struct sockaddr *) &sockstructun , sunlen ) ){ esx("nok connect."); }
t("got socket and connected");
  *(msgbufnext++)= 0 ; len++ ; /* NOTE THAT TERMINATING ZERO IS NECESSARY ! 20091024 */
  ck = write( sockfd, msgbuf, len );
  if( ck != len ){ e("wrote %d of %d bytes", ck, len ); }
t("did write");
  /* wait for server to send a reply, by doing a blocking read : */
  ck = read( sockfd, replybuf, REPLYBUFSIZE - 1 ); /* leave space for terminator */
  if( ck < 0 ){ es("failed read."); return ; }
  replybuf[ck] = 0 ; /* no guarantee that received string is terminated, so add extra terminator */
  t("got reply, size=%d string=%s", ck, replybuf );
t("did read");
  ck = close( sockfd ); if( ck != 0 ){ es("nok close socket"); }
  msgbufnext = msgbuf ;
  }

void Append( void* newmsg ){
  int len ;
  len = ustrlen( (uchar*)newmsg );
  if( MSGBUFSIZE - (msgbufnext - msgbuf) - 1 < len ){ Send(); }
  if( len > MSGBUFSIZE - 1 ){ ex("message too long"); }
  if( msgbufnext != msgbuf ){ *(msgbufnext++) = '\n' ; }
  memcpy( msgbufnext, newmsg, len ); msgbufnext += len ;
  }


/*----------------- routines that draw result using image coordinates --------------------------*/

#define margin 20
#define coordscale 100.0
#define forcescale 2
/* forcescale is number of pixels length of half lineseg (which has a single color) for force 1.0 */
int imgorix, imgoriy ; /* (x;y) where math coords are (0;0) */
char cmdbuf[100];

void InitCoords( void ){
  imgorix = (int) (margin - minrangex * coordscale) ;
  imgoriy = (int) (margin - minrangey * coordscale) ;
  Append("flushlayer\nclearimobuf");
  }

void DrawCharges( void ){
  sprintf( cmdbuf, "drawdisc %4.1f %4.1f %4.1f",
    chargemx * coordscale + imgorix, chargemy * coordscale + imgoriy, chargeradius * coordscale );
  Append( cmdbuf );
  Append("setdrawcolor 255 255 0 0\nflushlayer");
  sprintf( cmdbuf, "drawdisc %4.1f %4.1f %4.1f",
    chargepx * coordscale + imgorix, chargepy * coordscale + imgoriy, chargeradius * coordscale );
  Append( cmdbuf );
  Append("setdrawcolor 255 0 255 0\nflushlayer");
t("going to send");
  Append("showimobuf"); Send(); /* debug */
t("did send");
  }

void DrawFaps( void ){
  int	 i, n			 ;
  Fap	 fap			 ;
  double x1, y1, x2, y2		 ;
  double dx, dy, l2, maxl2	 ;

  Append("setdrawwidth 1");
  n = rowpoints * colpoints ;
  /* drawing green parts : */
  for( i=0 ; i < n ; i++ ){ fap = faplist[i] ;
    if( abs(fap.fx * forcescale) > gridstep * coordscale ){ continue ; }
    if( abs(fap.fy * forcescale) > gridstep * coordscale ){ continue ; }

    if( fap.fx == 0 && fap.fy == 0 ){ continue ; }
    x1 = fap.x * coordscale + imgorix ;
    y1 = fap.y * coordscale + imgoriy ;
    x2 = x1 + fap.fx * forcescale ;
    y2 = y1 + fap.fy * forcescale ;
    sprintf( cmdbuf,"drawlineseg %4.1f %4.1f %4.1f %4.1f", x1, y1, x2, y2 ); Append( cmdbuf );
    }
  Append("setdrawcolor 255 0 255 0\nflushlayer");
  /* drawing red parts : */
  for( i=0 ; i < n ; i++ ){ fap = faplist[i] ;
    if( abs(fap.fx * forcescale) > gridstep * coordscale ){ continue ; }
    if( abs(fap.fy * forcescale) > gridstep * coordscale ){ continue ; }

    if( fap.fx == 0 && fap.fy == 0 ){ continue ; }
    x1 = fap.x * coordscale + imgorix ;
    y1 = fap.y * coordscale + imgoriy ;
    x2 = x1 - fap.fx * forcescale ;
    y2 = y1 - fap.fy * forcescale ;
    sprintf( cmdbuf,"drawlineseg %4.1f %4.1f %4.1f %4.1f", x1, y1, x2, y2 ); Append( cmdbuf );
    }
  Append("setdrawcolor 255 255 0 0\nflushlayer");
  Append("showimobuf"); Send(); /* debug */
  }

void ShowAndSave( void ){
  int ck ;

  sprintf( cmdbuf, "showimobuf\nsavearea 0 0 %4.1f %4.1f /tmp/dipolefield.imo",
    (maxrangex - minrangex) * coordscale + margin + margin,
    (maxrangey - minrangey) * coordscale + margin + margin );
  Append( cmdbuf );
  Send();
  ck = system("ImoToJpeg /tmp/dipolefield.imo /tmp/dipolefield.jpg");
  if( ck != 0 ){ es("nok ImoToJpeg"); }
  }

/*----------------- main -----------------------------------------------------------------------*/

int main( void ){
  verbose = 1 ;
  InitCoords();		v("did initcoords");
  MakeFaplist();	v("did makefaplist");
  DrawFaps();		v("did drawfaps");
  DrawCharges();	v("did drawcharges");
  ShowAndSave();	v("did showandsave");
  return(0);
  }
