/* TODO: #define CLICKDRAGLISTSIZE 500 */
/* TODO: stringclickdrag's result has 2 unused chars in front of it */
/*----------------- extern --------------------------------------------------------------*/

#include "defines.h"		 /* gnu extensions, v(), ex(), uint16_t, etc	 */
#include <sys/time.h>		 /* struct timeval, gettimeofday()		 */
#include <stdio.h>
#include "clickdragT.h"		 /* #defines of PRESS etc.			 */

extern int verbose ;

/*----------------- public --------------------------------------------------------------*/

/*----------------- filescope -----------------------------------------------------------*/

/* a click is only valid if only that button was clicked and released within 0.2 seconds */

typedef struct {
  int	 type	 ;
  int	 button	 ; /* button flags : button1=2 , button2=4, etc ; buttonX=(1<<X) */
  double time	 ; /* timestamp, since 1970, resolution = 100 us */
  int	 x1	 ;
  int	 y1	 ;
  int	 x2	 ;
  int	 y2	 ;
  } clickdragev ;

#define CLICKTIMEOUT 0.2
#define DCLICKTIMEOUT 0.3
#define CLICKDRAGSTRINGBUFSIZE 16000

static struct timeval	 clickdragreftime	 ;
static clickdragev	 clickdraglist[500]	 ;
static int		 clickdraglistcel	 ;
static uchar		 clickdragstring[CLICKDRAGSTRINGBUFSIZE] ;

/*----------------- SetClickdragReftime --------------------------------------------------*/

void SetClickdragReftime( void ){
  if( 0 > gettimeofday( &clickdragreftime, NULL ) ){ es("nok gettimeofday") ; }
  }

/*----------------- RegisterClickdrag ----------------------------------------------------*/

void RegisterClickdrag( int type, int button, int x1, int y1, int x2, int y2 ){
  /* does not return status because event poller can not do anything usefull with that */
  clickdragev	 cde, cde2, cdli, cdlj	 ;
  struct timeval tv			 ;
  static uchar	 buttonsdown		 ; /* per-button flags */
  double	 time, dt		 ;
  int		 i, j, foundit, dx, dy	 ;
  int		 secs, usecs		 ;

  /*v("got type %d for button %d while buttonsdown is %d", type, button, buttonsdown );*/
  /*v("~buttonsdown=0x%X 1<<button=0x%X", ~buttonsdown, 1<<button );*/
  if( type == PRESS   && (buttonsdown    & (1<<button)) ){ return ; } /* discard */
  if( type == RELEASE && ((~buttonsdown) & (1<<button)) ){ return ; } /* discard */
  if( type == LEAVE   && buttonsdown == 0               ){ return ; } /* nothing to do */

  if(   type == PRESS   ){ buttonsdown |=  (1<<button) ; }
  elif( type == RELEASE ){ buttonsdown &= ~(1<<button) ; }
  elif( type == LEAVE   ){ buttonsdown = 0 ; }

  if( clickdraglistcel >= 500       ){ e("clickdraglist full"); return ; }
  if( 0 > gettimeofday( &tv, NULL ) ){ es("nok gettimeofday") ; return ; }
  /*v("clickdragreftime = %d sec , %d usec",
  clickdragreftime.tv_sec, clickdragreftime.tv_usec );*/
  usecs = tv.tv_usec - clickdragreftime.tv_usec ;
  secs =  tv.tv_sec  - clickdragreftime.tv_sec  ;
  time = (double) secs + (double) usecs * 0.000001 ;
  /*v("time=%19.17g", time );*/
  cde.type	 = type		 ;  cde.button	 = button	 ;
  cde.time	 = time		 ;
  cde.x1	 = x1		 ;  cde.y1	 = y1		 ;
  cde.x2	 = x2		 ;  cde.y2	 = y2		 ;

  clickdraglist[clickdraglistcel++] = cde ;
  /*v("clickdraglistcel = %d", clickdraglistcel );*/
  /* evaluate previous events, to detect click, dclick, and drag ; */
  /* if detected, replace basic events with parsed one ; */
  /* if found sequence that can not be valid result, delete it */

  if( cde.type == RELEASE ){ /* can be end of click or drag */
    /*v("release");*/
    if( buttonsdown != 0 ){ return ; } /* ignore if any other buttons pressed */
    /*v("valid release");*/
    /* find corresponding press event : */
    for( i=clickdraglistcel-2 ; i>=0 ; i-- ){
      cdli = clickdraglist[i] ;
      /*v("i=%d type=%d button=%d", i, cdli.type, cdli.button );*/
      if( cdli.type   != PRESS  ){ continue ; }
      if( cdli.button != button ){ continue ; }
      break ;
      }
    if( i < 0 ){ clickdraglistcel-- ; return ; } /* ignore. also dont store */
    /*v("found match");*/
    dt = cde.time - cdli.time ; dx = abs(x1 - cdli.x1) ; dy = abs(y1 - cdli.y1) ;
    if( dx >= 1 || dy >= 1 ){ /* drag */
      /*v("drag");*/
      cde2.type = DRAG ;
      cde2.time = time ;
      cde2.button = button ;
      cde2.x1 = cdli.x1 ; cde2.x2 = x1 ;
      cde2.y1 = cdli.y1 ; cde2.y2 = y1 ;
      clickdraglist[i] = cde2 ;
      clickdraglistcel = i + 1 ;
      return ;
      }
    if( dt >= CLICKTIMEOUT ){	 /* invalid click ; */
      /*v("invalid click");*/
      clickdraglistcel = i ; /* remove entire sequence from matching press to this release */
      return ;
      }
    /* click */
    /*v("valid click");*/
    cde2.type = CLICK ;
    cde2.time = cde.time ;
    cde2.button = button ;
    cde2.x1 = x1 ; cde2.y1 = y1 ; cde2.x2 = cde2.y2 = 0 ;
    clickdraglist[i] = cde2 ; clickdraglistcel = i+1 ;

    /* is this click second click of a doubleclick ? */
    /* ie it was preceded by another click of same button,
       which occurred less than timeout ago */
    foundit = 0 ;
    for( j=clickdraglistcel-2 ; j>=0 ; j-- ){
      cdlj = clickdraglist[j] ;
      /*v("j=%d type=%d button=%d time=%g nowtime=%g DCLICKTIMEOUT=%g",*/
      /*  j, cdlj.type, cdlj.button, cdlj.time, time, DCLICKTIMEOUT );*/
      if( time - cdlj.time > DCLICKTIMEOUT ){ break ; }
      if( cdlj.type   != CLICK  ){ continue ; }
      if( cdlj.button != button ){ continue ; }
      if( abs(cdlj.x1 - x1) > 1 || abs(cdlj.y1 - y1) > 1 ){ v("moved"); break ; }
      foundit = 1 ; break ;
      }
    if( ! foundit ){
      /*v("not second click of doubleclick");*/
      return ; }
    cde2.type = DCLICK ; /* rest of cde2 is already correct */
    clickdraglist[j] = cde2 ;
    clickdraglistcel = j + 1 ;
    /*v("completes doubleclick");*/
    return ;
    }
  elif( cde.type == LEAVE ){ /* did clear buttonsdown, discard any incomplete clicks/drags */
    /*v("leave");*/
    for( j=clickdraglistcel-2 ; j>=0 ; j-- ){
      cdlj = clickdraglist[j] ;
      if(  cdlj.type != PRESS && cdlj.type != RELEASE
        && cdlj.type != NOP   && cdlj.type != LEAVE   ){
        break ; }
      }
    clickdraglistcel = j + 1 ;
    }
  }

/*----------------- StringClickdrag -----------------------------------------------------*/

uchar* StringClickdrag( void ){
  /* returns pointer to string, or NULL if no events.			 */
  /* stringbuf is guaranteed to have an unused uchar in front of it.	 */
  /*   (so it can be used to put '+' character)				 */
  /* side-effect : clears clickdraglist */
  clickdragev	 cde		 ;
  uchar*	 oclickdragstring ;
  uchar*	 p		 ;
  int		 i		 ;
  struct timeval tv		 ;
  double	 time		 ;
  int		 secs, usecs	 ;
  uint		 buttonsdown	 ;
  int		 lastix		 ;

  clickdragstring[0] = '-' ; clickdragstring[1] = 0 ; /* default (for errors) */
  if( 0 > gettimeofday( &tv, NULL) ){ es("nok gettimeofday") ; return( clickdragstring + 1 ); }
  usecs = tv.tv_usec - clickdragreftime.tv_usec ;
  secs =  tv.tv_sec  - clickdragreftime.tv_sec  ;
  time = (double) secs + (double) usecs * 0.000001 ;
  oclickdragstring = clickdragstring + CLICKDRAGSTRINGBUFSIZE ;
  p = clickdragstring ;
  *(p++) = '+' ;	 /* content is not used, but uchar is guaranteed to be available */

  /* following is necessary because otherwise
     when client eventloop polls every 0.1 second
     nearly all of presses are discarded by clearlist of poll,
     so nearly none of clicks is ever notified to client.
     Note that drags do not time out, so a press without release can not be discarded.
  */
  /* loop over clickdraglist to find					 */
  /*   last event at start of which buttonsdown was zero		 */
  /*     (which is press of a possible unfinished drag)			 */
  /*     (this also catches all presses of yet-to-be-finished clicks)	 */
  /*   but break when a click is found that can still become a dclick	 */

  buttonsdown = 0 ; lastix = 0 ;
  for( i=0 ; i<clickdraglistcel ; i++ ){
    cde = clickdraglist[i] ;
    if(   cde.type == PRESS   ){
      buttonsdown |= (1<<cde.button) ;
      }
    elif( cde.type == RELEASE ){
      buttonsdown &= ~(1<<cde.button) ;
      if( buttonsdown == 0 ){ lastix = i + 1 ; }
      }
    else {
      if( buttonsdown == 0 ){ lastix = i + 1 ; }
      }
    if( cde.type == CLICK ){
      if( time - cde.time < DCLICKTIMEOUT ){ break ; }
      }
    }

  for( i=0 ; i<lastix ; i++ ){
    cde = clickdraglist[i] ;
    if( cde.type == NOP || cde.type == RELEASE || cde.type == LEAVE ){ continue ; }
    if( cde.type == PRESS ){
      if( time - cde.time < CLICKTIMEOUT ){ break ; }
      continue ;
      }
    if( cde.type == CLICK ){
      if( time - cde.time < DCLICKTIMEOUT ){ break ; }
      }
    if( oclickdragstring - p < 60 ){	 /* not enough for longest string of a clickdragev */
      w("stringbuf too small");
      memmove( clickdraglist, &(clickdraglist[i]), (clickdraglistcel-i)* sizeof(clickdragev) );
      clickdraglistcel -= i ;
      *p=0 ;
      return( clickdragstring + 1 );
      }
    if(   cde.type == CLICK  ){
      p += sprintf( (char*) p,"click %d %d %d\n",  cde.button, cde.x1, cde.y1 ); }
    elif( cde.type == DCLICK ){
      p += sprintf( (char*) p,"dclick %d %d %d\n", cde.button, cde.x1, cde.y1 ); }
    elif( cde.type == DRAG   ){
      p += sprintf( (char*) p,"drag %d %d %d %d %d\n",
        cde.button, cde.x1, cde.y1, cde.x2, cde.y2 );
      }
    else { ex("invalid clickdrageventtype"); }
    }
  *(p++) = 0 ;

  if( i < clickdraglistcel ){ /* encountered a possibly incomplete click or dclick */
    memmove( clickdraglist, &(clickdraglist[i]), (clickdraglistcel-i) * sizeof(clickdragev) );
    clickdraglistcel -= i ;
    }
  else { clickdraglistcel = 0 ; }

/*  v("returning %s", clickdragstring+1 );*/
  return( clickdragstring + 1 );
  }

/*----------------- ClearClickdrag ------------------------------------------------------*/

void ClearClickdrag( void ){		 /* is called when clickdrag recording is turned off */
  clickdraglistcel = 0 ; v("clear");
  }

