/* command.c : command interpreter and effector */
/* TODO: loadimofile's filename, if relative, should be considered relative to /tmp */

#include "defines.h"
#include "clickdragT.h"		 /* #defines of PRESS etc.	 */
#include "clickdrag.h"		 /* *Clickdrag()		 */
#include "fill.h"		 /* FillArea(), CountArea()	 */
#include "imobuf.h"		 /* imobuf			 */
#include "layer.h"		 /* *Layer*()			 */
#include "myincludesT.h"	 /* orgb */
#include "myincludes.h"		 /* ??? */
#include "text.h"
#include "x.h"

/*#include "command.h"*/	 /* DO I NEED IT ?		 */

#include <stdio.h>		 /* sscanf() NOT USED HERE	 */
#include <string.h>		 /* strncmp(), memmove()	 */
#include <unistd.h>		 /* write(), close()		 */
#include <fcntl.h>		 /* O_WRONLY etc		 */

/*----------------- extern --------------------------------------------------------------*/

extern uint32_t	 ButPrsFlag	 ; /* x.c		 */
extern uint32_t	 ButRelFlag	 ;
extern uint32_t	 LeaveFlag	 ;
extern int	 verbose	 ; /* myincludes.c	 */
extern int	 winhi		 ; /* main.c ?		 */
extern int	 winwi		 ; /* main.c ?		 */
extern double	 drawwidth	 ; /* main.c ?		 */
extern orgb	 drawcolor	 ; /* main.c ?		 */
extern int	 dummy		 ; /* main.c		 */
extern Filebuf	 imobuf		 ; /* imobuf.c		 */

/*----------------- filescope types -----------------------------------------------------*/

#define MAXCMDTEXT 1000

typedef struct drawcommand {
  uchar	 cmdtype	 ;
  uchar	 carg		 ; /* countof numeric arguments */
  double arg[8]		 ;
  uchar	 txt[MAXCMDTEXT] ; /* used for saveas loadimofile drawtext loadfont */
  } Dc ;

/*----------------- declares ---------------------------------------------------------------*/

static void	 FillKeywords(		 void						 );
static uchar*	 ParseKeyword(		 uchar* puc, Dc* pdc				 );
static uchar*	 SkipWhite(		 uchar* p					 );
static uchar*	 SkipToWhite(		 uchar* p					 );
static uchar*	 ParseFloat(		 uchar* p  , Dc* pdc				 );
static int	 ParseDrawcommand(	 uchar* cmd, Dc* pdc				 );
static int	 NokCarg(		 Dc* pdc, int needcarg				 );
static int	 NokArgval(		 Dc* pdc, int argix, double minval, double maxval );
static int	 HandleCommand(		 uchar* msg, uchar** reply			 );


/*================= MESSAGE PARSER =========================================================*/

/* to add a command :
 define number, increase size of keywords[], add keyword, write handlercode */
/*   also add a description of it to helptext. */

#define SETDRAWWIDTH	 0
#define SETDRAWCOLOR	 1
#define DRAWPOINT	 2
#define DRAWLINESEG	 3
#define DRAWDISC	 4
#define DRAWLINE	 5
#define DRAWCIRCLE	 6
#define DRAWARC		 7
#define ERASEPOINT	 8
#define ERASELINESEG	 9
#define ERASEDISC	 10
#define ERASELINE	 11
#define ERASECIRCLE	 12
#define ERASEARC	 13
#define SHOWLAYER	 14
#define FLUSHLAYER	 15
#define CLEARLAYER	 16
#define FILLAREA	 17
#define COUNTAREA	 18
#define LOADFONT	 19
#define DRAWTEXT	 20
#define SETAREA		 21
#define COPYAREA	 22
#define SHOWIMOBUF	 23
#define SETAFFECTED      24
#define CLEARIMOBUF	 25
#define LOADIMOFILE	 26
#define SAVEAS		 27
#define SAVEAREA	 28
#define STOREXIMAGE	 29
#define LOADXIMAGE	 30
#define FREEXIMAGEBUF	 31
#define RECORDCLICKDRAG	 32
#define NORECORDCLICKDRAG 33
#define REPORTCLICKDRAG  34
#define QUIT		 35
#define MAXKEYWORD	 36

uchar*	 keywords[MAXKEYWORD+1] ;

static void FillKeywords( void ){
  keywords[SETDRAWWIDTH		 ] = (uchar*) "setdrawwidth"	 ;
  keywords[SETDRAWCOLOR		 ] = (uchar*) "setdrawcolor"	 ;
  keywords[DRAWPOINT		 ] = (uchar*) "drawpoint"	 ;
  keywords[DRAWLINESEG		 ] = (uchar*) "drawlineseg"	 ;
  keywords[DRAWDISC		 ] = (uchar*) "drawdisc"	 ;
  keywords[DRAWLINE		 ] = (uchar*) "drawline"	 ;
  keywords[DRAWCIRCLE		 ] = (uchar*) "drawcircle"	 ;
  keywords[DRAWARC		 ] = (uchar*) "drawarc"		 ;
  keywords[ERASEPOINT		 ] = (uchar*) "erasepoint"	 ;
  keywords[ERASELINESEG		 ] = (uchar*) "eraselineseg"	 ;
  keywords[ERASEDISC		 ] = (uchar*) "erasedisc"	 ;
  keywords[ERASELINE		 ] = (uchar*) "eraseline"	 ;
  keywords[ERASECIRCLE		 ] = (uchar*) "erasecircle"	 ;
  keywords[ERASEARC		 ] = (uchar*) "erasearc"	 ;
  keywords[SHOWLAYER		 ] = (uchar*) "showlayer"	 ;
  keywords[FLUSHLAYER		 ] = (uchar*) "flushlayer"	 ;
  keywords[CLEARLAYER		 ] = (uchar*) "clearlayer"	 ;
  keywords[SHOWIMOBUF		 ] = (uchar*) "showimobuf"	 ;
  keywords[FILLAREA		 ] = (uchar*) "fillarea"	 ;
  keywords[COUNTAREA		 ] = (uchar*) "countarea"	 ;
  keywords[QUIT			 ] = (uchar*) "quit"		 ;
  keywords[CLEARIMOBUF		 ] = (uchar*) "clearimobuf"	 ;
  keywords[SAVEAS		 ] = (uchar*) "saveas"		 ;
  keywords[LOADIMOFILE		 ] = (uchar*) "loadimofile"	 ;
  keywords[STOREXIMAGE		 ] = (uchar*) "storeximage"	 ;
  keywords[LOADXIMAGE		 ] = (uchar*) "loadximage"	 ;
  keywords[FREEXIMAGEBUF	 ] = (uchar*) "freeximagebuf"	 ;
  keywords[RECORDCLICKDRAG	 ] = (uchar*) "recordclickdrag"	 ;
  keywords[NORECORDCLICKDRAG	 ] = (uchar*) "norecordclickdrag";
  keywords[REPORTCLICKDRAG	 ] = (uchar*) "reportclickdrag"	 ;
  keywords[SETAREA		 ] = (uchar*) "setarea"		 ;
  keywords[COPYAREA		 ] = (uchar*) "copyarea"	 ;
  keywords[SAVEAREA		 ] = (uchar*) "savearea"	 ;
  keywords[LOADFONT		 ] = (uchar*) "loadfont"	 ;
  keywords[DRAWTEXT		 ] = (uchar*) "drawtext"	 ;
  keywords[SETAFFECTED		 ] = (uchar*) "setaffected"	 ;
  keywords[MAXKEYWORD		 ] = NULL ;
  }

static uchar* ParseKeyword( uchar* puc, Dc* pdc ){
  T(uchar*)	 keyword, puco, argkeyword	 ;
  uchar		 uc				 ;
  int		 i				 ;

  for( puco=puc ; (uc=*puco) != ' ' && uc != '\t' && uc != '\n' && uc != 0 ; puco++ ){ ; }
  if( puco <= puc ){ e("white where keyword expected"); return(NULL); }
  argkeyword = (uchar*) alloca( puco - puc + 1 );
  memcpy( argkeyword, puc, puco-puc ); argkeyword[puco-puc] = 0 ;

  for( i=0 ; (keyword=keywords[i]) != NULL ; i++ ){
    if( 0 == ustrcmp( keyword, argkeyword ) ){ break ; }
    }
  if( keyword == NULL ){ e("invalid keyword at %s", puc); return(NULL); }
  pdc->cmdtype = i ;
  return( puco );
  }

static uchar* SkipWhite( uchar* p ){
  uchar c ;
  while( (c=*p) == ' ' || c == '\t' ){ p++ ; }
  return(p);
  }

static uchar* SkipToWhite( uchar* p ){
  uchar c ;
  while( (c=*p) != ' ' && c != '\t' && c != '\n' && c != 0 ){ p++ ; }
  return(p);
  }

static uchar* ParseFloat( uchar* p, Dc* pdc ){
  int	 negative	 ;
  int	 b, a		 ; /* before, after (decimal dot) */
  uchar* oldp		 ;
  double div, divpow	 ;
  double da, db, res	 ;
  int	 i		 ;

  /*v("p=%p *p=%s=",p, p );*/
  if(   *p == '-' ){ negative = 1 ; p++ ; }
  else             { negative = 0 ; if( *p == '+' ){ p++ ; } }
  p = ParseInt( p, &b );
  /*v("got int %d", b );*/
  if( p == NULL ){ e("invalid number"); return(NULL); }
  if( *p != '.' ){ da = 0.0 ; }
  else {
    p++ ;
    oldp = p ;
    p = ParseInt( p, &a );
    if( p == NULL || (divpow=(int)(p-oldp)) <= 0 ){ e("invalid float"); return(NULL); }
    for( i=0, div=1.0 ; i < divpow ; i++ ){ div *= 0.1 ; }
    da = ((double) a) * div ;
    }
  db = (double) b ;
  res = (negative)? - db - da : db + da ;
  /*v("result is %g", res );*/
  pdc->arg[(pdc->carg)++] = res ;
  return(p);
  }

static int ParseDrawcommand( uchar* cmd, Dc* pdc ){
  T(uchar*)	 p, p2	 ;
  int		 i, len	 ;

  vvv("got cmd=%s=", cmd );
  p = cmd		   ; if( *p == 0    ){ e("invalid msg : zerosize"); return(-1); }
  p = SkipWhite(p)	   ; if( *p == 0    ){ e("invalid msg : empty"   ); return(-1); }
  p = ParseKeyword(p, pdc) ; if( p  == NULL ){ e("invalid keyword"       ); return(-1); }
    /* also sets cmdtype */
  vvv("cmdtype=%d", pdc->cmdtype );
  if( pdc->cmdtype == SAVEAS || pdc->cmdtype == LOADIMOFILE ){
    p = SkipWhite(p);
    p2 = SkipToWhite(p);
    len = p2 - p ;
    if( len <= 0          ){ e("empty filename arg for loadimofile"); return(-1); }
    if( len >= MAXCMDTEXT ){ e("filename too long"); return(-1); }
    memcpy( pdc->txt, p, len ); pdc->txt[len] = 0 ;
    pdc->carg = 0 ;
    }
  elif( pdc->cmdtype == SAVEAREA ){
    for( i=0 ; i < 4 ; i++ ){
      p = SkipWhite(p) ; if( *p == 0 || *p == '\n' ){ break ; }
							 /* sometimes i get a newline !?! */
      p = ParseFloat(p, pdc) ; if( p == NULL  ){ e("nok parsefloat"        ); return(-1); }
      if( *p == 0 ){ e("no filename for savearea"); return(-1); }
      }
    p = SkipWhite(p);
    p2 = SkipToWhite(p);
    len = p2 - p ;
    if( len <= 0          ){ e("empty filename arg for savearea"); return(-1); }
    if( len >= MAXCMDTEXT ){ e("filename too long"); return(-1); }
    memcpy( pdc->txt, p, len ); pdc->txt[len] = 0 ;
    pdc->carg = 4 ;
    }
  elif( pdc->cmdtype == DRAWTEXT ){
    uchar c ;
    p = SkipWhite(p);
    c = *p ;
    if(   c == 'r' || c == 'd' || c == 'l' || c == 'u' ){ pdc->arg[0] = c ; pdc->carg = 1 ; }
    else { e("invalid direction %c", c ); return(-1); }
    p++ ;
    for( i=1 ; i < 3 ; i++ ){
      p = SkipWhite(p);
      if( *p == '\n' ){ e("unexpected newline");       return(-1); }
      if( *p == 0    ){ e("premature end of command"); return(-1); }
      p = ParseFloat(p, pdc) ;
      if( p == NULL  ){ e("nok parsefloat %d", i );    return(-1); }
      }
    p = SkipWhite(p);
    /* p2 = SkipToEnd */
    for( p2=p ; (c=*p2) != 0 ; p2++ ){ ; }
    /* p2 = SkipbackToNonwhite */
    for( --p2 ; (c=*p2) == ' ' || c == '\t' || c == '\n' ; p2-- ){ ; }
    len = p2 - p + 1 ;
    if( len <= 0   ){ e("empty text arg for drawtext"); return(-1); }
    if( len >= MAXCMDTEXT ){ e("text too long"); return(-1); }
    memcpy( pdc->txt, p, len ); pdc->txt[len] = 0 ;
    }
  elif( pdc->cmdtype == LOADFONT ){
    pdc->carg = 0 ;
    p = SkipWhite(p);
    p2 = SkipToWhite(p);
    len = p2 - p ;
    if( len <= 0          ){ e("empty fontname arg for loadfont"); return(-1); }
    if( len >= MAXCMDTEXT ){ e("filename too long"); return(-1); }
    memcpy( pdc->txt, p, len ); pdc->txt[len] = 0 ;
    }
  else {
    for( i=0 ; i < 8 ; i++ ){		 /* protect against too many arguments for struct */
      vvv("i=%d string=%s",i, p);
      p = SkipWhite(p) ; if( *p == 0 || *p == '\n' ){ break ; }
					 /* sometimes i get a newline !?! */
      vvv("c=*p=%c=%s=", *p, p );
      p = ParseFloat(p, pdc) ; if( p == NULL  ){ e("nok parsefloat"        ); return(-1); }
        /* also increments carg */
      if( *p == 0 ){ break ; }
      }
    }
  return(0);
  }

static int NokCarg( Dc* pdc, int needcarg ){
  if( pdc->carg < needcarg ){ e("not enough args for %s",keywords[pdc->cmdtype]);return(-1); }
  if( pdc->carg > needcarg ){ w("too many args for %s"  ,keywords[pdc->cmdtype]); }
  return(0);
  }

static int NokArgval( Dc* pdc, int argix, double minval, double maxval ){
  double arg ;
  arg = pdc->arg[argix] ;
  if( arg < minval ){
    e("value of arg %d of %s is less than minimum", argix, keywords[pdc->cmdtype] );
    return(-1); }
  if( arg > maxval ){
    e("value of arg %d of %s is more than maximum", argix, keywords[pdc->cmdtype] );
    return(-1); }
  return(0);
  }

static int HandleCommand( uchar* msg, uchar** preply ){
  /* returns 0:OK -1:error -10:quit, positive value for countarea */
  Dc	 dc	 ;
  int	 count	 ;
  int	 ck, i	 ;

  if( msg == NULL ){ ex("null arg"); }
  dc.carg = 0 ;
  if( 0 != ParseDrawcommand( msg, &dc ) ){ e("nok parsedrawcommand"); return(-1); }

  if( verbose ){
    v("cmdtype=%d cmd=%s carg=%d", dc.cmdtype, keywords[dc.cmdtype], dc.carg );
    for( i=0 ; i < dc.carg ; i++ ){ v(" arg[%d]=%g", i, dc.arg[i] ); }
    }

  if( dc.cmdtype == SETDRAWWIDTH ){
    if( NokCarg( &dc, 1 ) ){ return(-1); }
    if( NokArgval( &dc, 0, 0.0, (double) winhi ) ){ return(-1); }
    drawwidth = dc.arg[0] ;
    }
  elif( dc.cmdtype == SETDRAWCOLOR ){
    if( NokCarg( &dc, 4 ) ){ return(-1); }
    if( NokArgval( &dc, 0, 0.0, 255.99 ) ){ return(-1); }
    if( NokArgval( &dc, 1, 0.0, 255.99 ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, 255.99 ) ){ return(-1); }
    if( NokArgval( &dc, 3, 0.0, 255.99 ) ){ return(-1); }
    drawcolor.o = (int) dc.arg[0] ; drawcolor.r = (int) dc.arg[1] ;
    drawcolor.g = (int) dc.arg[2] ; drawcolor.b = (int) dc.arg[3] ;
    }
  elif( dc.cmdtype == DRAWPOINT ){
    if( NokCarg( &dc, 2 ) ){ return(-1); }
    DrawDiscToLayer( dc.arg[0], dc.arg[1], drawwidth * 0.55, 0 );
    }
  elif( dc.cmdtype == DRAWLINESEG ){
    if( NokCarg( &dc, 4 ) ){ return(-1); }
    if( 0 != DrawLinesegToLayer( dc.arg[0], dc.arg[1], dc.arg[2], dc.arg[3], 0 ) ){
      e("nok drawlinesegtolayer"); return(-1); }
    }
  elif( dc.cmdtype == DRAWDISC ){
    if( NokCarg( &dc, 3 ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, 1e6 ) ){ return(-1); }
    DrawDiscToLayer( dc.arg[0], dc.arg[1], dc.arg[2], 0 );
    }
  elif( dc.cmdtype == DRAWLINE ){
    if( NokCarg( &dc, 4 ) ){ return(-1); }
    if( 0 != DrawLinesegToLayer( dc.arg[0], dc.arg[1], dc.arg[2], dc.arg[3], 0 ) ){
      e("nok drawlinesegtolayer"); return(-1); }
    DrawDiscToLayer( dc.arg[0], dc.arg[1], drawwidth * 0.5, 0 );
    DrawDiscToLayer( dc.arg[2], dc.arg[3], drawwidth * 0.5, 0 );
    }
  elif( dc.cmdtype == DRAWCIRCLE ){
    if( NokCarg( &dc, 3 ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, 1e6 ) ){ return(-1); }
    DrawCircleToLayer( dc.arg[0], dc.arg[1], dc.arg[2], 0 );
    }
  elif( dc.cmdtype == DRAWARC ){
    if( NokCarg( &dc, 5 ) ){ return(-1); }
    DrawArcToLayer( dc.arg[0], dc.arg[1], dc.arg[2], dc.arg[3], dc.arg[4], 0 );
    }
  elif( dc.cmdtype == ERASEPOINT ){
    if( NokCarg( &dc, 2 ) ){ return(-1); }
    DrawDiscToLayer( dc.arg[0], dc.arg[1], drawwidth * 0.55, 1 );
    }
  elif( dc.cmdtype == ERASELINESEG ){
    if( NokCarg( &dc, 4 ) ){ return(-1); }
    if( 0 != DrawLinesegToLayer( dc.arg[0], dc.arg[1], dc.arg[2], dc.arg[3], 1 ) ){
      e("nok drawlinesegtolayer"); return(-1); }
    }
  elif( dc.cmdtype == ERASEDISC ){
    if( NokCarg( &dc, 3 ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, 1e6 ) ){ return(-1); }
    DrawDiscToLayer( dc.arg[0], dc.arg[1], dc.arg[2], 1 );
    }
  elif( dc.cmdtype == ERASELINE ){
    if( NokCarg( &dc, 4 ) ){ return(-1); }
    if( 0 != DrawLinesegToLayer( dc.arg[0], dc.arg[1], dc.arg[2], dc.arg[3], 1 ) ){
      e("nok drawlinesegtolayer"); return(-1); }
    DrawDiscToLayer( dc.arg[0], dc.arg[1], drawwidth * 0.5, 1 );
    DrawDiscToLayer( dc.arg[2], dc.arg[3], drawwidth * 0.5, 1 );
    }
  elif( dc.cmdtype == ERASECIRCLE ){
    if( NokCarg( &dc, 3 ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, 1e6 ) ){ return(-1); }
    DrawCircleToLayer( dc.arg[0], dc.arg[1], dc.arg[2], 1 );
    }
  elif( dc.cmdtype == ERASEARC ){
    if( NokCarg( &dc, 5 ) ){ return(-1); }
    DrawArcToLayer( dc.arg[0], dc.arg[1], dc.arg[2], dc.arg[3], dc.arg[4], 1 );
    }
  elif( dc.cmdtype == SHOWLAYER ){
    if( NokCarg( &dc, 2 ) ){ return(-1); } /* ShowLayer checks values of it's args itself */
    if( 0 != ShowLayer( dc.arg[0], dc.arg[1]) ){ e("nok showlayer"); return(-1); }
    }
  elif( dc.cmdtype == FLUSHLAYER ){
    if( NokCarg( &dc, 0 ) ){ return(-1); }
    if( LayerToImobuf()  != 0 ){ e("nok layertoimobuf"); } /* SHOULD MESSAGE TO COMMANDER */
    /* note that layertoimobuf calls clearlayer, so no need to do that here */
    }
  elif( dc.cmdtype == CLEARLAYER ){
    if( NokCarg( &dc, 0 ) ){ return(-1); }
    ClearLayer();
    }
  elif( dc.cmdtype == FILLAREA ){
    if( NokCarg( &dc, 2 ) ){ return(-1); }
    count = FillArea( (int) dc.arg[0], (int) dc.arg[1] );
    t("filled %d pixels", count );
    return(count);
    }
  elif( dc.cmdtype == COUNTAREA ){
    if( NokCarg( &dc, 2 ) ){ return(-1); }
    count = CountArea( (int) dc.arg[0], (int) dc.arg[1] );
    t("counted %d pixels", count );
    return(count);
    }
  elif( dc.cmdtype == LOADFONT ){
    if( dc.carg != 0             ){ e("too many arguments"); return(-1); }
    if( 0 > CkLoadFont( dc.txt ) ){ e("nok LoadFont()");     return(-1); }
    }
  elif( dc.cmdtype == DRAWTEXT ){
    if( NokCarg(   &dc, 3 )                         ){ return(-1); }
    if( NokArgval( &dc, 1, 0.0, (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, (double)(winhi-1) ) ){ return(-1); }
    if( 0 > DrawText( dc.arg[0], dc.arg[1], dc.arg[2], dc.txt ) ){ return(-1); }
    }
  elif( dc.cmdtype == SETAREA ){
    if( NokCarg( &dc, 4 ) ){ return(-1); }
    if( NokArgval( &dc, 0, 0.0, (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 1, 0.0, (double)(winhi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 3, 0.0, (double)(winhi-1) ) ){ return(-1); }
    SetArea((int)dc.arg[0],(int)dc.arg[1],(int)dc.arg[2],(int)dc.arg[3] );
    }
  elif( dc.cmdtype == COPYAREA ){
    if( NokCarg( &dc, 6 ) ){ return(-1) ; }
    if( NokArgval( &dc, 0, 0.0, (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 1, 0.0, (double)(winhi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 3, 0.0, (double)(winhi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 4, 0.0, (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 5, 0.0, (double)(winhi-1) ) ){ return(-1); }
    CopyArea((int)dc.arg[0],(int)dc.arg[1],(int)dc.arg[2],(int)dc.arg[3],
	(int)dc.arg[4],(int)dc.arg[5] );
    }
  elif( dc.cmdtype == SHOWIMOBUF ){
    if( NokCarg( &dc, 0 ) ){ return(-1); }
    ImobufToXimage();
    RenderXimage();
    }
  elif( dc.cmdtype == SETAFFECTED ){
    if( NokCarg( &dc, 4 )                                 ){ return(-1); }
    if( NokArgval( &dc, 0, 0.0      , (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 1, 0.0      , (double)(winhi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 2, dc.arg[0], (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 3, dc.arg[1], (double)(winhi-1) ) ){ return(-1); }
    if( imoaffx1 > dc.arg[0] ){ imoaffx1 = dc.arg[0] ; }
    if( imoaffy1 > dc.arg[1] ){ imoaffy1 = dc.arg[1] ; }
    if( imoaffx2 < dc.arg[2] ){ imoaffx2 = dc.arg[2] ; }
    if( imoaffy2 < dc.arg[3] ){ imoaffy2 = dc.arg[3] ; }
    }
  elif( dc.cmdtype == CLEARIMOBUF ){
    if( NokCarg( &dc, 0 ) ){ return(-1); }
    ClearImobuf();
    }
  elif( dc.cmdtype == LOADIMOFILE ){
    uchar*	 fqfilename	 ;
    int		 ck		 ;
    if( *(dc.txt) == 0 ){ e("no filename"); return(-1); }
    fqfilename = (uchar*) QualifyFilename( dc.txt );
    ck = FileExists( fqfilename, NULL );
    if( ck >  1 ){ e("file %s exists and is not an ordinary file", fqfilename ); return(-1); }
    if( ck <= 0 ){ e("file %s does not exist", fqfilename ); return(-1); }
    ck = GetImofileContent( fqfilename );
    if( ck < 0 ){ e("nok getimofilecontent %s", fqfilename ); }
    }
  elif( dc.cmdtype == SAVEAS ){
    uchar*	 fqfilename	 ;
    int		 ck, fh, size	 ;
    if( *(dc.txt) == 0 ){ e("no filename"); return(-1); }
    if( dc.txt[0] != '/' ){ /* a relative filename */
      /* i want it to be relative to /tmp, not to directory that has canvas executable */
      fqfilename = (uchar*) alloca( 5 + ustrlen(dc.txt) + 1 );
      memcpy( fqfilename    , "/tmp/", 5 );
      ustrcpy( fqfilename + 5, dc.txt );
      }
    else { fqfilename = dc.txt ; }
    fqfilename = (uchar*) QualifyFilename( fqfilename );
    ck = FileExists( fqfilename, NULL );
    if( ck > 1 ){ e("file %s exists and is not an ordinary file", fqfilename ); return(-1); }
    fh = open( (char*)fqfilename, O_WRONLY | O_CREAT, 0660 );
    if( fh <= 0 ){ es("nok open %s", fqfilename ); return(-1); }
    size = sizeof(imoheader) + winhi * winwi * sizeof(orgb) ;
    ck = write( fh, imobuf.buf, size );
    if( ck < size ){ es("nok write imofile"); close(fh); return(-1); }
    ck = close(fh); if( ck != 0 ){ es("nok close imofile %s", fqfilename ); }
    return(0);
    }
  elif( dc.cmdtype == SAVEAREA ){
    uchar*	 fqfilename	 ;
    if( NokCarg( &dc, 4 ) ){ return(-1); }
    if( NokArgval( &dc, 0, 0.0, (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 1, 0.0, (double)(winhi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 2, 0.0, (double)(winwi-1) ) ){ return(-1); }
    if( NokArgval( &dc, 3, 0.0, (double)(winhi-1) ) ){ return(-1); }
    if( *(dc.txt) == 0 ){ e("no filename"); return(-1); }
    if( dc.txt[0] != '/' ){ /* a relative filename */
      /* i want it to be relative to /tmp, not to directory that has canvas executable */
      fqfilename = (uchar*) alloca( 5 + ustrlen(dc.txt) + 1 );
      memcpy( fqfilename    , "/tmp/", 5 );
      ustrcpy( fqfilename + 5, dc.txt );
      }
    else { fqfilename = dc.txt ; }
    SaveArea((int)dc.arg[0],(int)dc.arg[1],(int)dc.arg[2],(int)dc.arg[3], fqfilename );
    }
  elif( dc.cmdtype == STOREXIMAGE ){
    if( NokCarg( &dc, 1 ) ){ return(-1); }
    ck = StoreXimage( dc.arg[0] );
    return(ck);
    }
  elif( dc.cmdtype == LOADXIMAGE ){
    if( NokCarg( &dc, 1 ) ){ return(-1); }
    ck = LoadXimage( dc.arg[0] );
    if( ck == 0 ){ RenderXimage(); }
    return(ck);
    }
  elif( dc.cmdtype == FREEXIMAGEBUF ){
    if( NokCarg( &dc, 1 ) ){ return(-1); }
    ck = FreeXimagebuf( dc.arg[0] );
    return(ck);
    }
  elif( dc.cmdtype == RECORDCLICKDRAG ){
    SetEventmask('+', ButPrsFlag | ButRelFlag | LeaveFlag );
    SetClickdragReftime();
    }
  elif( dc.cmdtype == NORECORDCLICKDRAG ){
    SetEventmask('-', ButPrsFlag | ButRelFlag | LeaveFlag );
    ClearClickdrag();
    }
  elif( dc.cmdtype == REPORTCLICKDRAG ){
    *preply = StringClickdrag() - 1 ;	 /* yields pointer to static buffer that be reused	 */
    }					 /* *preply is '+', followed by event strings		 */
  elif( dc.cmdtype == QUIT ){
    if( NokCarg( &dc, 0 ) ){ return(-1); }
    return(-10);
    }
  else { t("command (%d : %s) is valid but not yet supported",
    dc.cmdtype, keywords[dc.cmdtype] );
    }
  return(0);
  }

/*================= PUBLIC FUNCTIONS ===========================================================*/

void InitCommand( void ){
  FillKeywords();
  }


/* USED AS : ck = HandleMessage( buf, totalgot, &textresult ); */

/* CALLER MUST GUARANTEE THAT
   EITHER buf has at least one byte more than length of message
   OR message is null-terminated. <-- STRONGLY PREFERRED !
*/

int HandleMessage( uchar* buf, int totalgot, uchar** ptextresult ){
  int		 result, error, needquit ;
  int		 cmdix			 ;
  T(uchar*)	 ps, pe, po		 ; /* cmd start, cmd end, over msg */
  uchar		 uc			 ;
  int		 atend			 ;
  uchar*	 puc			 ;
  int		 ck			 ; /* tmp */

  /* ps pointsto first char of current command.						 */
  /* pe pointsto first char after current command,					 */
  /*   which can be newline, terminating zero or first byte after message.		 */

  /* split message in drawcommands, and execute each of them : */
  result = 0 ; error = 0 ; needquit = 0 ; atend = 0 ;
  if( buf == NULL	 ){ ex("buf is null"); }
  if( totalgot <= 0	 ){ ex("totalgot is non-positive"); }
  if( ptextresult == NULL){ ex("ptextresult is null"); }
  ps = buf ; po = ps + totalgot ;
  for( cmdix=0 ; 1 ; cmdix++ ){
    vvv("cmdix=%d", cmdix );
    for( pe=ps ; pe < po && (uc=*pe) != 0 && uc != '\n' ; pe++ ){ ; }
    if( pe == po ){ w("msg is not null-terminated"); atend = 1 ; }
    if( *pe == 0 ){ atend = 1 ; }

    /* ignore commands that are empty or completely whitespace : */
    for( puc=ps ; (uc=*puc) == ' ' || uc == '\t' ; puc++ ){ ; }
    if( puc == pe ){ w("empty command");
      if( atend ){ break ; }
      ps = pe+1 ; continue ;
      }

    *pe = 0 ;
    if( dummy ){ v("dummy: %s", ps ); ck = 0 ; }	 /* dummy run (for testing messaging) */
    else {
      ck = HandleCommand( ps, ptextresult );
      vvv("handlecommand returned %d", ck );
      if( ck == -10 ){ needquit = 1 ; vvv("needquit"); break ; }		 /* quit */
      if( ck <  0   ){ error    = 1 ; e("nok handlecommand"); break ; }
      if( ck >  0   ){ result = ck ; }
      }
    if( atend ){ break ; }

    /* find start of next command (if any) : */
    for( ps=pe+1 ; ps < po && ((uc=*ps) == '\n' || uc == '\t' || uc == ' ') ; ps++ ){;}
    if( ps >= po || uc == 0 ){ vvv("no more commands in message"); break ; }
    }

  if( needquit ){ return(-10); }
  if( error    ){ return( -1); }
  return( result );
  }
