/* TODO: move OverlayOrgb to text.c */

#include "defines.h"
#include <Xlib.h>		 /* XImage		 */
#include <fcntl.h>		 /* open()		 */
#include <unistd.h>		 /* write(), close()	 */
#include "myincludesT.h"	 /* Filebuf, orgb	 */
#include "myincludes.h"		 /* GetMmappedFilebuf()	 */

/*----------------- public --------------------------------------------------------------*/

orgb*	 imor					 ; /* pointo array of orgb in imobuf */
int	 imoaffx1, imoaffx2, imoaffy1, imoaffy2	 ; /* range of affected imobuf pixel coords */
Filebuf	 imobuf					 ; /* main needs it for saveas */
uchar*	 imobuffilename = (uchar*)"/tmp/canvasbuf.imo" ;

void	 MakeImobuf(		 void							 );
void	 ClearImobuf(		 void							 );
int	 GetImofileContent(	 uchar* filename					 );
void	 ImobufToXimage(	 void							 );
int	 SetArea(		 int x1, int y1, int x2, int y2				 );
int	 CopyArea(		 int x1, int y1, int x2, int y2, int x3, int y3		 );
int	 SaveArea(		 int x1, int y1, int x2, int y2, uchar* filename	 );
void	 OverlayOrgb(		 orgb* porig, orgb* povl				 );
void	 DiscardImobuf(		 void							 );

/*----------------- extern --------------------------------------------------------------*/

extern int	 verbose			 ; /* myincludes.c	 */
extern int	 winwi, winhi			 ; /* main.c		 */
extern orgb	 drawcolor			 ; /* main.c		 */
extern int	 ximgpixelsize			 ; /* x.c		 */
extern XImage*	 pxim				 ; /* x.c		 */
extern ulong	 (*OrgbToPixval)( orgb o )	 ; /* x.c		 */

/*----------------- MakeImobuf ------------------------------------------------------------*/

void ClearImobuf( void ){
  memset( imor, 255, winhi * winwi * sizeof(orgb) );
  imoaffx1 = 0 ; imoaffx2 = winwi - 1 ;
  imoaffy1 = 0 ; imoaffy2 = winhi - 1 ;
  }

void MakeImobuf( void ){	 /* imobuf is valid imofile for ease of allocing and saving it */
  int		 cpixels	 ;
  int		 size		 ;
  imoheader*	 pih		 ;
  int		 sharable	 = 1 ;
  int		 hassize	 ;
  int		 pagedsize	 ;
  int		 fh		 ;

  cpixels = winhi * winwi ;
  size =  sizeof(imoheader) + cpixels * sizeof(orgb) ;

  switch( FileExists( imobuffilename, &hassize ) ){
    case 0 :	   				 /* need to create it before i can truncate it */
    vvvv("case 0");
    fh = open( (char*) imobuffilename, O_CREAT | O_WRONLY, 0666 );
    if( fh <= 0                ){ es("nok open %s", imobuffilename    ); sharable = 0 ; break ; }
    if( 1 != write( fh,"\0",1) ){ es("nok write to %s",imobuffilename ); sharable = 0 ; break ; }
    if( 0 != close( fh )       ){ es("nok close %s", imobuffilename   ); sharable = 0 ; break ; }
    hassize = 1 ;
    /* no break here ; fall through to next case. */

    case 1 :
    vvvv("case 1");
    pagedsize = Pagedsize(size) ;
    if( hassize != pagedsize ){
      if( 0 != truncate( (char*)imobuffilename, pagedsize ) ){
        es("nok truncate %s", imobuffilename ); sharable = 0 ; break ; }
      FileExists( imobuffilename, &hassize );
      if( hassize != pagedsize ){ e("truncate cant enlarge files on this system"); sharable = 0 ; }
      }
    imobuf = GetMmappedFilebuf( imobuffilename, 1 );
    break ;

    case 2 : case 3 : case 4 : case 5 : case 6 : case 7 : case 8 : case 9 :
    e("file %s exists and is not an ordinary file", imobuffilename ); sharable = 0 ; break ;
    } /* end switch */
  if( ! sharable ){
    w("making non-sharable imobuf" );
    imobuf = GetMmappedBuf("imobuf", 0, size );
    }

  pih = (imoheader*) imobuf.buf ;
  vvvv("imobuf.buf = %p", imobuf.buf );
  pih->wi = winwi ; pih->hi = winhi ;
  pih->offx = pih->offy = 0 ; pih->scalex = pih->scaley = 1.0 ;
  memcpy( pih->sig, "imo0", 4 );
  pih->byteorder = 0x01020304 ;
  imor = (orgb*) (imobuf.buf + sizeof(imoheader));
  memset( imor, 255, cpixels * sizeof(orgb) );
  imoaffx1 = 0 ; imoaffx2 = winwi - 1 ;
  imoaffy1 = 0 ; imoaffy2 = winhi - 1 ;
  }

/*----------------- DiscardImobuf ----------------------------------------------------------*/

void DiscardImobuf( void ){ /* mainly removes /tmp/canvasbuf.imo */
  DiscardFilebuf( &imobuf );
  if( 1 == FileExists( imobuffilename, NULL ) ){
    if( 0 != unlink( (char*)imobuffilename ) ){ w("nok unlink %s", imobuffilename ); }
    }
  }

/*----------------- GetImofileContent ------------------------------------------------------*/

int GetImofileContent( uchar* filename ){
  Filebuf	 fbt		 ; /* temporary */
  imoheader*	 pih		 ;
  int		 copywi, copyhi	 ;
  int		 x, y		 ;
  T(orgb*)	 pi, po, pil, pol ;
  int		 f		 ; /* tmp */
  
  fbt = GetMmappedFilebuf( filename, 0 );	 /* was fully qualified in ParseArguments */
  pih = (imoheader*) fbt.buf ;
  if( 0 != memcmp( pih->sig, "imo", 3 )	 ){ e("not an imo file");		 return(-1); }
  if( pih->sig[3] != '0'		 ){ w("imofileversion not supported"); }
  if( pih->byteorder != 0x01020304	 ){ e("incorrect byteorder");		 return(-1); }
  f = (int) pih->wi ; copywi = (winwi < f)? winwi : f ;
  f = (int) pih->hi ; copyhi = (winhi < f)? winhi : f ;
  pil = (orgb*) (fbt.buf + sizeof(imoheader)) ;
  pol = imor ;
  for( y=0 ; y < copyhi ; y++, pil += pih->wi, pol += winwi ){
    po = pol ; pi = pil ;
    for( x=0 ; x < copywi ; x++ ){ *(po++) = *(pi++) ; }
    }
  DiscardFilebuf( &fbt );
  imoaffx1 = 0 ;
  if( imoaffx2 < copywi - 1 ){ imoaffx2 = copywi - 1 ; }
  imoaffy1 = 0 ;
  if( imoaffy2 < copyhi - 1 ){ imoaffy2 = copyhi - 1 ; }
  return(0);
  }

/*----------------- ImobufToXimage ---------------------------------------------------------*/

void ImobufToXimage( void ){
  T(orgb*)	 po, pol ;
  int		 x, y	 ;
v("imoaffx1=%d imoaffx2=%d imoaffy1=%d imoaffy2=%d",imoaffx1,imoaffx2,imoaffy1,imoaffy2);
  if( imoaffy1 <  0         ){ imoaffy1 = 0         ; }
  if( imoaffy1 >= winhi     ){ return ; }
  if( imoaffy2 <  0         ){ return ; }
  if( imoaffy2 >  winhi - 1 ){ imoaffy2 = winhi - 1 ; }

  if( ximgpixelsize <= 1 ){
    u8* pi ;
    for( y=imoaffy1 ; y <= imoaffy2 ; y++ ){
      x = imoaffx1 ;
      pi = (u8*) pxim->data + y * winwi + imoaffx1 ;
      po = imor             + y * winwi + imoaffx1 ;
      for( ; x <= imoaffx2 ; x++ ){ *(pi++) = (u8) OrgbToPixval(*(po++)) ; }
      }
    }
  elif( ximgpixelsize <= 2 ){ /* this version animates 5% faster */
    T(u16*) pi, pil ;
    y = imoaffy1 ;
    pol = imor              + winwi * y ;
    pil = (u16*) pxim->data + winwi * y ;
    for( y=imoaffy1 ; y <= imoaffy2 ; y++, pol+=winwi, pil+=winwi ){
      x = imoaffx1 ;
      pi = pil + imoaffx1 ;
      po = pol + imoaffx1 ;
      for( ; x <= imoaffx2 ; x++ ){ *(pi++) = (u16) OrgbToPixval(*(po++)) ; }
      }
    }
  else {
    u32* pi ;
    for( y=imoaffy1 ; y <= imoaffy2 ; y++ ){
      x = imoaffx1 ;
      pi = (u32*) pxim->data + y * winwi + imoaffx1 ;
      po = imor              + y * winwi + imoaffx1 ;
      for( ; x <= imoaffx2 ; x++ ){ *(pi++) = (u32) OrgbToPixval(*(po++)) ; }
      }
    }
  imoaffx1 = winwi ; imoaffx2 = -1 ;
  imoaffy1 = winhi ; imoaffy2 = -1 ;
  }

/*----------------- SetArea -------------------------------------------------------------*/

int SetArea( int x1, int y1, int x2, int y2 ){
  /* sets all pixels of area to current drawcolor */
  int	 x, y ;
  orgb*	 line ;

  for( y=y1 ; y <= y2 ; y++ ){
    line = imor + y * winwi ;
    for( x=x1 ; x <= x2 ; x++ ){ line[x] = drawcolor ; }
    }
  if( imoaffx1 > x1 ){ imoaffx1 = x1 ; }
  if( imoaffy1 > y1 ){ imoaffy1 = y1 ; }
  if( imoaffx2 < x2 ){ imoaffx2 = x2 ; }
  if( imoaffy2 < y2 ){ imoaffy2 = y2 ; }
  return(0);
  }

/*----------------- CopyArea ------------------------------------------------------------*/

int CopyArea( int x1, int y1, int x2, int y2, int x3, int y3 ){
  /* Copies area who'se topleft is x1/y1 and bottomright x2/y2	 */
  /*   to new location where topleft is x3/y3 ;			 */
  /* Old and new area can overlap !				 */
  /* values of arguments have already been verified by CkArgs(). */
  int		 offx, x, y		 ;
  T(orgb*)	 linefrom, lineto	 ;
  int		 wi, hi			 ;

  wi = x2 - x1 ; hi = y2 - y1 ; if( wi < 0 || hi < 0 ){ e("invalid coords"); return(-1); }
  if( imoaffx1 > x3      ){ imoaffx1 = x3      ; }
  if( imoaffx2 < x3 + wi ){ imoaffx2 = x3 + wi ; }
  if( imoaffy1 > y3      ){ imoaffy1 = y3      ; }
  if( imoaffy2 < y3 + hi ){ imoaffy2 = y3 + hi ; }
  offx = x3 - x1 ;

  if( y3 <= y1 ){
    linefrom = imor + y1 * winwi ;
    lineto   = imor + y3 * winwi ;
    if( x3 <= x1 ){
      for( y=y1 ; y <= y2 ; y++, linefrom+=winwi, lineto+=winwi ){
        for( x=x1 ; x <=x2 ; x++ ){ lineto[x+offx] = linefrom[x] ; }
        } }
    else {
      for( y=y1 ; y <= y2 ; y++, linefrom+=winwi, lineto+=winwi ){
        for( x=x2 ; x >=x1 ; x-- ){ lineto[x+offx] = linefrom[x] ; }
        } }
    }
  else {
    linefrom = imor + y2 * winwi ;
    lineto   = imor + (y3 + hi) * winwi ;
    if( x3 <= x1 ){
      for( y=y2 ; y >= y1 ; y--, linefrom-=winwi, lineto-=winwi ){
        for( x=x1 ; x <=x2 ; x++ ){ lineto[x+offx] = linefrom[x] ; }
        } }
    else {
      for( y=y2 ; y >= y1 ; y--, linefrom-=winwi, lineto-=winwi ){
        for( x=x2 ; x >=x1 ; x-- ){ lineto[x+offx] = linefrom[x] ; }
        } }
    }
  return(0);
  }

/*----------------- SaveArea ------------------------------------------------------------*/

int SaveArea( int x1, int y1, int x2, int y2, uchar* filename ){
  /* argument values have already been checked by ArgVal() */
  int		 size, areawi, areahi		 ;
  void*		 buf			 ;
  imoheader*	 pih			 ;
  T(orgb*)	 imor2, fromline, toline ;
  int		 w, x, y		 ;
  int		 ck, fh, written, wrote	 ;

  if( x2 < x1 || y2 < y1 ){ e("invalid coords"); return(-1); }
  if( filename == NULL   ){ e("no filename");    return(-1); }

  areawi = x2 - x1 + 1 ; areahi = y2 - y1 + 1 ;
  size = sizeof(imoheader) + areahi * areawi * sizeof(orgb) ;
  buf = alloca( size );

  pih = (imoheader*) buf ;
  pih->sig[0]='i'; pih->sig[1]='m'; pih->sig[2]='o'; pih->sig[3]='0';
  pih->byteorder = 0x01020304 ;
  pih->wi = areawi ; pih->hi = areahi ;
  pih->offx = pih->offy = 0 ;
  pih->scalex = pih->scaley = (float) 1.0 ;

  imor2 = (orgb*) (buf + sizeof(imoheader)) ;
  fromline = imor + y1 * winwi ;
  toline   = imor2 ; 
  for( y=y1 ; y<=y2 ; y++, fromline+=winwi, toline+=areawi ){
    for( x=x1, w=0 ; x<=x2 ; x++, w++ ){ toline[w] = fromline[x] ; }
    }

  filename = (uchar*) QualifyFilename(filename);
  ck = FileExists(filename,NULL);
  if( ck > 1 ){ e("file %s exists and is not an ordinary file",filename); return(-1); }
  fh = open( (char*)filename, O_WRONLY | O_CREAT, 0660 );
  if( fh < 0 ){ es("nok open %s", filename ); return(-1); }
  for( written = 0 ; written < size ; written += wrote ){
    wrote = write( fh, buf + written, size - written );
    if( wrote < 0 ){ es("nok write"); close(fh); return(-1); }
    /* failure could be caused by EINTR, but i don't care */
    }
v("wrote %d bytes to file %s", wrote, filename );
  if( 0 != close(fh) ){ es("nok close %s", filename ); }
  return(0);
  }

/*----------------- OverlayOrgb ---------------------------------------------------------*/
/* is not used in this file, is only used in text.c, but is generally usefull. */
/* HOWEVER it is faster if povl->r * povlo etc are precomputed. so maybe not so usefull. */

void OverlayOrgb( orgb* porig, orgb* povl ){ /* CAN BE INLINED ! */
  /* puts newly computed value into *porig */
  orgb			 orig		 ;
  register uchar	 povlo, npovlo	 ;
  register ushort	 t		 ;

  if( povl->o == 0 ){ return ; }
  if( povl->o == 0xFF || porig->o == 0 ){ *porig = *povl ; return ; }
  orig = *porig ;
  povlo = povl->o ; npovlo = 0xFF - povlo ;
  t = povl->r * povlo + orig.r * npovlo ; t += t >> 8 ; orig.r = t >> 8 ;
  t = povl->g * povlo + orig.g * npovlo ; t += t >> 8 ; orig.g = t >> 8 ;
  t = povl->b * povlo + orig.b * npovlo ; t += t >> 8 ; orig.b = t >> 8 ;
  t = 0xFF    * povlo + orig.o * npovlo ; t += t >> 8 ; orig.o = t >> 8 ;
  *porig = orig ;
  }
