/* layer.c : everything that works with subpixel layer */

#include "defines.h"		 /* gnu extensions, errordef, etc	 */
#include "myincludesT.h"	 /* orgb				 */
#include <math.h>		 /* sin()				 */
#include <Xlib.h>		 /* XImage				 */
#include "myincludes.h"		 /* ckalloc()				 */
#include "x.h"			 /* RenderXimage(), OrgbToPixval()	 */

/*----------------- declares of functions in this file ----------------------------------*/
/* (superfluous) */

void		 InitLayerGlobals(	 void						 );
void		 InitLayer(		 void						 );
void		 ClearLayer(		 void						 );
int		 LayerToImobuf(		 void						 );
static void	 SetBits(		 int isy, int isx1, int isx2, int erase		 );
int		 DrawLinesegToLayer(	 dob x1, dob y1, dob x2, dob y2, int erase	 );
void		 DrawDiscToLayer(	 dob x,  dob y,  dob r,  int erase		 );
void		 DrawCircleToLayer(	 dob x,  dob y,  dob r,  int erase		 );
void		 DrawArcToLayer(	 dob x,  dob y,  dob r,  dob a1, dob a2, int erase );
void		 DumpSubpixels(		 int x1, int y1, int x2, int y2			 );

/*----------------- public --------------------------------------------------------------*/

/*----------------- extern --------------------------------------------------------------*/

extern int	 verbose				 ; /* myincludes.c */
extern int	 imoaffx1, imoaffx2, imoaffy1, imoaffy2	 ; /* main.c	 */
extern int	 winwi, winhi				 ; /* main.c	 */
extern orgb	 drawcolor				 ; /* main.c	 */
extern orgb*	 imor					 ; /* imobuf.c	 */
extern dob	 drawwidth				 ; /* main.c	 */
extern XImage*	 pxim					 ; /* x.c	 */
extern int	 ximgpixelsize				 ; /* x.c	 */
extern ulong	 (*OrgbToPixval)( orgb o )		 ; /* call as OrgbToPixval(o)	 */

/*---------------- filescope globals ----------------------------------------------------*/

static uchar*	 sub					 ; /* bit array for layer */
static int	 sublinesize				 ; /* winwi * 2 */
static int	 subsize				 ; /* sublinesize * subhi */
static int	 winwi4, winhi4, winwi4mo, winhi4mo	 ;
static dob	 dwinwi4, dwinhi4			 ;
static dob	 dwinwi4mo, dwinhi4mo			 ;
static dob	 dwinwi, dwinhi				 ;
static dob	 lesux, risux, tosuy, bosuy		 ;
static int	 layaffx1, layaffx2, layaffy1, layaffy2	 ;
						 /* rangeof affected layersubpixel coords*/
static int	 nibbits[] = { 0, 1, 1,	2, 1, 2, 2, 3, 1, 2, 2, 3,  2,  3,  3,  4 } ;
			 /*    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 11, 12, 13, 14, 15 */
static char*	 nibblestring[] = { /* DEBUG */
		 " 0000", " 0001", " 0010", " 0011", " 0100", " 0101", " 0110", " 0111",
		 " 1000", " 1001", " 1010", " 1011", " 1100", " 1101", " 1110", " 1111" };

/*================= DRAW ROUTINES ==========================================================*/

/*---------------------------------------------------------------------------------------*/

void InitLayerGlobals( void ){
  lesux = -0.375 ; risux = winwi - 0.625 ;
  tosuy = -0.375 ; bosuy = winhi - 0.625 ;
  winwi4 = winwi << 2 ;
  winhi4 = winhi << 2 ;
  winwi4mo = winwi4 - 1 ;
  winhi4mo = winhi4 - 1 ;
  dwinwi = (dob) winwi ;
  dwinhi = (dob) winhi ;
  dwinwi4 = dwinwi * 4.0 ;
  dwinhi4 = dwinhi * 4.0 ;
  dwinwi4mo = dwinwi4 - 1.0 ;
  dwinhi4mo = dwinhi4 - 1.0 ;
  }

/*---------------------------------------------------------------------------------------*/

void InitLayer( void ){
  sublinesize = winwi >> 1 ;
  subsize = sublinesize * winhi4 ;
  sub = (uchar*) ckalloc( subsize );
  memset( sub, 0, subsize );
  layaffx1 = winwi4 ; layaffx2 = -1 ;
  layaffy1 = winhi4 ; layaffy2 = -1 ;
  }

/*---------------------------------------------------------------------------------------*/

void ClearLayer( void ){
  /* memset is only a millisecond, not modifying it for now */
  memset( sub, 0, subsize );
  layaffx1 = 0 ; layaffx2 = winwi4 - 1 ;
  layaffy1 = 0 ; layaffy2 = winhi4 - 1 ;
  }

/*---------------------------------------------------------------------------------------*/

void DumpSubpixels( int x1, int y1, int x2, int y2 ){ /* DEBUG */
  T(uchar*)	 psb, ps, pso, p ;
  int		 bytelen	 ;
  int		 y, i		 ;
  uchar		 b		 ;

  if( x1 & 1 ){ x1-- ; }
  v("x1=%d y1=%d x2=%d y2=%d", x1, y1, x2, y2 );
  bytelen = ((x2 - x1 + 1) >> 1) + 1 ;
  v("bytelen=%d", bytelen );
  for( y=y1 ; y <= y2 ; y++ ){
    psb = sub + (y << 2) * sublinesize + (x1 >> 1) ;
    for( i=0, ps=psb ; i < 4 ; i++, ps+=sublinesize ){
      pso = ps + bytelen ;
      for( p=ps ; p < pso ; p++ ){ b = *p ;
        /*v("sub=%p sublinesize=%d p=%p b=%X", sub, sublinesize, p, b );*/
        printf("%s%s", nibblestring[b>>4], nibblestring[b&0xF] );
        }
      printf("\n");
      }
    printf("\n");
    }
  }

/*---------------------------------------------------------------------------------------*/

int LayerToImobuf( void ){ /* also clears layer, because layer must be added exactly once */
  int		 x, y, i		 ;
  orgb*		 pimgline		 ;
  uchar*	 psubline		 ;
  orgb*		 po			 ;
  T(uchar*)	 psub, ps		 ;
  u8		 u, v, w		 ;
  orgb		 R			 ; /* result */
  orgb		 o			 ; /* old value (of orgb in imor) */
  int		 dco12			 ;
  int		 effdrawo12, efforigo12	 ;
  int		 reso12, INVreso11	 ;
  int		 firsty, lasty, firstx, lastx ;

  if( drawcolor.o == 0 ){			 /* would cause divide by zero */
    e("drawcolor is transparent");
    ClearLayer();
    layaffx1 = winwi4 ; layaffx2 = -1 ;
    layaffy1 = winhi4 ; layaffy2 = -1 ;
    return(0);
    }

  /* precomputed constants for speedup : */
  dco12 = (drawcolor.o<<4)|(drawcolor.o>>4) ;

  firstx = (layaffx1 >> 2) - 1 ; if( firstx <  0     ){ firstx = 0         ; }
  lastx  = (layaffx2 >> 2) + 1 ; if( lastx  >= winwi ){ lastx  = winwi - 1 ; }
  firsty = (layaffy1 >> 2) - 1 ; if( firsty <  0     ){ firsty = 0         ; }
  lasty  = (layaffy2 >> 2) + 1 ; if( lasty  >= winhi ){ lasty  = winhi - 1 ; }

  if( imoaffx1 > firstx ){ imoaffx1 = firstx ; }
  if( imoaffx2 < lastx  ){ imoaffx2 = lastx  ; }
  if( imoaffy1 > firsty ){ imoaffy1 = firsty ; }
  if( imoaffy2 < lasty  ){ imoaffy2 = lasty  ; }

  y = firsty ;
  pimgline = imor + firsty * winwi ;
  psubline = sub  + firsty * (sublinesize << 2) ;
  for( ; y <= lasty ; y++, pimgline += winwi, psubline += (sublinesize << 2) ){
    if( firstx & 1 ){ firstx-- ; }
    x    = firstx ;
    po   = pimgline + firstx ;
    psub = psubline + (firstx>>1) ;
    for( ; x <= lastx ; x += 2, psub++, po += 2 ){				 /* 53 */
      v = w = 0 ;								 /*  2 */
      for( i=0, ps=psub ; i < 4 ; i++, ps+=sublinesize ){ u = *ps ;
        if( u == 0 ){ continue ; }
        v += nibbits[ u >> 4  ] ;
        w += nibbits[ u & 0xF ] ;
        }									 /* 288 */
      if( v == 0 && w == 0 ){ continue ; }					 /*  34 */
      if( v != 0 ){								 /*  17 */
        o = *po ;								 /*   4 */
        if( v == 16 && (o.o == 0 || drawcolor.o == 255) ){ R = drawcolor ; }
        else {
          effdrawo12 = (dco12 * ((v<<8) | (v<<4) | v)) >> 12 ;			 /* 44 */
          efforigo12 = ((0xFFF - effdrawo12) * ((o.o<<4)|(o.o>>4)) ) >> 12 ;	 /* 57 */
          reso12 = (effdrawo12 + efforigo12) ;					 /* 17 */
	  R.o = reso12 >> 4 ;							 /*  4 */
          INVreso11  = 8384511 / reso12 ;					 /* 87 */
          R.r = ((drawcolor.r * effdrawo12 + o.r * efforigo12) * INVreso11) >> 23 ; /* 80 */
          R.g = ((drawcolor.g * effdrawo12 + o.g * efforigo12) * INVreso11) >> 23 ;
          R.b = ((drawcolor.b * effdrawo12 + o.b * efforigo12) * INVreso11) >> 23 ;
          }
        *po = R ;								 /*   4 */
        }									 /* 357 */
      if( w != 0 ){
        o = *(po+1) ;
        if( w == 16 && (o.o == 0 || drawcolor.o == 255) ){ R = drawcolor ; }
        else {
          effdrawo12 = (dco12 * ((w<<8) | (w<<4) | w)) >> 12 ;			 /* 44 */
          efforigo12 = ((0xFFF - effdrawo12) * ((o.o<<4)|(o.o>>4)) ) >> 12 ;	 /* 57 */
          reso12 = (effdrawo12 + efforigo12) ;					 /* 17 */
	  R.o = reso12 >> 4 ;							 /*  4 */
          INVreso11  = 8384511 / reso12 ;					 /* 87 */
          R.r = ((drawcolor.r * effdrawo12 + o.r * efforigo12) * INVreso11) >> 23 ; /* 80 */
          R.g = ((drawcolor.g * effdrawo12 + o.g * efforigo12) * INVreso11) >> 23 ;
          R.b = ((drawcolor.b * effdrawo12 + o.b * efforigo12) * INVreso11) >> 23 ;
          }
        *(po+1) = R ;
        }									 /* 357 */
      } /* next x */
    } /* next y */
  ClearLayer();
  layaffx1 = winwi4 ; layaffx2 = -1 ;
  layaffy1 = winhi4 ; layaffy2 = -1 ;
  return(0);
  } /* end function */							/* wi * hi * 552 */

/*---------------------------------------------------------------------------------------*/

static void SetBits( int isy, int isx1, int isx2, int erase ){
  /* args are : linix, firstix, lastix */				 /* 34 */
  int		 bytix1, bytix2		 ;
  uchar		 mask1, mask2		 ;
  T(uchar*)	 pline, pu1, pu2, pu	 ;

  /*v("isy=%d isx1=%d isx2=%d", isy, isx1, isx2 );*/
  bytix1 = isx1 >> 3 ; mask1 = 0xFF >> (isx1 & 7) ;			 /* 17 */
  bytix2 = isx2 >> 3 ; mask2 = 0xFF << (7 - (isx2 & 7)) ;		 /* 33 */
  pline = sub + isy * sublinesize ;					 /* 36 */
  if( ! erase ){
    if( bytix1 == bytix2 ){ *(pline + bytix1) |= (mask1 & mask2) ; }	 /* 17 : +27 */
    else {
      pu1 = pline + bytix1 ;						 /* 17 */
      pu2 = pline + bytix2 ;						 /* 17 */
      *pu1 |= mask1 ;							 /*  4 */
      for( pu=pu1+1 ; pu < pu2 ; pu++ ){ *pu = 0xFF ; }			 /* 23 N */
      *pu2 |= mask2 ;							 /*  4 */
      }									 /* branch=88 */
    }
  else { /* erase */
    if( bytix1 == bytix2 ){ *(pline + bytix1) &= ~(mask1 & mask2) ; }	 /* 17 : +27 */
    else {
      pu1 = pline + bytix1 ;						 /* 17 */
      pu2 = pline + bytix2 ;						 /* 17 */
      *pu1 &= ~mask1 ;							 /*  4 */
      for( pu=pu1+1 ; pu < pu2 ; pu++ ){ *pu = 0 ; }			 /* 23 N */
      *pu2 &= ~mask2 ;							 /*  4 */
      }									 /* branch=88 */
    }
  }									 /* avg=175 */

/*----------------- DrawLinesegToLayer --------------------------------------------------*/

/* actually draws a rectangular block */

int DrawLinesegToLayer( dob x1, dob y1, dob x2, dob y2, int erase ){
  /* returns 0 for success, -1 for error */			 /* 2088 + N*517 */
  dob	 minx, maxx, miny, maxy			 ;
  dob	 minsx, maxsx, minsy, maxsy		 ;
  dob	 dx, dy					 ;
  int	 iminsx, imaxsx, iminsy, imaxsy		 ;
  int	 isy					 ;
  dob	 dydx, dxdy, hyp, hlw, hlwx, hlwy	 ;
  dob	 sya, syb, syc, syd			 ;
  dob	 sxa, sxb, sxc, sxd			 ;
  dob	 disya, disyb, disyc, disyd, disy	 ;
  int	 isya					 ;
  dob	 sx1, sx2				 ;
  int	 i					 ;

  /*v("x1=%g y1=%g x2=%g y2=%g", x1, y1, x2, y2 );*/

  dy = y2 - y1 ; dx = x2 - x1 ;							 /*  38 */
  if( x1 < x2 ){ minx = x1 ; maxx = x2 ; } else { minx = x2 ; maxx = x1 ; }	 /*  80 */
  if( y1 < y2 ){ miny = y1 ; maxy = y2 ; } else { miny = y2 ; maxy = y1 ; }	 /*  80 */

  /*v("dy=%g dx=%g", dy, dx );*/

  if( fabs( dy ) < 0.1 || fabs( dx ) < 0.1 ){ /* horizontal or vertical line */
    /*v("horvert abs(dy)=%g abs(dx)=%g",fabs(dy),fabs(dx));*/
    minsx = minx * 4.0 + 1.5 ; maxsx = maxx * 4.0 + 1.5 ;
    minsy = miny * 4.0 + 1.5 ; maxsy = maxy * 4.0 + 1.5 ;
    if(   abs( dy ) < 0.1 ){ minsy -= drawwidth * 2.0 ; maxsy += drawwidth * 2.0 ; }
    elif( abs( dx ) < 0.1 ){ minsx -= drawwidth * 2.0 ; maxsx += drawwidth * 2.0 ; }
    /*v("minsx=%g maxsx=%g minsy=%g maxsy=%g", minsx, maxsx, minsy, maxsy );*/

    /* FOLLOWING REALLY NEED ceil() and floor() instead of adding 0.999 or nothing */
    iminsx = (int) (minsx + 0.9999) ; if( iminsx <  0      ){ iminsx = 0 ; }
    iminsy = (int) (minsy + 0.9999) ; if( iminsy <  0      ){ iminsy = 0 ; }
    imaxsx = (int) maxsx            ; if( imaxsx >= winwi4 ){ imaxsx = winwi4mo ; }
    imaxsy = (int) maxsy            ; if( imaxsy >= winhi4 ){ imaxsy = winhi4mo ; }

    i = minsx ; if( layaffx1 > i ){ layaffx1 = i ; }
    i = maxsx ; if( layaffx2 < i ){ layaffx2 = i ; }
    i = minsy ; if( layaffy1 > i ){ layaffy1 = i ; }
    i = maxsy ; if( layaffy2 < i ){ layaffy2 = i ; }

    if( iminsx > winwi4mo || imaxsx < 0 || iminsy > winhi4mo || imaxsy < 0 ){
      w("nothing to draw"); }
    else {
      for( isy=iminsy ; isy <= imaxsy ; isy++ ){
        /*v("setting bits : isy=%d iminsx=%d imaxsx=%d", isy, iminsx, imaxsx );*/
        SetBits( isy, iminsx, imaxsx, erase );
        }
      }
    return(0); /* prevent divide by zero furtheron */
    }

  /* SLANTING LINE */
  dydx = dy / dx ; dxdy = dx / dy ;						 /* 170 */
  hyp  = sqrt(dx*dx + dy*dy) ;							 /* 166 */
  hlw  = drawwidth * 0.5 ;							 /*  20 */
  hlwx = fabs( hlw * dy / hyp );						 /*  96 */
  hlwy = fabs( hlw * dx / hyp );						 /*  96 */
  /*v("dydx=%g dxdy=%g hyp=%g hlw=%g hlwx=%g hlwy=%g", dydx, dxdy, hyp, hlw, hlwx, hlwy );*/

  /* a, b, c, d are cornerpoints, a is top and d is bottom corner,			 */
  /*   b is other corner on same line-end as a, c is other corner on same line-end as d. */
  /* Y COORD INCREASES DOWNWARD !							 */
  /* -dxdy is really dydx of boundary on ends of line					 */

  sya = (miny - hlwy) * 4.0 + 1.5 ; syb = (miny + hlwy) * 4.0 + 1.5 ;
  syc = (maxy - hlwy) * 4.0 + 1.5 ; syd = (maxy + hlwy) * 4.0 + 1.5 ;

  if( dydx > 0.0 ){ /* DESCENDING line */
    sxa = (minx + hlwx) * 4.0 + 1.5 ; sxb = (minx - hlwx) * 4.0 + 1.5 ;
    sxc = (maxx + hlwx) * 4.0 + 1.5 ; sxd = (maxx - hlwx) * 4.0 + 1.5 ;

    i = (int) sxb ; if( layaffx1 > i ){ layaffx1 = i ; }
    i = (int) sxc ; if( layaffx2 < i ){ layaffx2 = i ; }
    i = (int) sya ; if( layaffy1 > i ){ layaffy1 = i ; }
    i = (int) syb ; if( layaffy2 < i ){ layaffy2 = i ; }
    }
  else { /* ascending line */
    sxa = (maxx - hlwx) * 4.0 + 1.5 ; sxb = (maxx + hlwx) * 4.0 + 1.5 ;
    sxc = (minx - hlwx) * 4.0 + 1.5 ; sxd = (minx + hlwx) * 4.0 + 1.5 ;

    i = (int) sxc ; if( layaffx1 > i ){ layaffx1 = i ; }
    i = (int) sxb ; if( layaffx2 < i ){ layaffx2 = i ; }
    i = (int) sya ; if( layaffy1 > i ){ layaffy1 = i ; }
    i = (int) syd ; if( layaffy2 < i ){ layaffy2 = i ; }
    }

  /*v("sa=(%g;%g) sb=(%g;%g) sc=(%g;%g) sd=(%g;%g)", sxa, sya, sxb, syb, sxc, syc, sxd, syd );*/

  disya = ceil( sya ); disyb = ceil( syb );
  disyc = ceil( syc ); disyd = ceil( syd );
  if( disya < 0.0 ){ disya = 0.0 ; } if( disya > dwinhi4 ){ disya = dwinhi4 ; }
  if( disyb < 0.0 ){ disyb = 0.0 ; } if( disyb > dwinhi4 ){ disyb = dwinhi4 ; }
  if( disyc < 0.0 ){ disyc = 0.0 ; } if( disyc > dwinhi4 ){ disyc = dwinhi4 ; }
  if( disyd < 0.0 ){ disyd = 0.0 ; } if( disyd > dwinhi4 ){ disyd = dwinhi4 ; }
  isya = (int) (disya + 0.5) ;							 /* 113 */
  /*v("disya=%g disyb=%g disyc=%g disyd=%g dwinhi4=%g", disya, disyb, disyc, disyd, dwinhi4 );*/

  if( dydx > 0.0 ){ /* DESCENDING LINE */
    if( disyc <= disyb ){ /* very short or nearly horizontal line */		 /*  12 */
      for( disy=disya, isy=isya ; disy < disyc ; disy+=1.0, isy++ ){		 /* .28 */
        sx1 = sxa - dydx * (disy - sya) + 0.999 ; if( sx1 > dwinwi4mo ){ continue ; } /*CEIL */
        sx2 = sxa + dxdy * (disy - sya)         ; if( sx2 < 0.0       ){ continue ; } /*FLOOR*/
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      for( ; disy < disyb ; disy+=1.0, isy++ ){
        sx1 = sxa - dydx * (disy - sya) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxc - dydx * (disy - syc)         ; if( sx2 < 0.0       ){ continue  ; }
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      for( ; disy < disyd ; disy+=1.0, isy++ ){					 /*  16 */
        sx1 = sxb + dxdy * (disy - syb) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxc - dydx * (disy - syc)         ; if( sx2 < 0.0       ){ continue  ; }
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
		 /* 3ass=3 2dcmp=24 2dadd=4 2cast=186 1call=34 1func=175 total=423 */
        }									 /* 517 */
      }
    else { /* has a zone of parallel boundaries in dydx direction */
      for( disy=disya, isy=isya ; disy < disyb ; disy+=1.0, isy++ ){
        sx1 = sxa - dydx * (disy - sya) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxa + dxdy * (disy - sya)         ; if( sx2 < 0.0       ){ continue  ; }
        /*if( isy>=0 && isy<40 ){ v("setting bits, isy=%d disy=%g sx1=%g sx2=%g",isy,disy,sx1,sx2);}*/
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      for( ; disy < disyc ; disy+=1.0, isy++ ){
        sx1 = sxb + dxdy * (disy - syb) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxa + dxdy * (disy - sya)         ; if( sx2 < 0.0       ){ continue  ; }
        /* v("setting bits, isy=%d disy=%g sx1=%g sx2=%g", isy, disy, sx1, sx2 );*/
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      for( ; disy < disyd ; disy+=1.0, isy++ ){
        sx1 = sxb + dxdy * (disy - syb) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxc - dydx * (disy - syc)         ; if( sx2 < 0.0       ){ continue  ; }
        /*if( isy>=0 && isy<40 ){ v("setting bits, isy=%d disy=%g sx1=%g sx2=%g",isy,disy,sx1,sx2);}*/
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      }
    }
  else { /* ASCENDING LINE */
    if( disyc <= disyb ){ /* very short or nearly horizontal line */
      for( disy=disya, isy=isya ; disy < disyc ; disy+=1.0, isy++ ){
        sx1 = sxa + dxdy * (disy - sya) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxa - dydx * (disy - sya)         ; if( sx2 < 0.0       ){ continue  ; }
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      for( ; disy < disyb ; disy+=1.0, isy++ ){
        sx1 = sxc - dydx * (disy - syc) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxa - dydx * (disy - sya)         ; if( sx2 < 0.0       ){ continue  ; }
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      for( ; disy < disyd ; disy+=1.0, isy++ ){					 /*  16 */
        sx1 = sxc - dydx * (disy - syc) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxb + dxdy * (disy - syb)         ; if( sx2 < 0.0       ){ continue  ; }
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }									 /* 517 */
      }
    else { /* has a zone of parallel boundaries in dydx direction */
      for( disy=disya, isy=isya ; disy < disyb ; disy+=1.0, isy++ ){
        sx1 = sxa + dxdy * (disy - sya) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxa - dydx * (disy - sya)         ; if( sx2 < 0.0       ){ continue  ; }
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      for( ; disy < disyc ; disy+=1.0, isy++ ){
        sx1 = sxa + dxdy * (disy - sya) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxb + dxdy * (disy - syb)         ; if( sx2 < 0.0       ){ continue  ; }
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      for( ; disy < disyd ; disy+=1.0, isy++ ){
        sx1 = sxc - dydx * (disy - syc) + 0.999 ; if( sx1 > dwinwi4mo ){ continue  ; }
        sx2 = sxb + dxdy * (disy - syb)         ; if( sx2 < 0.0       ){ continue  ; }
        SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
        }
      }
    }
  return(0);
  }

/*----------------- DrawDiscToLayer -----------------------------------------------------*/

/* most often used to draw endpoints and continuation points on a line */

void DrawDiscToLayer( dob x, dob y, dob r, int erase ){
  dob stop, sbot	 ;
  dob disy		 ;
  dob distop, disbot	 ;
  int isy		 ;
  dob sxc, syc		 ; /* center, in subpixel coords */
  dob r4sq		 ;
  dob dy, absdx		 ;
  dob sx1, sx2		 ;
  int i			 ;

  if( x+r < lesux || x-r > risux || y+r < tosuy || y-r > bosuy ){ w("nothing to draw"); return ; }
  syc = y * 4.0 + 1.5 ;								 /*  22 */
  sxc = x * 4.0 + 1.5 ;								 /*  22 */
  r4sq = r * r * 16.0 ;								 /*  17 */
  stop = (y - r) * 4.0 + 1.5 ; distop = (stop < 0.0      )? 0.0       : ceil( stop) ; /* 204 */
  sbot = (y + r) * 4.0 + 1.5 ; disbot = (sbot > dwinhi4mo)? dwinhi4mo : floor(sbot) ; /* 204 */
  /*v("sxc=%g syc=%g r4sq=%g stop=%g sbot=%g", sxc, syc, r4sq, stop, sbot );*/

  i = (int)(sxc - r*4.0) ; if( layaffx1 > i ){ layaffx1 = i ; }
  i = (int)(sxc + r*4.0) ; if( layaffx2 < i ){ layaffx2 = i ; }
  i = (int)(stop       ) ; if( layaffy1 > i ){ layaffy1 = i ; }
  i = (int)(sbot       ) ; if( layaffy2 < i ){ layaffy2 = i ; }

  for( disy=distop, isy=(int)disy ; disy <= disbot ; disy+=1.0, isy++ ){	 /*121+21 */
    dy    = disy - syc ;							 /*  19 */
    if( r4sq < dy * dy ){ absdx = 0.0 ;
      v("taking sqrt of negative number. absdx");
      /* this happens due to truncation to integer. it is harmless if absdx is set to 0 */
      }
    else { absdx = sqrt( r4sq - dy * dy ); }					 /* 165 */
    sx1 = sxc - absdx + 0.999 ; if( sx1 > dwinwi4mo ){ continue ; } /* CEIL  */
    sx2 = sxc + absdx         ; if( sx2 < 0.0       ){ continue ; } /* FLOOR */
    SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase ); /* */
    }										 /* 580 */
  }									 /* 646 + N*580 */

/*----------------- DrawCircleToLayer --------------------------------------------------*/

void DrawCircleToLayer( dob x, dob y, dob r, int erase ){
  dob	 ri, ro		 ;
  dob	 r4sqi, r4sqo	 ;
  dob	 dy		 ;
  dob	 syc, sxc	 ;
  dob	 stopo, distopo	 ;
  dob	 sboto, disboto	 ;
  dob	 stopi, distopi	 ;
  dob	 sboti, disboti	 ;
  dob	 disy		 ;
  int	 isy		 ;
  dob	 absdxi, absdxo	 ;
  dob	 sx1, sx2	 ;
  int	 i		 ;

  ri = r - drawwidth * 0.5 ;							 /* 25 */
  ro = r + drawwidth * 0.5 ;							 /* 25 */
  if( x+ro < lesux || x-ro > risux || y+ro < tosuy || y-ro > bosuy ){		 /* 62 */
    w("nothing to draw"); return ;
    }
  if( ri <= 0.0 ){ DrawDiscToLayer( x, y, ro, erase ); return ; }
  syc = y * 4.0 + 1.5 ;								 /* 22 */
  sxc = x * 4.0 + 1.5 ;								 /* 22 */
  r4sqi = ri * ri * 16.0 ;							 /* 26 */
  r4sqo = ro * ro * 16.0 ;							 /* 26 */
  stopo = (y - ro) * 4.0 + 1.5 ; distopo = (stopo < 0.0      )? 0.0      : ceil( stopo);/*231*/
  sboto = (y + ro) * 4.0 + 1.5 ; disboto = (sboto > dwinhi4mo)? dwinhi4mo: floor(sboto);/*231*/
  stopi = (y - ri) * 4.0 + 1.5 ; distopi = (stopi < 0.0      )? 0.0      : ceil( stopi);/*231*/
  sboti = (y + ri) * 4.0 + 1.5 ; disboti = (sboti > dwinhi4mo)? dwinhi4mo: floor(sboti);/*231*/
  /*v("stopo=%g stopi=%g sboti=%g sboto=%g", stopo, stopi, sboti, sboto );*/
  /*v("distopo=%g distopi=%g disboti=%g diboto=%g", distopo, distopi, disboti, disboto );*/

  i = (int)(sxc - r*4.0) ; if( layaffx1 > i ){ layaffx1 = i ; }
  i = (int)(sxc + r*4.0) ; if( layaffx2 < i ){ layaffx2 = i ; }
  i = (int)(stopo      ) ; if( layaffy1 > i ){ layaffy1 = i ; }
  i = (int)(sboto      ) ; if( layaffy2 < i ){ layaffy2 = i ; }

  for( disy=distopo, isy=(int)(disy + 0.5) ; disy < distopi ; disy+=1.0, isy++ ){/*106+21 */
    dy = disy - syc ;								 /*  19 */
    if( r4sqo < dy * dy ){
      absdxo = 0.0 ;
      w("taking sqrt of negative number. absdxo"); } /* DEFENSIVE. REMOVELATER. */
    else { absdxo = sqrt( r4sqo - dy * dy ); }					 /* 165 */
    sx1 = sxc - absdxo + 0.999 ; if( sx1 > dwinwi4mo ){ continue ; }
    sx2 = sxc + absdxo         ; if( sx2 < 0.0       ){ continue ; }
    /*v("setting bits @top : isy=%d sx1=%g sx2=%g", isy, sx1, sx2 );*/
    SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
    }
  for( ; disy <= disboti ; disy+=1.0, isy++ ){					 /*  21 */
    dy = disy - syc ;								 /*  19 */
    if( r4sqo < dy * dy ){
      absdxo = 0.0 ;
      w("taking sqrt of negative number. 517"); } /* DEFENSIVE. REMOVELATER. */
    else { absdxo = sqrt( r4sqo - dy * dy ); }					 /* 165 */
    if( r4sqi < dy * dy ){
      absdxi = 0.0 ;
      w("taking sqrt of negative number. absdxi"); } /* DEFENSIVE. REMOVELATER. */
    else { absdxi = sqrt( r4sqi - dy * dy ); }					 /* 165 */
    /*v("r4sqo=%g r4sqi=%g dy=%g absdxo=%g absdxi=%g", r4sqo, r4sqi, dy, absdxo, absdxi );*/
    sx1 = sxc - absdxo + 0.999 ; if( sx1 > dwinwi4mo ){ continue ; }
    sx2 = sxc - absdxi         ;
    if( sx2 >= 0.0 ){
      /*v("setting bits @midl : isy=%d sx1=%g sx2=%g", isy, sx1, sx2 );*/
      SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
      }										 /* */
    sx1 = sxc + absdxi + 0.999 ; if( sx1 > dwinwi4mo ){ continue ; }
    sx2 = sxc + absdxo         ; if( sx2 < 0.0       ){ continue ; }
    /*v("setting bits @midr : isy=%d sx1=%g sx2=%g", isy, sx1, sx2 );*/
    SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase ); /* 375 */
    }										 /* 1120 */
  for( ; disy <= disboto ; disy+=1.0, isy++ ){
    dy    = disy - syc ;
    if( r4sqo < dy * dy ){
      absdxo = 0.0 ;
      w("taking sqrt of negative number. 539"); } /* DEFENSIVE. REMOVELATER. */
    else { absdxo = sqrt( r4sqo - dy * dy ); }
    sx1 = sxc - absdxo + 0.999 ; if( sx1 > dwinwi4mo ){ continue ; }
    sx2 = sxc + absdxo         ; if( sx2 < 0.0       ){ continue ; }
    /*v("setting bits @bot : isy=%d sx1=%g sx2=%g", isy, sx1, sx2 );*/
    SetBits( isy, (sx1<0.0)?0:(int)sx1, (sx2>dwinwi4mo)?winwi4mo:(int)sx2, erase );
    }
  }									 /* 1242 + N*1120 */

/*----------------- DrawArcToLayer -------------------------------------------------------*/

/* angles are expressed as fractions of a whole turn.		 */
/* angles are relative to 12 o'clock, and are counted clockwise. */

void DrawArcToLayer( dob x, dob y, dob r, dob a1, dob a2, int erase ){
  dob	 rad			 ; /* angle, in radians */
  dob	 arg0, arg1, arg2, arg3	 ;
  dob	 csegs			 ;
  dob	 arctot			 ; /* angle of entire arc in full turns */
  dob	 radtot			 ;
  dob	 radseg			 ; /* angle (of arcsegment) in radians */
  dob	 rada1, rada2		 ;
  dob	 lim			 ;

  /*v("args: x=%g y=%g r=%g a1=%g a2=%g", x, y, r, a1, a2 );*/
  if( r <= 0.0 ){ e("invalid radius"); return ; }
  if( fabs(a2 - a1) >= 1.0 ){ DrawCircleToLayer( x, y, r, erase ); return ; }
  if( a1 > 1.0 || a1 < 0.0 ){ a1 -= floor( a1 ); }
  if( a2 > 1.0 || a2 < 0.0 ){ a2 -= floor( a2 ); }
  /*v("modified angles : a1=%g a2=%g", a1, a2 );*/

  arctot = a2 - a1 ;
  if( arctot <= 0.0 ){ arctot += 1.0 ; }
  if( arctot <= 0.0 ){ ex("impossible"); }
  radtot = 2.0 * M_PI * arctot ;
  radseg = sqrt(1/r) ;				 /* maxlen=sqrt(r) */
  csegs  = ceil( radtot / radseg ) ;
  radseg = radtot / csegs ;
  rada1  = 2.0 * M_PI * a1 ;
  rada2  = rada1 + csegs * radseg ;

  /*v("arctot=%g radtot=%g radseg=%g csegs=%g rada1=%g rada2=%g",
    arctot, radtot, radseg, csegs, rada1, rada2);*/
  /*
    LIMITS NOT CHECKED
    drawlineseg checks limits of lineseg, and drawdisc does same,
    but both emit warnings if nothing to draw
  */
  arg2 = x + sin( rada1 ) * r ;
  arg3 = y - cos( rada1 ) * r ;
  /*v("first point : x=%g y=%g", arg2, arg3 );*/
  DrawDiscToLayer( arg2, arg3, drawwidth * 0.5, erase );

  lim = rada2 + 0.5 * radseg ;
  for( rad=rada1+radseg ; rad < lim ; rad+=radseg ){
    arg0 = arg2 ;
    arg1 = arg3 ;
    arg2 = x + sin( rad ) * r ;
    arg3 = y - cos( rad ) * r ;
    /*v("next point : x=%g y=%g", arg2, arg3 );*/
    /*NextDrawcolor();*/
    DrawLinesegToLayer( arg0, arg1, arg2, arg3, erase );
    DrawDiscToLayer( arg2, arg3, drawwidth * 0.5, erase );
    /*LayerToImobuf();*/
    }
  }

/* DEBUG */
/*----------------- ShowLayer -----------------------------------------------------------*/

void SetXimagePixel( int x, int y, ulong val ){
  if( ximgpixelsize == 1 ){
    u8* pximpixel8 ;
    pximpixel8 = ((u8*)pxim->data) + winwi * y + x ;
    *pximpixel8 = (u8 ) val ;
    }
  elif( ximgpixelsize == 2 ){
    u16* pximpixel16 ;
    pximpixel16 = ((u16*)pxim->data) + winwi * y + x ;
    *pximpixel16 = (u16) val ;
    }
  else {
    u32* pximpixel32 ;
    pximpixel32 = ((u32*)pxim->data) + winwi * y + x ;
    *pximpixel32 = (u32) val ;
    }
  }

/*---------------------------------------------------------------------------------------*/

int ShowLayer( int offx, int offy ){ /* for debugging (and user insight) */
  /* args are pixelcoords */
  T(uchar*)	 puc, psubline		 ;
  int		 sx1, sx2, sy1, sy2, sx, sy	 ;
  uchar		 bit, byte			 ;
  orgb		 o				 ;
  u32		 blackval, whiteval		 ;
  int		 i, lim				 ;
  int		 ix, iy				 ;

  sx1 = offx << 2            ; if( sx1 < 0       ){ e("arg offx negative" ); return(-1); }
  sy1 = offy << 2            ; if( sy1 < 0       ){ e("arg offy negative" ); return(-1); }
  sx2 = sx1 + (winwi>>2) - 1 ; if( sx2 >= winwi4 ){ e("arg offx too large"); return(-1); }
  sy2 = sy1 + winhi      - 1 ; if( sy2 >= winhi4 ){ e("arg offy too large"); return(-1); }

  o.o=0xFF ; o.r = o.g = o.b = 0    ; blackval = OrgbToPixval(o);
  o.o=0xFF ; o.r = o.g = o.b = 0xFF ; whiteval = OrgbToPixval(o);

  for( sy=sy1, iy=0 ; sy <= sy2 ; sy++, iy++ ){
    psubline = sub + sy * sublinesize ;
    for( sx=sx1, ix=0 ; sx <= sx2 ; sx+=2, ix+=8 ){
      puc = psubline + (sx >> 1);
      byte = *puc ;
      lim = 8 ; if( sx2 - sx <= 1 ){ lim = 4 ; }
      for( i=0 ; i < lim ; i++ ){
        bit = byte & (0x80 >> i) ;
        if( bit ){ SetXimagePixel(ix+i,iy,blackval) ; }
        else     { SetXimagePixel(ix+i,iy,whiteval) ; }
        }
      }
    }
  RenderXimage();
  return(0);
  }
