#ifndef inccanvasC
#define inccanvasC

/* canvas : read draw-commands from stdin and execute them */

/*----------------- include ----------------------------------------------------------------*/

#include "defines.h"
#include "command.h"			 /* InitCommand(), HandleMessage()		 */
#include "clickdragT.h"			 /* #defines of PRESS etc.			 */
#include "clickdrag.h"			 /* *Clickdrag()				 */
#include "fill.h"			 /* FillArea(), CountArea()			 */
#include "imobuf.h"
#include "layer.h"			 /* *Layer*()					 */
#include "myincludesT.h"		 /* orgb */
#include "myincludes.h"			 /* ??? */
#include "socket.h"			 /* InitSocket(), ExitSocket(),	HandleChannels() */
#include "text.h"
#include "x.h"

#include <signal.h>
#include <stdio.h>			 /* sscanf()					 */
#include <string.h>			 /* strncmp(), memmove()			 */
#include <sys/types.h>			 /* select()					 */
#include <sys/socket.h>			 /* socket(), bind(), listen(), accept()	 */
#include <sys/un.h>			 /* struct sockaddr_un				 */
#include <sys/time.h>			 /* struct timeval				 */
#include <unistd.h>			 /* sleep(), read(), write(), close(), unlink()	 */
#include <fcntl.h>			 /* open()					 */

/*----------------- defines -------------------------------------------------------------*/

/*----------------- extern --------------------------------------------------------------*/

extern int	 verbose	 ; /* myincludes.c	 */
extern Filebuf	 imobuf		 ; /* imobuf.c		 */

/*----------------- declares ---------------------------------------------------------------*/

void		 ParseArguments( int argc, char** argv	 );
int		 main(		 int argc, char** argv	 );

/*----------------- GLOBALS -------------------------------------------------------------*/

int		 winwi, winhi	 ;
uchar*		 sockfilename	 ;
int		 dummy		 ;
orgb		 drawcolor	 = { 0xFF, 0, 0, 0 } ; /* == oblack */
double		 drawwidth	 = 1.0 ;

static uchar*	 imofilename	 ;

/*----------------- help -------------------------------------------------------------------*/

/* TODO : reduce helptext to basic syntax, make extra text page with full doc. */

char* help[] = {
  "canvas : make a window and listen for draw-commands                            ",
  "syntax : canvas (-v)* -wNUMBER -hNUMBER -sSOCKETFILE -iIMOFILE                 ",
  "  where -w and -h specify sizes of canvas (in pixels) (default 760x570)        ",
  "  and -i specifies file to read for initial content of canvas (default none)   ",
  "    which (if specified) must be an imo file.                                  ",
  "  Every -v increases verbosity.                                                ",
  "                                                                               ",
  "MODE OF DRAWING OPERATIONS :                                                   ",
  "Drawcommands draw to subpixel layer,                                           ",
  "  which for each pixel of canvas contains a 4x4 array of subpixels             ",
  "  in which each subpixel is either 0 or 1.                                     ",
  "Drawing uses value of 'drawwidth' variable, which determines linewidth,        ",
  "  and which can be set with command 'setdrawwidth'.                            ",
  "When command 'flushlayer' is given                                             ",
  "  shape of subpixel layer is combined with color of 'drawcolor' variable       ",
  "    (which can be set with command 'setdrawcolor')                             ",
  "  and added to 'imobuf', which contains opacity/red/green/blue for each pixel  ",
  "  by treating it as an overlay whose opacity                                   ",
  "    is proportional to fraction of subpixels (of pixel) that is 1              ",
  "    and is also proportional to opacity of drawcolor                           ",
  "  After this, 'flushlayer' erases content of layer.                            ",
  "'imobuf' is essentially a filebuffer, of a file in 'imo' fileformat            ",
  "  which is native fileformat of this program                                   ",
  "  and consists of a small header and an array of opacity/red/green/blue values.",
  "Contents of imobuf can be stored in a file by command 'saveas'.                ",
  "Contents of imobuf are made visible in window by command 'showimobuf'.         ",
  "Background of drawing can be from an imofile,                                  ",
  "  if this file is specified at invocation of canvas program ;                  ",
  "  If no background was specified, or imobuf has been cleared,                  ",
  "    background is fully opaque white                                           ",
  "    (in future versions this may be transparent instead).                      ",
  "Subpixel layer supports erasing,                                               ",
  "  which is same as drawing, except bits of layer are cleared instead of set.   ",
  "Imobuf does not generally support erasing,                                     ",
  "  since every addition to it is a possibly transparent overlay,                ",
  "  but drawing with drawcolor set to backgroundcolor has similar effect.        ",
  "Imobuf supports filling areas bounded by dark lines, where 'dark' means that   ",
  "  at least one of r/g/b components of color is less than half of maximum.      ",
  "Filling subpixel layer is planned for a future version.                        ",
  "                                                                               ",
  "SUPPORTED COMMANDS :                                                           ",
  "  setdrawwidth  w                                                              ",
  "  setdrawcolor  o r g b                                                        ",
  "  drawpoint     x  y                    (uses drawwidth)                       ",
  "  drawlineseg   x1 y1 x2 y2                                                    ",
  "  drawdisc      x  y  r                                                        ",
  "  drawline      x1 y1 x2 y2                                                    ",
  "  drawcircle    x  y  r                                                        ",
  "  drawarc       x  y  r  a1 a2                                                 ",
  "    angles are expressed as fractions of a whole turn, clockwise               ",
  "  erasepoint    x  y                                                           ",
  "  eraseline     x1 y1 x2 y2                                                    ",
  "  erasedisc     x  y  r                                                        ",
  "  erasecircle   x  y  r                                                        ",
  "  erasearc      x  y  r  a1 a2                                                 ",
  "  showlayer     x  y           (shows subpixel bits in window, for debugging)  ",
  "  flushlayer                   (copies layer content to imobuf, clears layer)  ",
  "  clearlayer                   (erases layercontent without affecting imobuf)  ",
  "  fillarea      x  y           (operates on imobuf, not on layer)              ",
  "    fills any adjacent points that have a light color                          ",
  "      starting with point (x,y), which must have a light color ;               ",
  "    reports number of pixels that it filled                                    ",
  "  countarea     x  y                                                           ",
  "    reports how many pixels would be filled, but does not fill anything.       ",
  "  loadfont      fontfilename                                                   ",
  "    loads a client-side font, some of which are contained in canvas package.   ",
  "  drawtext      direction x y text                                             ",
  "    draws text in direction r (right), l (left), d (down) or u (up),           ",
  "    such that origin of first glyph is at x/y                                  ",
  "  setarea       x1 y1 x2 y2            (sets area to drawcolor)                ",
  "  copyarea      x1 y1 x2 y2 x2 y3      (copies x1/y1 to x3/y3)                 ",
  "  showimobuf                           (calls imobuftoximage and renderximage) ",
  "  setaffected   x1 y1 x2 y2                                                    ",
  "    is used when content of imobuf has been changed by an external program,    ",
  "    because showimobuf only refreshes parts that it knows have changed.        ",
  "  clearimobuf                                                                  ",
  "  loadimofile   filename                                                       ",
  "  saveas        filename                                                       ",
  "  savearea      x1 y1 x2 y2 filename                                           ",
  "  storeximage   N       (allocates an extra buffer to store current image)     ",
  "  loadximage    N       (loads stored image number N, and renders it)          ",
  "  freeximagebuf N       (frees memory of image buffer number N)                ",
  "  recordclickdrag       (makes canvas start recording events)                  ",
  "  norecordclickdrag     (makes canvas stop recording events)                   ",
  "  reportclickdrag       (replies with string of all events received so far)    ",
  "  quit                                                                         ",
  "Most arguments are decimal floating point numbers in non-exponential format    ",
  "                                                                               ",
  "TIPS :                                                                         ",
  "To draw a block, use 'drawlineseg',                                            ",
  "  which draws a rectangular block in any orientation ;                         ",
  "  it's coords are like that of a line, and width of block is drawwidth.        ",
  "To draw a fluently curved line, use multiple 'drawlineseg' commands,           ",
  "  each followed by a 'drawdisc x y drawwidth/2' to ensure good connection.     ",
  "To get image into a graphics file, save it as imo with saveas or savearea      ",
  "  and then use ImoToJpeg or ImoToPng to convert them ;                         ",
  "  these programs, and also JpegToImo, are distributed with canvas program.     ",
  "                                                                               ",
  "COMMUNICATION :                                                                ",
  "canvas reads drawcommands from a local socket, by default /tmp/canvas.sock .   ",
  "For use by hand, or by programs not capable of using sockets directly,         ",
  "  there is commandline utility 'canvassock',                                   ",
  "  which sends it's (string) argument as message to socket /tmp/canvas.sock .   ",
  NULL } ;

/*----------------- ParseArguments ---------------------------------------------------------*/

void ParseArguments( int argc, char ** argv ){
  int		 i		 ;
  char*		 arg		 ; /* tmp */
  int		 ck		 ;

  /* set defaults : */
  dummy		 =   0 ;
  verbose	 =   0 ;
  winwi		 = 750 ;
  winhi		 = 570 ;
  imofilename	 = NULL ;
  sockfilename	 = (uchar*)"/tmp/canvas.sock" ;

  /* parse arguments : */
  for( i=1 ; i<argc ; i++ ){
    arg = argv[i] ;
    if( arg[0] == '-' ){
      if( arg[1] == 'v' ){ verbose++ ; continue ; }
      if( arg[1] == 'd' ){ dummy = 1 ; continue ; }
      if( arg[1] == 'i' ){ imofilename  = (uchar*) (argv[i] + 2) ; continue ; }
      if( arg[1] == 's' ){ sockfilename = (uchar*) (argv[i] + 2) ; continue ; }
      if( arg[1] == '?' ){ phe(); }
      if( arg[1] == 'w' ){
        if(ParseInt((uchar*)(arg+2),&winwi)==NULL){ e("invalid -w option"); phe(); }
        else { continue ; }
        }
      if( arg[1] == 'h' ){
        if(ParseInt((uchar*)(arg+2),&winhi)==NULL){ e("invalid -h option"); phe(); }
        else { continue ; }
        }
      ex("unrecognized option %s ; try 'canvas ?' for help", arg );
      }
    phe();
    }

  if( winwi <= 0 || winhi <= 0 ){ ex("invalid windowsize"); }
  if( imofilename != NULL ){
    imofilename = (uchar*) QualifyFilename( (char*)imofilename );
    ck = FileExists( (char*) imofilename, NULL );
    if( ck == 0 ){ ex("imofile %s does not exist", imofilename ); }
    if( ck != 1 ){ ex("imofile %s exists but is not a regular file", (char*)imofilename ); }
    }
  v("sockfilename=%s", sockfilename );
  sockfilename = (uchar*) QualifyFilename( (char*)sockfilename );
  ck = FileExists( (char*) sockfilename, NULL );
  if( ck != 0 && ck != 7 ){
    ex("sockfile %s exists but is not a socket", (char*)sockfilename ); }
  if( ck == 7 ){
    if( 0 > unlink((char*)sockfilename) ){ ex("nok unlink existing socket"); }
    }

  v("dummy=%d verbose=%d winwi=%d winhi=%d imofilename=%s sockfilename=%s",
    dummy, verbose, winwi, winhi, imofilename, sockfilename );
  }



/*----------------- main ----------------------------------------------------------------*/

int main( int argc, char** argv ){

  ParseArguments( argc, argv );		 v("did parse arguments");
  InitLayerGlobals();			 v("did init globals");
  InitCommand();			 v("did init command");
  InitX();				 v("did initx");
  MakeImobuf();				 v("did makeimobuf");
  if( imofilename != NULL ){
    GetImofileContent(imofilename);	 v("got imofilecontent");
    ImobufToXimage();			 v("did imobuftoximage");
    RenderXimage();			 v("did renderximage");
    }
  InitLayer();				 v("did init layer");
  AllocTagmap();			 v("did alloc tagmap");
  InitSocket();	 /* also makes sigpipe be ignored */
					 v("did init socket");
  InitEvents();				 v("did init events");

  HandleChannels();			 v("returned from HandleChannels");

  ExitSocket();				 v("did exit socket");
  DiscardImobuf();			 v("did discard imobuf");
  return(0);
  }


#endif
