
/*----------------- includes ---------------------------------------------------------------*/

#include "defines.h"
#include "myincludesT.h"

#include <alloca.h>	 /* alloca()				 */
#include <fcntl.h>	 /* open() */
#include <malloc.h>	 /* malloc()				 */
#include <signal.h>	 /* signal(), SIGALRM, struct sigaction	 */
#include <string.h>	 /* strcpy() etc			 */
#include <stdint.h>	 /* uint32_t etc */
#include <stdio.h>	 /* cuserid()				 */
#include <stdlib.h>	 /* exit()				 */
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>	 /* ulong, uint32_t etc */
#include <unistd.h>	 /* getcwd(), getlogin()		 */

/*----------------- define --------------------------------------------------------------*/

/*----------------- public --------------------------------------------------------------*/

int		 verbose	 ;

/*----------------- extern --------------------------------------------------------------*/

extern char*	 help[]		 ;

/*----------------- filescope globals ---------------------------------------------------*/

static int	 DoExitAtAlarm	 = 0 ;
static char*	 exitmsg	 = NULL ; /* msg @ timeout of window-mapped event */
static void	 (*funcp)(void)	 = NULL ;

/*----------------- ph ------------------------------------------------------------------*/

void ph( void ){
  int i ;
  for( i=0 ; ; i++ ){
    if( help[i] == NULL ){ break ; }
    printf("%s\n", help[i] );
    }
  }

void phe( void ){ ph(); exit(0); }

/*----------------- ckalloc -------------------------------------------------------------*/

void* ckalloc( int size ){ void * pv ;
				 if( size <= 0  ){ ex("size nonpositive"); }
  pv = malloc( size );		 if( pv == NULL ){ esx("nok malloc"); }
  return( pv );
  }

/*----------------- ckfree --------------------------------------------------------------*/

void ckfree( void* pv ){ free(pv); }

/*----------------- ParseInt -------------------------------------------------------------------*/

uchar* ParseInt( uchar* string, int* presult ){
  uchar	 dv	 ;
  int	 v	 = 0 ;
  uchar* s	 ;

  s = string ;
  for( ; (dv=(*s - '0')) <= 9 ; s++ ){ v = 10 * v + dv ; }
  if( s == string ){ e("no digits"); return(NULL); }
  *presult = v ;
  return(s);
  }

/*----------------- open_rw -------------------------------------------------------------*/

int open_rw( char* filename, int exitonfail ){
  int fh ;
  fh = open( filename, O_RDWR | O_CREAT, 0660 );
  if( fh < 0 ){
    es("failed open_rw %s", filename );
    if( exitonfail ){ exit(-1); }
    }
  return( fh );
  }

/*----------------- HandleTimeout -------------------------------------------------------*/

/* NOTE that there can only be one timeout running at a time ! */

void HandleTimeout( int signum ){
  if( signum != SIGALRM	 ){ e("wrong signum"); return ; }
  if( exitmsg != NULL	 ){ t("%s", exitmsg); } else { t("TIMEOUT"); }
  if( DoExitAtAlarm	 ){ exit(0); }
  if( funcp != NULL	 ){ funcp(); }
  }

/*----------------- SetTimeout ----------------------------------------------------------*/

int SetTimeout( int secs, int usecs, char * msg, int doexit, void (*fp)(void) ){
  /* returns status : 0=OK negative=error */
  struct sigaction	 newact  ;
  int			 ck	 ;
  struct itimerval	 delay	 ;
  exitmsg	 = msg	 ;
  DoExitAtAlarm	 = doexit ;
  funcp		 = fp	 ;
  /* establish handler : */
  newact.sa_handler = &HandleTimeout	 ; /* sighandler_t , == void (*func)(int SIGNUM) */
  sigemptyset( &(newact.sa_mask) )	 ; /* no signals to be blocked when handler runs */
  newact.sa_flags   = 0			 ; /* no special cases				 */
  ck = sigaction( SIGALRM, &newact, NULL ); /* only support one alarm at a time.	 */
  if( ck != 0 ){ esx("failed sigaction"); }
  /* set timer : */
  delay.it_interval.tv_sec  = 0 ;		 /* don't repeat alarm		 */
  delay.it_interval.tv_usec = 0 ;		 /* don't repeat alarm		 */
  delay.it_value.tv_sec  = secs ;
  delay.it_value.tv_usec = usecs ;
  ck = setitimer( ITIMER_REAL, &delay, NULL );
  if( ck != 0 ){ esx("failed setitimer"); }
  return(0);
  }

/*----------------- UnsetTimeout --------------------------------------------------------*/

void UnsetTimeout( void ){
  struct sigaction	 defact	 ;
  int			 ck	 ;
  defact.sa_handler = SIG_IGN ; /* not SIG_DFL, because default is to terminate process */
  sigemptyset( &(defact.sa_mask) ); /* sets mask to 0 portably */
  defact.sa_flags   = 0 ;
  ck = sigaction( SIGALRM, &defact, NULL );
  if( ck != 0 ){ esx("failed sigaction"); }
  }

/*----------------- StringContains --------------------------------------------------------*/

char* StringContains( char* string, char* substring ){
  char	 c1	 ; /* first char of substring */
  char	 c	 ; /* tmp */
  char*	 pc	 ;

  /* caller must make sure that string and substring are not null */
  c1 = *substring ;
  pc = string ;
  while(1){
    for( ; (c=*pc) != c1 && c!= 0 ; pc++ ){ ; }
    if( c == 0 ){ return(NULL); }
    if( 0 == strncmp( pc, substring, (int) strlen(substring) ) ){ return( pc ); }
    ++pc ;
    }
  return( NULL ); /* dummy ; never reached. */
  }

/*----------------- StringContainsAtend ---------------------------------------------------*/

char* StringContainsAtend( char* string, char* substring ){
  int offset ;

  offset = strlen(string) - strlen(substring);
  if( 0==strcmp( string+offset, substring) ){ return( string+offset ); }
  return(NULL);
  }

/*----------------- SubstringReplace ------------------------------------------------------*/

/* can only be used if result is shorter than original ! */

void SubstringReplace( char* targetstring, int lendel, char* repstring, int lenrep ){
  /* 'rep' stands for 'replacement', 'del' for part at start to be deleted. */
  int lentargetstring	 ;
  int lenrepstring	 ;
  int i			 ;
  int shrink		 ;
  int lenmove		 ;

  lentargetstring = (int) strlen( targetstring );
  lenrepstring    = (int) strlen( repstring    );
  if( lentargetstring < lendel	 ){
    ex("string=%s= shorter than length to be deleted", targetstring ); }
  if( lenrepstring    < lenrep	 ){
    ex("replacement=%s= shorter than part to use of it", repstring  ); }
  shrink = lendel - lenrep ;
  if( shrink < 0 ){ ex("replacement=%s= is longer than original", repstring); }
  lenmove = lentargetstring - lendel + 1 ;		 /* also move terminating zero */
  for( i=0 ; i<lenrep ; i++ ){ targetstring[i] = repstring[i] ; }
  if( shrink > 0 ){
    for( i=0 ; i<lenmove ; i++ ){ targetstring[lenrep+i] = targetstring[lenrep+shrink+i] ; }
    }
  }

/*----------------- FindLeftwardChar ------------------------------------------------------*/

char* FindLeftwardChar( char* string, char* start, char c ){
  char* pc ;
  for( pc=start ; pc!=string && *pc!=c ; pc-- ){ ; }
  if(  *pc != c ){ return(NULL); }
  return(pc);
  }

/*----------------- QualifyFilename --------------------------------------------------------*/

char * QualifyFilename( void* vfilename ){ /* donates malloced buffer */
  /* argument should be char or uchar */
  int		 namelen, cwdlen			 ;
  T(char*)	 fullname, newname, username, cwd	 ;
  T(char*)	 pc, pc2, filenamecopy, ret, filename	 ;

  filename = (char*) vfilename ;
  if( filename == NULL ){ ex("filename argument is null"); }
  if( (namelen=(int)strlen(filename)) <= 0 ){ ex("filename argument is empty"); }
  /* first copy filename, to avoid changing argument itself : */
  filenamecopy = alloca( namelen + 5 );			 /* 4 bytes too much for defense */
  if( filenamecopy == NULL ){ ex("alloca failed"); }
  memcpy( filenamecopy, filename, namelen+1 );
  filename = filenamecopy ;
  /* remove all trailing slashes, but not name "/" : */
  while( namelen > 1 && filename[namelen-1] == '/' ){
    filename[namelen-1] = 0 ; --namelen ;
    }
  /* if filename starts with ~, replace it with name of user's home directory : */
  if( (filename[0] == '~') && (filename[1] == '/') ){
    username = cuserid(NULL);
    newname = (char*) ckalloc( strlen(filename) + strlen(username) + 8 );
    if( newname == NULL ){ ex("alloca failed"); }
    pc = newname ;
    if( 0 == strcmp( username, "root" ) ){
      strcpy( pc, "/root"); pc += (int)strlen("/root");
      }
    else {
      strcpy( pc, "/home/" ); pc += (int)strlen( "/home/" );
      strcpy( pc, username ); pc += (int)strlen( username );
      }
    strcpy( pc, filename + 1 ); /* starts with / */
    filename = newname ;
    }
  /* if filename does not start with '/', prepend current working directory to it : */
  if( filename[0] != '/' ){
    cwd = getcwd( NULL, 0 ); if( cwd == NULL ){ ex("getcwd failed"); }
    cwdlen = (int)strlen( cwd );
    fullname = (char*) alloca( cwdlen + 1 + namelen + 1 );
    if( fullname == NULL ){ ex("alloca failed"); }
    memcpy( fullname, cwd, cwdlen );
    fullname[cwdlen] = '/' ;
    stpcpy( fullname + cwdlen + 1 , filename );
    free( cwd );
    }
  else {
    fullname = (char*) alloca( namelen + 1); if( fullname == NULL ){ ex("OOM"); }
    strcpy( fullname, filename );
    }
  while( (pc=StringContains( fullname,"//" )) != NULL ){ SubstringReplace( pc+1, 1, "", 0 ); }
  while( (pc=StringContains( fullname,"/./")) != NULL ){ SubstringReplace( pc+1, 2, "", 0 ); }
  while( (pc=StringContainsAtend( fullname,"/."  )) != NULL ){ pc[0] = pc[1] = 0 ; }
  while( (pc=StringContains(      fullname,"/../")) != NULL ){
    if( pc == fullname ){ SubstringReplace( pc+1, 3, "", 0); continue ; }
    pc2=FindLeftwardChar( fullname, pc-1, '/') ;
    if( pc2 == NULL ){ ex("fullname doesnt start with /"); }
    SubstringReplace( pc2+1, (pc-pc2)+3, "", 0 );
    }
  while( (pc=StringContainsAtend( fullname,"/.." )) != NULL ){
    if( pc == fullname ){ SubstringReplace( pc+1, 2, "", 0); continue ; }
    pc2=FindLeftwardChar( fullname, pc-1, '/') ;
    if( pc2 == NULL ){ ex("fullname doesnt start with /"); }
    if( pc2 == fullname ){ SubstringReplace( pc2+1, (pc-pc2)+2, "", 0 ); }
    else                 { SubstringReplace( pc2  , (pc-pc2)+3, "", 0 ); }
    }
  /* malloc a buffer for fullname, to donate to caller : */
  namelen = (int)strlen( fullname );
  ret = ckalloc( namelen + 1 );
  memcpy( ret, fullname, namelen+1 );
  return( ret );
  }

/*----------------- FileExists ----------------------------------------------------------*/

int FileExists( void* vfqname, int* pfilesize ){
  struct stat	 stats	 ;
  char*		 fqname	 ;

  fqname = (char*) vfqname ;
  if( 0 != stat( fqname, &stats ) ){ /* allow symlinks ! */
    if( errno != ENOENT ){ esx("stat failed"); }
    return(0);
    }
  if( pfilesize != NULL ){ *pfilesize = stats.st_size ; }
  if( S_ISREG( stats.st_mode ) ){ return(1); }
  if( S_ISDIR( stats.st_mode ) ){ return(2); }
  if( S_ISLNK( stats.st_mode ) ){ return(3); }
  if( S_ISCHR( stats.st_mode ) ){ return(4); }
  if( S_ISBLK( stats.st_mode ) ){ return(5); }
  if( S_ISFIFO(stats.st_mode ) ){ return(6); }
  if( S_ISSOCK(stats.st_mode ) ){ return(7); }
  return(10);
  }

/*----------------- Pagedsize --------------------------------------------------------------*/

int Pagedsize( int needsize ){
  static int pagesize = 0 ;	 /* always a power of two */

  if( needsize <= 0 ){ ex("nonpositive requested size = %d",needsize);}
  if( pagesize <= 0 ){
    if( (pagesize=(int) sysconf(_SC_PAGESIZE)) <= 0 ){ esx("nok sysconf pagesize"); }
    }
  return( ((needsize + pagesize - 1)/pagesize) * pagesize );
  }

/*----------------- GetMmappedFilebuf ------------------------------------------------------*/

/* Gets an mmap of a whole file.							 */
/* Consequently size of map is just large enough to hold entire file.			 */
/* Therefore empty or nonexistent files can not be mmapped with this function.		 */

Filebuf GetMmappedFilebuf( void* vfqfilename, int domakename ){
  /* domakename is bool 'do malloc b.name <- fqfilename'. */
  int		 ck		 ;
  Filebuf	 b		 ;
  int		 needsize	 ;
  int		 filesize	 ; /* can not be size_t because size_t is unsigned */
  int		 fh		 ;
  char*		 fqfilename	 ;

  fqfilename = (char*) vfqfilename ;
  if( fqfilename == NULL	 ){ ex("filename is null"); }
  if( domakename ){
    b.name = ckalloc( strlen( fqfilename ) + 2 );
    strcpy( b.name, fqfilename );
    b.nameisprivate = 1 ;
    }
  else { b.name = fqfilename ; b.nameisprivate = 0 ; }
  ck = FileExists( b.name, &filesize );
  if( ck >  1		 ){ ex("file %s is not a regular file, so can't mmap",fqfilename ); }
  if( ck == 0		 ){ ex("file %s does not exist, so has no size",fqfilename); }
  if( filesize <= 0	 ){ ex("filesize of %s is nonpositive: %d", b.name, filesize ); }
  needsize = Pagedsize( filesize );		 /* can only allocate multiple of pagesize */
  fh = open_rw( fqfilename, 1 );		 /* includes test for returnvalue */
  b.buf = mmap( NULL, needsize, PROT_READ | PROT_WRITE, MAP_SHARED, fh, 0 );
  if( b.buf == MAP_FAILED	 ){ esx("nok mmap %s",fqfilename); }
  if( close( fh ) < 0		 ){ es("failed close"); }
  b.obuf	 = b.buf + needsize ;
  b.cont	 = b.buf ;
  b.ocont	 = b.buf + filesize ;
  b.ismmapped = 1 ; b.hasfile = 1 ; b.iswritable = 1 ; b.isshared = 1 ;
  b.pPrev = b.pNext = NULL ;
  return( b );
  }

/*----------------- GetMmappedBuf ----------------------------------------------------------*/

Filebuf GetMmappedBuf( void* vname, int domakename, int size ){
  Filebuf	 b		 ;
  int		 needsize	 ;
  char*		 name		 ;

  name = (char*) vname ;
  if( name == NULL		 ){ ex("name is null"); }
  if( size <= 0			 ){ ex("size nonpositive"); }
  if( domakename ){
    b.name = ckalloc( strlen(name) + 1 );
    strcpy( b.name, name );
    b.nameisprivate = 1 ;
    }
  else { b.name = name ; b.nameisprivate = 0 ; }
  needsize = Pagedsize( size );
  b.buf = mmap( NULL, needsize, PROT_WRITE|PROT_READ, MAP_PRIVATE | MAP_ANON, 0, 0 );
  if( b.buf == MAP_FAILED	 ){ esx("nok mmap %s", name ); }
  b.obuf	 = b.buf + needsize ;
  b.cont	 = b.buf ;
  b.ocont	 = b.buf ;
  b.ismmapped = 1 ; b.hasfile = 0 ; b.iswritable = 1 ; b.isshared = 0 ;
  b.pPrev = b.pNext = NULL ;
  return( b );
  }

/*----------------- DiscardFilebuf ---------------------------------------------------------*/

void DiscardFilebuf( Filebuf* pfb ){
  if( pfb == NULL	 ){ ex("pfb is null"); }
  if( pfb->name == NULL	 ){ e("name is null"); pfb->name="<noname>" ; pfb->nameisprivate=0;}
  if( pfb->buf  == NULL	 ){ ex("null buf of %s",pfb->name); }
  if( pfb->obuf <= pfb->buf ){ ex("nonpositive length of filebuf of %s",pfb->name); }
  if( pfb->ismmapped ){
    if( munmap( pfb->buf, pfb->obuf - pfb->buf) < 0 ){ es("nok munmap %s",pfb->name ); }
    }
  else { free(pfb->buf); }				 /* free() does not return status. */
  if( pfb->nameisprivate && pfb->name != NULL ){ free( pfb->name ); }
  /* DONT free() filebuffer struct itself !  It is a local variable normally. */
  }

/*----------------- open_w ----------------------------------------------------------------------*/

int open_w( char * filename, int exitonfail ){
  int fh ;
  fh = open( filename, O_WRONLY | O_CREAT, 0660 );
  if( fh < 0 ){
    es("failed open_w %s", filename );
    if( exitonfail ){ exit(-1); }
    }
  return( fh );
  }

/*----------------- ckwrite ---------------------------------------------------------------------*/

void ckwrite( int fh, void* buf, int size ){
  int totalwritten = 0	 ;
  int ck		 ;

  while( totalwritten < size ){
    ck = write( fh, buf+totalwritten, size-totalwritten );
    if( ck == 0 ){ sleep(1); continue ; }	 /* returnvalue 0 not specified in docs. */
    if( ck <  0 ){
      if( errno == EAGAIN ){ sleep(1) ; continue ; }
      if( errno == EINTR  ){ continue ; }
      esx("write error");
      }
    totalwritten += ck ;
    }
  }

/*----------------- WriteImoheader ---------------------------------------------------------*/

/* file must be open for write, and fileposition must be at start of file */

void WriteImoheader( int filehandle, uint wi, uint hi, int offx, int offy,
  float scalex, float scaley ){
  imoheader	h ;

  if( sizeof(imoheader) != 32 ){ ex("imoheader not 32 byte !"); }

  if( sizeof(float) > 4 ){ w("sizeof(float) is larger than 4 !  Writing malformed header !"); }
  if( sizeof(float) < 4 ){ w("sizeof(float) is smaller than 4 !  Writing nonstandard header !"); }

  h.sig[0] = 'i' ; h.sig[1] = 'm' ; h.sig[2] = 'o' ; h.sig[3] = '0' ;
  h.byteorder	 = 0x01020304 ;
  h.wi		 = (uint32_t) wi ;
  h.hi		 = (uint32_t) hi ;
  h.offx	 = (int32_t) offx ;	 /* offset always zero when converting a jpeg */
  h.offy	 = (int32_t) offy ;
  h.scalex	 = scalex ;		 /* scale always 1 when converting a jpeg */
  h.scaley	 = scaley ;

  ckwrite( filehandle, &h, 32 );
  }


/*----------------- WriteMallocedFilebuf --------------------------------------------------*/

void WriteMallocedFilebuf( Filebuf* pfb ){ /* really is WritePrivateFilebuf */
  int	 size	 ; /* can't be size_t because size_t is unsigned */
  int	 fh	 ;

  if( pfb == NULL		 ){ ex("pfb is null"); 	}
  if( pfb->name == NULL		 ){ ex("name is null"); }
  if( pfb->buf  == NULL		 ){ ex("buf is null ; name=%s",pfb->name); }
  size = pfb->ocont - pfb->cont ;
  if( size      <= 0		 ){ ex("contentlength is nonpositive ; name=%s",pfb->name); }
  if( pfb->ismmapped && pfb->isshared ){ ex("is mmapped shared ; name=%s",pfb->name ); }
  if( ! pfb->iswritable		 ){ ex("is not writable ; name=%s",pfb->name ); }
  if( ! pfb->hasfile		 ){ ex("filebuf %s is anonymous", pfb->name ); }

  if( (fh= open( pfb->name, O_WRONLY|O_CREAT, 0660)) < 0 ){
				    esx("failed open_w %s", pfb->name ); }

  /* Filebufs are never opened in append mode, so simply writing after opening suffices. */
  /* Because i use GNU #defines, write is restarted when it is interrupted,		 */
  /*   so no need to check for EINTR.							 */

  ckwrite( fh, pfb->buf, size);		 /* exits if read never succeeds. */
  if( close( fh) != 0		 ){ es("failed close %s", pfb->name ); }
  }

/*----------------- WriteMmappedFilebuf ----------------------------------------------------*/

 /* Doesn't really write, but changes filesize to an appropriate value and syncs.	 */
 /* Writing a filebuf is only usefull if content was changed,				 */
 /*   which implies that (some) content was written,					 */
 /*   which with current libc implies that file is available for read/write access.	 */
 /* Impossible is that some other process deleted file while my program has it mmapped,	 */
 /*   (at least that's what libc manual seems to be saying),				 */
 /*   so not really any need to add O_CREAT to open mode.				 */
 /*   It doesn't hurt to supply it, though.						 */
 /* Because open(w) + mmap(w) = permission denied,					 */
 /*   while open(rw) + mmap(rw) = ok							 */
 /*   i think it safest to open rw,							 */
 /*   even though in this routine i dont mmap but only ftruncate with filehandle gotten. */

void WriteMmappedFilebuf( Filebuf* pfb ){ /* really is WriteSharedFilebuf */
  int	 fh	 ;

  if( pfb == NULL		 ){ ex("pfb is null"); }
  if( pfb->name == NULL		 ){ ex("filebuf without name"); }
  if( ! pfb->ismmapped		 ){ ex("is not mmapped"); }
  if( ! pfb->isshared		 ){ e("is not shared"); WriteMallocedFilebuf(pfb); return ; }
  if( ! pfb->hasfile		 ){ ex("buffer has no file"); }
  if( ! pfb->iswritable		 ){ ex("filebuf %s was not alloced writable", pfb->name ); }
  if( pfb->ocont - pfb->cont <= 0 ){ ex("nonpositive size for %s", pfb->name ); }

  if( (fh= open( pfb->name, O_RDWR, 0660)) < 0	 ){ esx("nok open %s", pfb->name ); }
  if( ftruncate( fh, pfb->ocont - pfb->cont) != 0 ){ esx("nok ftruncate %s", pfb->name ); }
  if( close( fh) != 0				 ){ esx("nok close %s", pfb->name ); }
  if( msync( pfb->cont, pfb->ocont - pfb->cont, 0) < 0 ){ es("nok msync %s", pfb->name ); }
  }

