/* socket.c : socket code for canvas. handles client connections and watches x connection. */

#include "defines.h"
#include "myincludesT.h"	 /* orgb		 */
#include "myincludes.h"		 /* ???			 */
#include "command.h"		 /* HandleMessage()	 */
#include "x.h"			 /* PollEvents()	 */

#include <stdio.h>		 /* sscanf()					 */
#include <string.h>		 /* strncmp(), memmove()			 */
#include <sys/types.h>		 /* select()					 */
#include <sys/socket.h>		 /* socket(), bind(), listen(), accept()	 */
#include <sys/un.h>		 /* struct sockaddr_un				 */
#include <sys/time.h>		 /* struct timeval				 */
#include <unistd.h>		 /* sleep(), read(), write(), close(), unlink()	 */
/*#include <fcntl.h>*/		 /* open()					 */
#include <signal.h>

/*--------------------- typedef ------------------------------------------------------------*/

typedef struct ConnectionS {
  struct ConnectionS*	 pNext	 ;
  int			 fh	 ; /* filehandle */
  uchar			 isx	 ; /* bool 'is connection to x server */
  uchar			 isserv	 ; /* bool 'is server socket' */
  } Connection ;

/*----------------- global variables -----------------------------------------------------------*/

extern int	 xfh		 ; /* defined in x.c				 */
extern uchar*	 sockfilename	 ; /* defined in main, for ParseArguments()	 */
extern int	 verbose	 ; /* myincludes.c				 */

/*----------------- filescope variables --------------------------------------------------------*/

static int		 sockfh		 ; /* fh of server socket */
static Connection*	 conlist	 ;
static Connection*	 conlast	 ;

/*================= UTIL =======================================================================*/

static void AddConnection( int fh, int isx, int isserv ){
  v("adding connection fh=%d isx=%d isserv=%d", fh, isx, isserv );
  if( conlist == NULL ){ conlast = conlist        = (Connection*) ckalloc( sizeof(Connection)); }
  else                 { conlast = conlast->pNext = (Connection*) ckalloc( sizeof(Connection)); }
  conlast->pNext = NULL ;
  conlast->fh = fh ; conlast->isx = isx ; conlast->isserv = isserv ;
  }

/*================= INIT CODE ==================================================================*/

static void SetSigpipeHandler( void ){
  if( SIG_ERR == signal(SIGPIPE, SIG_IGN) ){ esx("nok signal"); }
  }

static void MakeServSock( void ){
  int			 size		 ; /* tmp */
  struct sockaddr_un	 sockstruct	 ; /* main server socket */

  if( 0 > (sockfh = socket( PF_UNIX, SOCK_STREAM, 0 )) ){ esx("cant create socket."); }
    /* creates local connectable socket, using only available (nameless) protocol. */
  sockstruct.sun_family = AF_UNIX ;
  strcpy( sockstruct.sun_path, (char*)sockfilename );
  size = SUN_LEN( &sockstruct ) ;
    vvv("SUN_LEN is %d", size ); /* SUN_LEN is 18 for /tmp/canvas.sock */
  if( 0 != bind( sockfh, (struct sockaddr *) &sockstruct, size ) ){ esx("failed bind."); }
    /* associates filename with socket */
  if( 0 != listen( sockfh, 3 )                                   ){ esx("failed listen"); }
    /* starts kernel listening on socket, max 3 pending connections before ECONNREFUSED. */
  }

static void DeleteServsock( void ){
  if( 0 != close( sockfh )		 ){ es("nok close sockfh"); }
  if( 0 != unlink( (char*)sockfilename ) ){ es("nok unlink sockfilename"); }
  }

void InitSocket( void ){
  if( conlist != NULL ){ ex("conlist is not null"); }
  if( xfh     == 0    ){ ex("must be called After X has been initialized"); }
  SetSigpipeHandler();
  MakeServSock();
  AddConnection( xfh   , 1, 0 );
  AddConnection( sockfh, 0, 1 );
  }

void ExitSocket( void ){
  DeleteServsock();
  }


/*================= LOOP CODE ==================================================================*/

/*----------------- AcceptNewConnection --------------------------------------------------------*/

static void AcceptNewConnection( void ){
  struct timeval	 tv		 ;
  int			 ck		 ;
  fd_set		 set		 ;
  struct sockaddr	 consockstruct	 ;
  socklen_t		 consocklen	 ;
  int			 confd		 ;

  /* check that there is a pending connection : */
  FD_ZERO( &set ); FD_SET( sockfh, &set );
    /* FD_SET is changed by select(), so must be reinitialized, every time it is used ! */
  tv.tv_sec = 0 ; tv.tv_usec = 0 ;	 /* select should not blocking wait at all */
  ck = select( FD_SETSIZE, &set, NULL, NULL, &tv );
  if( ck == 0 ){ w("non c'e pending connection"); return ; }
  if( ck <  0 ){ es("nok select sockfh"); return ; }

  /* accept pending connection : */
  consocklen = 0 ;
  consockstruct.sa_family = 0 ;
  consockstruct.sa_data[0] = 0 ;
  if( 0 > (confd = accept( sockfh, &consockstruct, &consocklen )) ){ es("nok accept."); return ; }

  /* register new connection in connectionslist : */
  AddConnection( confd, 0, 0 );

  /* if there is a new connection, it would have a message,				 */
  /*   so i could try to handle it immediately,						 */
  /* but HandleConnection does an extra check to see whether socket still has a message, */
  /* so i let that handle new connection too.						 */
  }

/*----------------- DiscardConnection ----------------------------------------------------------*/

static void DiscardConnection( Connection* pcon ){
  /* canvas never closes a connection.					 */
  /* discard happens at broken pipe due to client closing connection.	 */
  /* discard connection removes connection from connectionlist.		 */
  T(Connection*) pc, pcp ;

  if( conlist == NULL ){ ex("empty list"); }
  if( conlast == NULL ){ ex("broken list"); }
  for( pc=conlist, pcp=NULL ; pc != pcon && pc != NULL ; pcp=pc, pc=pc->pNext ){ ; }
  if( pc == NULL ){ e("connection not found in list"); return ; }
  v("filehandle=%d isx=%d isserv=%d", pcon->fh, pcon->isx, pcon->isserv );
  if( pcp == NULL ){ conlist = pcon->pNext ; } else { pcp->pNext = pcon->pNext ; }
  if( conlast == pcon ){ conlast = pcp ; }
  if( 0 != close( pcon->fh ) ){ es("nok close"); }
  ckfree( pcon );
  }

/*----------------- readall --------------------------------------------------------------------*/

static int readall( int fh, uchar** pmsg ){
  /* Returns length of message read, or -1 for error.						 */
  /* Donates malloced buffer to caller.								 */
  /* Buffer is always at least one byte larger than message, and byte after message is zero.	 */
  /*   (for ease of printing it in verbose mode).						 */

  int		 got		 ;
  T(char*)	 buf, mbuf	 ;
  int		 bufsize	 ;

  if( pmsg == NULL ){ ex("null pmsg"); }

  /* set initial buffer size to 1 MB, to hopefully have enough size for any message */
  bufsize = 1000000 ; buf = alloca(bufsize);
  while( 1 ){						 /* retry in case of EINTR */
    errno = 0 ;
    got = read( fh, buf, bufsize-1 ); vvv("read %d bytes", got );
    if( got >  0 ){ break ; }
    if( got == 0 ){ return(0); } /* this happens at read from closed pipe. is kernel or libc bug. */
    if( got <  0 && errno == EINTR ){ continue ; }
    if( got <  0 ){ es("nok read"); return(-1); }
    }
  if( buf[got-1] != 0 ){ w("message is not null-terminated"); }

  mbuf = (char*) ckalloc( got+1 );
  memcpy( mbuf, buf, got ); mbuf[got] = 0 ;
  *pmsg = (uchar*)mbuf ; return(got);
  }

/*----------------- writeall -------------------------------------------------------------------*/

static int writeall( int fh, uchar* msg, int msglen ){ /* returns 1 if success, -1 if error */
  int	 wrote, totalwrote	 ;

  for( totalwrote=0 ; totalwrote < msglen ; totalwrote += wrote ){
    wrote = write( fh, msg + totalwrote, msglen - totalwrote );
    vvv("wrote %d bytes", wrote );
    if( wrote < 0 && errno != EINTR ){ es("nok write"); return(-1); }
    if( wrote == 0 ){ w("wrote nothing"); }
    }
  return(1);
  }

/*----------------- HandleConnection -----------------------------------------------------------*/

int HandleConnection( Connection* pcon ){
  /* It is known that connection needs to be handled.					 */
  /* This could be because it has an error to report, or because it has a message.	 */
  /* returns positive for success, negative for error.					 */
  T(uchar*)	 buf			 ; /* pointo msg read from socket		 */
  uchar		 replybuf[16]		 ; /* plenty for "+%d\0"			 */
  uchar*	 textresult		 ;
  uchar*	 reply			 ; /* equals either replybuf or textresult	 */
  int		 got, replen, ck	 ;
  int		 error, needquit, result ;
  int		 status			 ; /* superfluous ?				 */

  error = needquit = result = 0 ;
  got = readall( pcon->fh, &buf );			 /* donates ckalloced storage */
  if( got < 0 ){					 /* failed read(), not due to EINTR */
    if( errno == EPIPE ){ v("got EPIPE for filehandle %d", pcon->fh );
      DiscardConnection( pcon ); return(1);
      }
    es("nok read from fh %d", pcon->fh ); return(-1);
    }
  if( got == 0 ){
    /* this apparently means that other end of connection was closed	!!!	 */
    /*   and, equally apparently, server never gets EPIPE for this.	!!!	 */
    /* Since this always happens, i don't emit a warning for it.		 */
    vv("no message received from fh %d . (kernel/libc bug).", pcon->fh );
    DiscardConnection( pcon ); return(0);
    }
 
  /* buf now contains pointer to ckallocked message, got contains size of message */
  vvv("got message length=%d string=%s=", got, buf );

  textresult = NULL ;
  result = HandleMessage( buf, got, &textresult );	 /* in command.c */
  /* if there is a textresult, it points to scratchpad storage that is permanently allocated */
  /*   this storage has '+', followed by event strings */
  if(   result == -10 ){ needquit = 1 ; }
  elif( result <    0 ){ error    = 1 ; }

  /* free storage of message : */
  ckfree( buf );

  /* send reply : */
  vvv("sending reply. error=%d result=%d needquit=%d", error, result, needquit );
  if( textresult != NULL ){
    reply = textresult ;
    reply[0] = (error)? '-' : '+' ;
    replen = 2 + ustrlen(textresult+1);
    }
  else {
    reply = replybuf ;
    reply[0] = (error)? '-' : '+' ;
    if( result > 0 ){ replen = 2 + sprintf( (char*)replybuf+1,"%d",result ); }
    else            { replen = 2 ; reply[1] = 0 ; }
    }
  vvv("reply=%s replen=%d fh=%d", replybuf, replen, pcon->fh );
  if(   needquit ){ status = -10 ; } /* only this state needs to be signalled to caller */
  elif( error    ){ status =  -1 ; }
  else            { status =   1 ; }
  ck = writeall( pcon->fh, reply, replen );

  if( ck < 0 ){
    if( errno == EPIPE ){	 /* this error can happen in normal operation */
      v("EPIPE at write to filehandle %d", pcon->fh );
      DiscardConnection( pcon );
      }
    else { es("nok writeall to filehandle %d", pcon->fh ); if( status > -1 ){ status = -1 ; } }
    }
  if( ck == 0){ e("writeall returned 0"); if( status > -1 ){ status = -1 ; } }
  return(status);
  }

/*----------------- HandleChannels -------------------------------------------------------------*/

void HandleChannels( void ){
  fd_set	 set		 ;
  T(Connection*) pcon, pnext	 ;
  int		 ck		 ;
  int		 iterix		 ;

  for( iterix=0 ; 1 ; iterix++ ){
    vv("iterix=%d",iterix);
    FD_ZERO( &set );
    for( pcon=conlist ; pcon != NULL ; pcon=pcon->pNext ){ FD_SET( pcon->fh, &set ); }

    ck = select( FD_SETSIZE, &set, NULL, NULL, NULL ); /* blocks until activity */
    if( ck == 0 ){ e("select returned 0 (maybe process got a signal ?)"); continue ; }
      /* would normally indicate timeout, but is also caused by ANY signal */
    if( ck <  0 ){ es("nok select"); return ; }
    /* set now contains only those filedescriptors that need handling */

    for( pcon=conlist ; pcon != NULL ; pcon=pnext ){
      pnext = pcon->pNext ; /* because *pcon itself may be removed from connectionlist */
      if( ! FD_ISSET( pcon->fh, &set) ){ continue ; }
      if(   pcon->isx    ){ PollEvents(); }
      elif( pcon->isserv ){ AcceptNewConnection(); }
      else                { ck = HandleConnection( pcon ); if( ck == -10 ){ return ; } }
      }
    }
  }
