
/*----------------- include ----------------------------------------------------------------*/

#include "defines.h"		 /* gnu-extensions, v(), etc.	 */
#include <fcntl.h>		 /* open()			 */
#include <sys/mman.h>		 /* munmap()			 */
#include <unistd.h>		 /* ftruncate()			 */
#include "myincludesT.h"	 /* orgb, imo			 */
#include "myincludes.h"
#include "imobuf.h"		 /* OverlayOrgb()		 */
#include "textT.h"		 /* fontmeta, glyphmeta, etc.	 */

/*----------------- extern -----------------------------------------------------------------*/

extern int	 verbose				 ; /* myincludes.c	 */
extern int	 imoaffx1, imoaffx2, imoaffy1, imoaffy2	 ; /* imobuf.c		 */
extern int	 winwi, winhi				 ; /* main.c		 */
extern orgb	 drawcolor				 ; /* main.c		 */
extern orgb*	 imor					 ; /* imobuf.c		 */

/*----------------- public -------------------------------------------------------------*/

/*----------------- filescope -----------------------------------------------------------*/

static Filebuf	 fontfb		 ;
static int	 havefont	 ;
static fontmeta* pfm		 ;

void DumpGlyphmeta( glyphmeta* pgm );

/*----------------- LoadFont ------------------------------------------------------------*/

/*#define STDGLYPHDIR (uchar*)"/usr/share/glyph"*/
#define STDGLYPHDIR (uchar*)"./glyph/"

/* If file doesn't exist, GetMmappedFilebuf will ex().					 */
/* If it does, but is not a valid gly0 file, FindGlyph() will probably segfault.	 */
/* Therefore, user should run CheckGlyfile after load, and react appropriately.		 */

static Filebuf LoadFont( uchar* fileroute ){ /* fileroute is alloca'd by caller */
  Filebuf	 fb	 ;
  fontmeta*	 pfm	 ;

  if( fileroute == NULL					 ){ ex("null arg"); }
  if( 0 != strncmp( (char*) fileroute, "/usr/share/glyph/", 17 ) ){
    w("directory wrong=%s= should be =/usr/share/glyph=", fileroute ); }
  fb = GetMmappedFilebuf( (char*) fileroute, 1 ) ;

  pfm = fb.cont + sizeof(glyheader) ;
  pfm->xid = 0 ;
  return( fb );
  }

/*----------------- DiscardFont ------------------------------------------------------------*/

static void DiscardFont( Filebuf* pfb ){
  int fd, ok, size ;

  if( pfb == NULL ){ ex("null arg"); }
  ok = 1 ;
  if( pfb->hasfile && pfb->ismmapped ){				 /* normally this is true */
    if( pfb->name == NULL ){ e("filebuf has no name"); ok = 0 ; }
    size = pfb->obuf - pfb->buf ;
    if( size <= 0 ){ e("filebuf bufsize is nonpositive"); ok = 0 ; }
    size = pfb->ocont - pfb->cont ;
    if( size <= 0 ){ e("filebuf bufsize is nonpositive"); ok = 0 ; }
    if( ok ){ 
      if( 0 > (fd = open( pfb->name, O_RDWR )) ){ esx("nok open %s", pfb->name ); ok = 0 ; }
      if( 0 != munmap( pfb->buf, pfb->obuf - pfb->buf ) ){ esx("nok munmap"); }
      if( 0 > ftruncate( fd, size ) ){ esx("nok ftruncate %s to size %d", pfb->name, size ); }
      }
    }
  else { w("not an mmapped file buffer") ; }
  DiscardFilebuf( pfb );
  }

/*----------------- Glyphdatasize ----------------------------------------------------------*/

/* returns size of storage currently used by data of a glyph in a gly0 file */

static int Glyphdatasize( glyphmeta* pgm ){
  int	 cpixels	 ;
  int	 datasize	 ;

  cpixels = pgm->inkboxhi * pgm->inkboxwi ;

  if( ! pgm->gry ){
    if( ! pgm->toggle ){ datasize = (cpixels + 7) >> 3 ; }
    else               { datasize = pgm->toggle * sizeof(bwtoggle) ; }
    }
  else {
    if( ! pgm->toggle ){ datasize = (cpixels + 1) >> 1 ; }
    else               { datasize = pgm->toggle * sizeof(grytoggle) ; }
    }
  return( datasize );
  }

/*----------------- CheckGlyfile -----------------------------------------------------------*/

/* returns +1 for ok, -1 for nok. */

static int CheckGlyfile( Filebuf* pfb ){
  glyheader*	 pgh				 ;
  fontmeta*	 pfm				 ;
  T(glyphmeta*)	 pgms, pogms, pgm		 ;
  int		 minsize			 ;
  u32		 cglyphs, i			 ;
  u32		 datablockoff, odatablockoff	 ;
  u32		 prevdataoff			 ;
  int		 charno, prevcharno		 ;
  int		 datasize, prevdatasize		 ;
  int		 ok				 ;
  int		 cpixels, ctoggles, count	 ;
  T(void*)	 pdata, podata			 ;

  if( pfb == NULL				 ){ e("null arg"); return(-1); }
  minsize = sizeof(glyheader) + sizeof(fontmeta) ;
  if(  pfb->buf  == NULL || pfb->obuf  == NULL
    || pfb->cont == NULL || pfb->ocont == NULL
    || pfb->obuf - pfb->buf <= minsize || pfb->ocont - pfb->cont <= minsize
    || pfb->cont - pfb->buf < 0 || pfb->obuf - pfb->ocont < 0
   						 ){ e("invalid Filebuf");	 return(-1); }

  pgh = (glyheader*) pfb->cont ;
  if( 0 != memcmp( pgh->sig, "gly0", 4 )	 ){ e("not a gly0 file");	 return(-1); }
  if( pgh->order != 0x01020304			 ){ e("nok byteorder");		 return(-1); }

  pfm = pfb->cont + sizeof(glyheader) ;
  cglyphs = pfm->cglyphs ;
  pgms  = (glyphmeta*) ((void*) pfm + sizeof(fontmeta)) ;
  pogms = pgms + cglyphs ;
  if( (void*) pogms > pfb->ocont		 ){ e("r glyphmetas no fit");	 return(-1); }

  datablockoff  = (u32) ((void*) pogms - pfb->cont) ;
  odatablockoff = (u32) (pfb->ocont    - pfb->cont) ;

  prevdataoff  = datablockoff ;
  prevdatasize = 0 ;
  prevcharno = -1 ;
  ok = 1 ;
  for( pgm=pgms, i=0 ; i < cglyphs ; i++, pgm++ ){
    charno = (int) pgm->charno ;
    if( charno < prevcharno  ){
     	 e("glyphix=%d prevcharno=%d charno=%d", i, prevcharno, charno );	 ok=0; }
    if( charno == prevcharno			 ){ e("duplicate charno");	 ok=0; }
    prevcharno = charno ;

    datasize = Glyphdatasize( pgm );
    if( datasize <= 0 ){ continue ; }
    if(  pgm->data <  datablockoff
      || pgm->data >= odatablockoff
      ){		 e("glyphix=%d invalid dataoffset %d",i,pgm->data); return(-1); }
    if( pgm->data + datasize > odatablockoff	 ){ e("data beyond buffer"); return(-1); }
    if( prevdataoff + prevdatasize != pgm->data	 ){ e("non-consecutive data");	 ok=0; }
    prevdataoff = pgm->data ; prevdatasize = datasize ;

    if( pgm->toggle ){ /* verify cpixels/ctoggles */
      pdata  = pfb->cont + pgm->data ;
      podata = pdata + datasize ;
      cpixels  = pgm->inkboxwi * pgm->inkboxhi ;
      ctoggles = pgm->toggle ;
      if( ! pgm->gry ){
        bwtoggle* pbt ;
        for( pbt=(bwtoggle*)pdata, count=0 ;
          count < cpixels && pbt < (bwtoggle*)podata ; pbt++ ){
          count += pbt->rep ;
          }
        if( count != cpixels			 ){ e("cpixels wrong");		 ok=0; }
        if( pbt - (bwtoggle*)pdata != ctoggles	 ){ e("ctoggles wrong");	 ok=0; }
        }
      else { /* grytoggle */
        grytoggle* pgt ;
        for( pgt=(grytoggle*)pdata, count=0 ;
          count < cpixels && pgt < (grytoggle*)podata ; pgt++){
          count += pgt->rep ;
          }
        if( count != cpixels			 ){ e("cpixels wrong");		 ok=0; }
        if( pgt - (grytoggle*)pdata != ctoggles	 ){ e("ctoggles wrong");	 ok=0; }
        }
      }
    }
  if( prevdataoff + prevdatasize != odatablockoff ){
    w("end of datablock not at end of data"); }
  if( ok ){ vvv("ok"); return(1); }
  e("nok"); return(-1);
  }

/*----------------- CkLoadFont ----------------------------------------------------------*/

int CkLoadFont( uchar* filename ){ /* arg points to txt field of TEMPORARY Dc struct */
  /* returns 1 for ok, -1 for nok */
  int		 dirnamelen		 ;
  T(uchar*)	 fqfilename, newname, p	 ;
  int		 ck			 ;

  dirnamelen = ustrlen(STDGLYPHDIR);
  if(   filename[0] == '/' ){ fqfilename = filename ; }
  elif( filename[0] == '.' ){
    fqfilename = uQualifyFilename( filename );			 /* mallocs ! */
    newname = (uchar*) alloca( ustrlen(fqfilename) + 1 );
    ustrcpy( newname, fqfilename );
    ckfree( fqfilename );
    fqfilename = newname ;
    }
  else {
    fqfilename = (uchar*) alloca( dirnamelen + 1 + ustrlen(filename) + 1 );
    p = ustpcpy( fqfilename, STDGLYPHDIR ); *(p++) = '/' ;
    p = ustpcpy( p, filename ); *p = 0 ;
    }
  ck = FileExists( fqfilename, NULL );
  if( ck <= 0 ){ e("fontfilename %s does not exist", fqfilename ); return(-1); }
  if( ck >  1 ){ e("fontfilename %s is not an ordinary file", fqfilename ); return(-1); }

  if( havefont ){ DiscardFont(&fontfb); havefont = 0 ; }
  fontfb = LoadFont( fqfilename ); 
  if( 0 > CheckGlyfile( &fontfb ) ){
    e("nok checkglyfile");
    DiscardFont(&fontfb); havefont = 0 ;
    return(-1);
    }
  pfm = (fontmeta*) (fontfb.cont + sizeof(glyheader)) ;
  havefont = 1 ;
  vvv("ok load font %s", fqfilename );
  return(1);
  }

/*----------------- FindGlyph --------------------------------------------------------------*/

static glyphmeta* FindGlyph( Filebuf* pfb, int charno ){
  fontmeta*	 pfm		 ;
  glyphmeta*	 pgms		 ;
  int		 cn		 ;
  int		 cglyphs	 ;
  int		 hi, lo, half	 ;

  if( pfb == NULL ){ ex("null pfb"); }
  if( charno < 0  ){ ex("negative charno"); }

  pfm = (fontmeta*) (pfb->cont + sizeof(glyheader)) ;
  cglyphs = pfm->cglyphs ;
  pgms = (glyphmeta*) (pfb->cont + pfm->glyphs) ;
  lo = 0 ;
  cn = pgms[lo].charno ;
  if( cn >  charno ){ return( NULL      ); }
  if( cn == charno ){ return( pgms + lo ); }
  hi = cglyphs - 1 ;
  cn = pgms[hi].charno ;
  if( cn == charno ){ return( pgms + hi ); }
  if( cn <  charno ){ return( NULL      ); }
  while( hi - lo > 0 ){
    half = (hi + lo) >> 1 ;
    cn = pgms[half].charno ;
    if( cn == charno ){ return( pgms + half ); }
    if( cn <  charno ){ lo = half ; }
    else              { hi = half ; }
    }
  return( NULL );
  }

/*----------------- MemrenderGlyph ---------------------------------------------------------*/

static int MemrenderGlyph( glyphmeta* pgm, void* pcont, u8* inkbox ){
  /* returns size of consumed inputdata */
  int		 cpixels	 ;
  int		 gry		 ;
  int		 toggle		 ;
  void*		 data		 ;
  T(u8*)	 puc, pbyte	 ;
  u8		 byte, val	 ;
  bwtoggle*	 pbt		 ;
  grytoggle*	 pgt		 ;
  int		 i, j, rep	 ;

  if( pgm == NULL || pcont == NULL || inkbox == NULL ){ ex("null arg"); }

  cpixels = pgm->inkboxhi * pgm->inkboxwi ;	 if( cpixels <= 0 ){ return(0) ; }
  gry = pgm->gry ; toggle = pgm->toggle ;
  data = pcont + pgm->data ;

  puc   = (u8*) inkbox ;
  if( gry == 0 ){
    if( toggle == 0 ){ /* bwmap */
      pbyte = (u8*) data ;
      for( i=0 ; ; pbyte++ ){ byte = *pbyte ;
        if( i >= cpixels ){ break ; } *(puc++) = (byte & 0x80)? 0xFF : 0 ; i++ ;
        if( i >= cpixels ){ break ; } *(puc++) = (byte & 0x40)? 0xFF : 0 ; i++ ;
        if( i >= cpixels ){ break ; } *(puc++) = (byte & 0x20)? 0xFF : 0 ; i++ ;
        if( i >= cpixels ){ break ; } *(puc++) = (byte & 0x10)? 0xFF : 0 ; i++ ;
        if( i >= cpixels ){ break ; } *(puc++) = (byte & 0x08)? 0xFF : 0 ; i++ ;
        if( i >= cpixels ){ break ; } *(puc++) = (byte & 0x04)? 0xFF : 0 ; i++ ;
        if( i >= cpixels ){ break ; } *(puc++) = (byte & 0x02)? 0xFF : 0 ; i++ ;
        if( i >= cpixels ){ break ; } *(puc++) = (byte & 0x01)? 0xFF : 0 ; i++ ;
        }
      return( (cpixels+7) >> 3 );
      }
    else { /* bwtoggle */
      pbt = (bwtoggle*) data ;
      for( i=0 ; i < cpixels ; pbt++ ){
        rep = pbt->rep ;
        if( i + rep > cpixels ){
          e("bwt overrun: charno=? i=%d rep=%d cpixels=%d", i, rep, cpixels );
          rep = cpixels - i ;
          }
        val = (pbt->val==1)? 0xFF : 0 ;
        for( j=0 ; j < rep ; j++ ){ *(puc++) = val ; }
        i += rep ;
        }
      return( ((void*)pbt) - data );
      }
    }
  else { /* gry */
    if( toggle == 0 ){ /* grymap */
      pbyte = (u8*) data ;
      for( i=0 ; ; pbyte++ ){ byte = *pbyte ;
        val = byte >> 4  ;
        val |= val << 4 ;
        *(puc++) = val ;
        i++ ;
        if( i >= cpixels ){ break ; }
        val = byte & 0xF ;
        val |= val << 4 ;
        *(puc++) = val ;
        i++ ;
        if( i >= cpixels ){ break ; }
        }
      return( (cpixels + 1) >> 1 );
      }
    else { /* grytoggle */
      pgt = (grytoggle*) data ;
      for( i=0 ; i < cpixels ; pgt++ ){
        rep = pgt->rep ;
        if( i + rep > cpixels ){ e("gt overrun"); rep = cpixels - i ; }
        val = pgt->val ;
        val |= val << 4 ;
        for( j=0 ; j < rep ; j++ ){ *(puc++) = val ; }
        i += rep ;
        }
      return( ((void*)pgt) - data );
      }
    }
  return(0); /* dummy */
  }

/*----------------- DrawText -----------------------------------------------------------*/

int DrawText( uchar direction, int x, int y, uchar* text ){
  /* x and y coords are relative to topleft of window (where x/y is 0/0) */
  int		 gox, goy		 ; /* current glyph's origin */
  int		 x1, y1			 ; /* topleftmost inked pixel of current glyph */
  orgb*		 plino			 ; /* first orgb of inkboxline */
  orgb*		 po			 ; /* current position in imobuf */
  int		 inkboxhi, inkboxwi	 ; /* tmp */
  ushort	 t			 ; /* tmp */
  int		 i, v, h		 ; /* loopvars */
  glyphmeta*	 pgm			 ;
  T(uchar*)	 mem, pmem		 ;
  orgb		 o			 ;
  int		 interline		 ;
  int		 textlen		 ;
  int		 memsize, newmemsize	 ;
  int		 rightest, lowest	 ; /* tmp */
  int		 stride			 ;

  stride = (pfm->maxstride + pfm->minstride) >> 1 ;

  vvv("direction=%c x=%d y=%d text=%s=", direction, x, y, text );
  if( havefont == 0 ){ e("no font loaded"); return(-1); }
  if( x > winwi || y > winhi ){ e("coord out of bounds"); return(-1); }
  textlen = ustrlen(text) ;
  if( textlen <= 0 ){ e("empty text"); return(-1); }

  if( direction == 'u' || direction == 'd' ){
    /* compute interline space of font */
    interline = pfm->fontheight - (pfm->inkhighest - pfm->inklowest + 1) ;
    vvv("fontheight=%d inkhighest=%d inklowest=%d interline=%d",
      pfm->fontheight, pfm->inkhighest, pfm->inklowest, interline );
    if( interline < 2 ){ interline = 2 ; }
    }

  gox = x ; goy = y ; memsize = 0 ;
  for( i=0 ; i < textlen ; i++ ){ pgm = FindGlyph( &fontfb, text[i] );
    vvvv("i=%d rendering char %c, pgm=%p", i, text[i], pgm );
    if( pgm != NULL ){
      if( verbose >= 3 ){ DumpGlyphmeta(pgm); }
      if( pgm->data != 0 ){ /* else is inkless, so no need to draw anything */
        if( direction == 'l' ){ gox -= pgm->logiwi ; } /* looks better */
        if( direction == 'd' ){ goy += pgm->inkboxoffy ; }
        if( direction == 'u' ){ goy -= (pgm->inkboxhi - pgm->inkboxoffy) ; }
        inkboxwi = pgm->inkboxwi ; inkboxhi = pgm->inkboxhi ;
        x1 = gox + pgm->inkboxoffx ;
        y1 = goy - pgm->inkboxoffy ;
        rightest = x1 + inkboxwi - 1 ;
        lowest   = y1 + inkboxhi - 1 ;
        vvvv("gox=%d goy=%d inkboxwi=%d inkboxhi=%d x1=%d y1=%d rightest=%d lowest=%d",
          gox, goy, inkboxwi, inkboxhi, x1, y1, rightest, lowest );
        if(  x1 >= 0 && rightest < winwi && y1 >= 0 && lowest < winhi ){
          newmemsize = inkboxwi * inkboxhi * sizeof(uchar) ;
          if( newmemsize > memsize ){
            mem = (uchar*) alloca( newmemsize ); memsize = newmemsize ;
            }
          MemrenderGlyph( pgm, fontfb.cont, mem );
          x1 = gox + pgm->inkboxoffx ;
          y1 = goy - pgm->inkboxoffy ;
		v("rendering glyph %c at x=%d y=%d", text[i], gox, goy );
          if( imoaffx1 > x1       ){ imoaffx1 = x1	 ; }
          if( imoaffy1 > y1       ){ imoaffy1 = y1	 ; }
          if( imoaffx2 < rightest ){ imoaffx2 = rightest ; }
          if( imoaffy2 < lowest   ){ imoaffy2 = lowest	 ; }

          plino = imor + y1 * winwi + x1 ; pmem = mem ;
          for( v=0 ; v < inkboxhi ; v++, plino+=winwi ){
            for( h=0, po=plino ; h < inkboxwi ; h++, po++, pmem++ ){
              o = drawcolor ; t = o.o * *pmem ; t += t >> 8 ; o.o = t >> 8 ;
              OverlayOrgb( po, &o );
              }
            }
          }
        else { w("omitted char %c because it does not completely fit on canvas", text[i]); }
        vvvv("drew char to imobuf");
        }
      else { vvvv("this char is inkless"); }

      if( pgm->data != 0 ){
        if(   direction == 'r' ){ gox += pgm->logiwi ; }
        elif( direction == 'l' ){ /* gox -= pgm->logiwi */ ; /* already done at start */ }
        elif( direction == 'd' ){
          if( pgm->inkboxhi <= 0 ){ goy += interline + pgm->logiwi ; }
          else                    { goy += interline + pgm->inkboxhi - pgm->inkboxoffy ; }
	  }
        else {
          if( pgm->inkboxhi <= 0 ){ goy -= interline + pgm->logiwi ; }
          else                    { goy -= interline + pgm->inkboxoffy ; }
          }
        }
      else { /* space char, or not in font */
        if(   direction == 'r' ){ gox += stride ; }
        elif( direction == 'l' ){ gox -= stride ; }
        elif( direction == 'd' ){ goy += stride ; }
        else                    { goy -= stride ; }
        }
      }
    else { /* no glyph for this charno */
      if(   direction == 'r' ){ gox += stride ; }
      elif( direction == 'l' ){ gox -= stride ; }
      elif( direction == 'd' ){ goy += stride ; }
      else                    { goy -= stride ; }
      }
    } /* next character */
  return(1);
  }

/*----------------- DumpGlyphmeta -------------------------------------------------------*/

void DumpGlyphmeta( glyphmeta* pgm ){
  t("  inkboxwi   = %d", pgm->inkboxwi );
  t("  inkboxhi   = %d", pgm->inkboxhi );
  t("  inkboxoffx = %d", pgm->inkboxoffx );
  t("  inkboxoffy = %d", pgm->inkboxoffy );
  t("  logiwi     = %d", pgm->logiwi );
  t("  gry        = %d", pgm->gry );
  t("  toggle     = %d", pgm->toggle );
  t("  charno     = 0x%X", pgm->charno );
  t("  data       = 0x%X", pgm->data );
  }
