#ifndef inctextTH
#define inctextTH

/* this file is to be included by text.c and by users of text.h */

typedef struct {   /* continuous encoding of array of bitvalues */
  u8	val:1	 ; /* 1: was ink ; 0 : was no ink . (ie inkedness before this pixel) */
  u8	rep:7	 ; /* offset from pixel in previous bwtoggle (or from 0) along glyphscanline*/
  } bwtoggle ;

typedef struct {   /* continuous encoding of array of bytevalues */
  u16	val:4	 ; /* gryscale value of borderpoint		 */
  u16	rep:12	 ;
  } grytoggle ;

/* bwmap is a continuous list of b/w pixelvaluebits, stored in uchar[],		 */
/*   MSb represents leftmost pixel, not an integer number of bytes per line !	 */

/* grymap is a continuous list of 4bit gry values, stored in uchar[],		 */
/*   most significant nibble holding leftmost pixel,				 */
/*   not an integer number of bytes per line !					 */

typedef struct { /* 20 byte */
  u16	 inkboxwi	 ; /* if glyph has no ink, then inboxwi/hi are zero		 */
  u16	 inkboxhi	 ;
  s16	 inkboxoffx	 ; /* x/y offsets of topleft of inkbox				 */
  s16	 inkboxoffy	 ; /*   relative to (0;0) of glyph.				 */
			   /*   in GRAPHDRAWING coords (ie higher pos has higher y)	 */
  u32	 logiwi:12	 ;
   u32	 gry:1		 ; /* dataformat: is determined by gry/b&w and toggle/pixmap	 */
   u32	 toggle:19	 ; /* DOUBLE USE: as param to StoreGlyph: bool: 0 or 1 ;	 */
			   /*   after StoreGlyph: countof toggles,			 */
			   /*   for easy compute storagesize				 */
  u32	 charno		 ; /* unicode number of character				 */
  u32	 data		 ; /* see GetFont.doc for dataformats. inkless glyph has data 0	 */
  } glyphmeta ;		   /* data and charno MUST be last, for comparing glyphs	 */

typedef struct {
  s32	 xid		 ; /* negative means that it is not loaded by X			 */
  uchar	 xlfd[102]	 ;
  uchar	 foundry[51]	 ; /* there are two 'terminal' fonts, from different foundries	 */
  uchar	 style[51]	 ; /* usually called 'fontname'. includes alt/bold/italique	 */
  u32	 firstchar	 ; /* Only valid if cglyphs is nonzero				 */
  u32	 lastchar	 ;
  u32	 cglyphs	 ; /* can be null (for using this struct for fontselect)	 */
			   /* countof non-default glyphs. is size of array of fontmetas	 */
  u32	 glyphs		 ; /* offset of array of glyphmeta				 */
  u16	 unused3	 ;
  u16	 nomheight	 ; /* 'pixelsize' as reported by xlfd				 */
  u16	 fontheight	 ; /* advised height, from fontabove+fontbelow.			 */
  u16	 avgstride	 ; /* averaged over all chars in font.				 */
  u16	 fontabove	 ; /* non-negative */
  u16	 fontbelow	 ; /* non-negative */
  s16	 inkhighest	 ;
  s16	 inklowest	 ;
  s16	 inkleftest	 ;
  s16	 inkrightest	 ;
  u16	 maxstride	 ;
  u16	 minstride	 ;
  u8	 resx		 ; /* from xlfd */
  u8	 unused1	 ;
  u8	 weight		 ; /* weight averaged over all glyphs. really float. range [0;1] */
  s8	 slant		 ; /* cos of slant of 'verticals'. really a float. range [-1;1]	 */
  u8	 serif		 ; /* ratio of seriflength/inkboxheight of'I'.float. range [0;1] */
  u8	 beauty		 ; /* beauty-contest not formalized yet. range [0;255]		 */
  u8	 usecount	 ;
  u8	 unused2:3	 ;
   u8	 bold:1		 ; /* bool, from xlfd						 */
   u8	 italic:1	 ; /* bool, from xlfd						 */
   u8	 equiwi:1	 ; /* 0: variable-width font. 1:fixed-widht font.		 */
   u8	 encoding:2	 ; /* 1:unicode, 2:iso8859-1, 0: other				 */
  } fontmeta ; /* 256 byte */

typedef struct {
  u8	sig[4]		 ; /* 'g' 'l' 'y' '0' ; glyphfile formatversion 0 */
  u32	order		 ; /* 0x01020304 */
  } glyheader ;


#endif
