/* X-related functions for canvas */

/* InitXimageGlobals MUST BE RUN BEFORE NewXimage() is used	 */
/* InitXimageGlobals MUST NOT be run BEFORE Connect()		 */

/*----------------- include ------------------------------------------------------------*/

#include "defines.h"		 /* make GNU extensions available	 */

/* X's headers : */
#include <X11/Xlib.h>
#include <X11/Xutil.h>		 /* XVisualInfo				 */

#include "myincludesT.h"	 /* orgb				 */
/*#include "xT.h"*/		 /* gcval, mkwindata, new X names	 */
#include "myincludes.h"		 /* error/debug routines etc		 */
#include "clickdragT.h"		 /* #defines of PRESS etc.		 */
#include "clickdrag.h"		 /* RegisterClickdrag()			 */

/*----------------- public --------------------------------------------------------------*/

int		 xfh				 ; /* filehandle for connection to X	 */
uint32_t	 ButPrsFlag = ButtonPressMask	 ; /* these are defined in X.h		 */
uint32_t	 ButRelFlag = ButtonReleaseMask	 ; /*   and i need to use them in main.c */
uint32_t	 LeaveFlag  = LeaveWindowMask	 ;
int		 ximgpixelsize			 ; /* size of pixels in data of XImage	 */
XImage*		 pxim				 ;
ulong		 (*OrgbToPixval)( orgb o )	 ; /* to be called as OrgbToPixval(o)	 */

/*----------------- extern --------------------------------------------------------------*/

extern int	 verbose			 ; /* myincludes.c	 */
extern int	 winwi, winhi			 ; /* main.c		 */

/*----------------- types ---------------------------------------------------------------*/

typedef struct {		 /* data for making a window */
  Window	 win		 ; /* filled in by makewindow	 */
  int		 wi		 ; /* specified to makewindow	 */
  int		 hi		 ; /* specified to makewindow	 */
  orgb		 bg		 ; /* specified to makewindow	 */
  int		 backstore	 ; /* specified to makewindow NotUseful WhenMapped Always */
  int		 bitgrav	 ; /* specified to makewindow ForgetGravity	 */
  } mkwindata ;

typedef struct gcvalS {
  int	 func		 ; /* 0 logical operation [GXcopy]			 */
  ulong	 planes		 ; /* 1 plane mask [ all 1s]				 */
  ulong	 fg		 ; /* 2 foreground pixel [0]				 */
  ulong	 bg		 ; /* 3 background pixel [1]				 */
  int	 linewi		 ; /* 4 line width (in pixels) [0]			 */
  int	 linestyle	 ; /* 5 [LineSolid], LineOnOff Dash, LineDoubleDash	 */
  int	 capstyle	 ; /* 6 CapNotLast, [CapButt], CapRound, CapProjecting	 */
  int	 joinstyle	 ; /* 7 [JoinMiter], JoinRound, JoinBevel		 */
  int	 fillstyle	 ; /* 8 [FillSolid], FillTiled, FillStippled FillOpaqueStippled	 */
  int	 fillrule	 ; /* 9 [EvenOddRule], WindingRule				 */
  int	 arcmode	 ; /*22 ArcChord, [ArcPieSlice]					 */
  Pixmap tile		 ; /*10 tile pixmap for tiling operations [see after typedef]	 */
  Pixmap stipple	 ; /*11 stipple 1 plane pixmap for stippling [see after typedef] */
  int	 tsxorigin	 ; /*12 offset for tile or stipple operations [0]		 */
  int	 tsyorigin	 ; /*13 "							 */
  Font	 font		 ; /*14 default text font [implementation dependent]		 */
  int	 subwinmode	 ; /*15 [ClipByChildren], IncludeInferiors			 */
  Bool   graphexp	 ; /*16 boolean, "exposures should be generated [true]		 */
  int	 clipxorigin	 ; /*17 origin for clipping [0]					 */
  int	 clipyorigin	 ; /*18 "							 */
  Pixmap clipmask	 ; /*19 bitmap clipping; (or use XSetClipRectangles) [None]	 */
  int	 dashoffset	 ; /*20 patterned/dashed line information [0]			 */
  char	 dashes		 ; /*21 [ 4 (ie: list[4,4]) ]					 */
  /* values for my additions : */
  GC		 gc	 ; /*31 keep X's identifier in struct.			    GCGC */
  ulong		 mask	 ; /*30 register which elements have been set in this gc.   GCMA */
  struct gcvalS* pNext	 ; /*29 handle lists of GCs.				    GCNE */  
  void*		 clips	 ; /*28 clip to restrict to area.			    GCCL */
  void*		 ppifw	 ; /*27 register info about font			    GCFI */
  void*		 colt	 ; /*26 NULLterminated list of colors. bg=[0] fg=[1]	    GCCT */
  void *	 link	 ; /*25 for application to keep additional data.	    GCLI */
  int		 cuse	 ; /*24 count of users of this gcval. free after no users.  GCCU */
  /* maybe i'll ever need tile and/or stipple, but haven't needed them yet. #24, #23 */
  } gcval ; /* NOTE that i changed my elements to void* here. */

/*----------------- defines -------------------------------------------------------------*/

#define XConDat				 Display
#define XOpenConnection			 XOpenDisplay
#define XPaltypDat			 Visual
#define XPaltypId			 VisualID
#define XPaltypDatToo			 XVisualInfo
#define XPalId				 Colormap
#define XPalDat				 XVisualInfo
#define XPaltypDatP			 visual
#define XGetDefaultpaletteCel		 XDisplayCells
#define XGetDefaultpaletteId		 XDefaultColormap
#define XGetPalettecolors		 XQueryColors
#define XAllocateSharableColor		 XAllocColor
#define XAllocatePrivateColors(a,b,c,d)	 XAllocColorCells(a,b,0,NULL,0,c,d)
#define XUnallocateColors(a,b,c,d)	 XFreeColors(a,b,c,d,0ul)
#define XCreatePalette			 XCreateColormap
#define XWinId				 Window
#define XDrawsetId			 GC
#define XDrawsetDat			 XGCValues

/* these are defines for bits in mask for gcval values : */
/* standard X bits for GCValues mask : */
#define GCFU GCFunction
#define GCPL GCPlaneMask
#define GCFG GCForeground
#define GCBG GCBackground
#define GCLW GCLineWidth
#define GCLS GCLineStyle
#define GCCS GCCapStyle
#define GCJS GCJoinStyle
#define GCFS GCFillStyle
#define GCFR GCFillRule
#define GCTI GCTile
#define GCST GCStipple
#define GCTX GCTileStippleXOrigin
#define GCTY GCTileStippleYOrigin
#define GCFO GCFont
#define GCSM GCSubwindowMode
#define GCGE GCGraphicsExposures
#define GCCX GCClipXOrigin
#define GCCY GCClipYOrigin
#define GCCM GCClipMask
#define GCDO GCDashOffset
#define GCDL GCDashList
#define GCAM GCArcMode
/* for my additions : */
#define GCGC (1L << 31)
#define GCMA (1L << 30)
#define GCNE (1L << 29)
#define GCCL (1L << 28)
#define GCFI (1L << 27)
#define GCCT (1L << 26)
#define GCLI (1L << 25)
#define GCCU (1L << 24)

/*----------------- filescope declares -------------------------------------------------*/

void		 InitColor(			 orgb* publicolors, int cel	 );
static void	 MakeWindow(			 mkwindata * pmwd		 );
static gcval	 SimpleGcval(			 void				 );
static void	 GetGC(				 gcval* pg, Window w		 );
static void	 SwitchFuncpointersForBpp(	 void				 );
static void	 MakePalette666(		 void				 );
static void	 Harmonize(			 void				 );
static void	 FillImpOf565(			 void				 );
static uchar	 ImpOfOrgb666(			 orgb o				 );
static ulong	 OrgbToU16(			 orgb o				 );
static orgb	 U16ToOrgb(			 ulong u			 );
static int	 Bubble(			 int require			 );
static uchar	 ComputeImpOfOrgb666(		 orgb o				 );

/*
.static void Connect( int screenno ){
.static void InitXimageGlobals( void ){
.static XImage* NewXimage( int getwi, int gethi ){
.void InitX( void ){
.void RenderXimage( void ){
.static int Coldif( orgb o1, orgb o2 ){
.static int ColdifRandom( orgb o ){
.static void ComputeMRow( short imp ){
.static void ComputeM( void ){
.static void GetGC( gcval * pg, Window w ){
.int StoreXimage( int bufix ){
.int LoadXimage( int bufix ){
.int FreeXimagebuf( int bufix ){
.void PollEvents( void ){
.void SetEventmask( char c, ulong newmask ){
.void InitEvents( void ){
*/

/*----------------- filescope ----------------------------------------------------------*/

static XConDat*		 pCon		 ;
static int		 screenix	 ; /* normally 0, but may be changed		 */
static XPaltypDatToo	 xdefpaltypdat	 ; /* pointo XVisualInfo of default visual	 */
static int		 bpp		 ;
static Window		 myw		 ;
static Window		 parw		 ;
static Window		 rootw		 ; /* actually identical to parw.		 */
static gcval		 gcv		 ;
static GC		 mygc		 ; /* a general purpose gc, usable on myw .	 */
static int		 rootwi		 ;
static int		 roothi		 ;
static uint8_t		 ximgpixsizepow	 ; /* pixelsize::pixsizepow : 1:0 2:1 4:2	 */
static int		 ximgunitsize	 ; /* pixels per line must be multiple of this	 */
static orgb		 owhite		 = { 0xFF, 0xFF, 0xFF, 0xFF };
static orgb		 oblack		 = { 0xFF, 0, 0, 0 } ;

/* bpp-dependent funcpointers : */
static orgb	 (*PixvalToOrgb)( ulong	 p );	 /* to be called as PixvalToOrgb(u) */

static	 orgb	 xdefpal[256]		 ; /* XDefaultPalette as orgbs. .o is validity	 */
static	 orgb	 OrgbOfImp[256]		 ; /* my (un-reordered) palette. o has validity. */
static	 uchar	 (*ImpOfOrgb)(orgb o)	 ; /* pointo lookup() in my (un-reordered) palette */
static	 uchar	 IxpOfImp[256]		 ; /* table of (desired) reorderings.		 */
static	 orgb	 OrgbOfIxp[256]		 ; /* my re-ordered palette.			 */
static	 XPalId	 mycolmap		 ; /* X id for my private color table. not used. */

/* shared temporaries : */
static	 uchar	 IxpOfImpValid[256]	 ; /* keeps status of elements of IxpOfImp[]	 */
static	 uchar	 IxpAssigned[256]	 ; /* ie: assigned to some Imp of my colortable	 */
static	 int	 M[256][256]		 ; /* "Match", same as Colordiff[][]		 */

static uchar grytable[] = {					 /* faster than a switch */
	      0, /* black	 */
  217, 218,  43, /* grey 111	 */
  220, 221,  86, /* grey 222	 */
  223, 224, 129, /* grey 333	 */
  226, 227, 172, /* grey 444	 */
  229, 230, 215	 /* white	 */
  };

static uchar ImpOf565[0x10000] ;

static void*	 ximagebufs[100]	 ; /* speedup animations */

/*----------------- ConnectToXserver -------------------------------------------------------*/

static void Connect( int screenno ){
  char		 screenstring[]	 = ":0.0" ;
  int		 i		 ;
  XVisualInfo	 xvi		 ;
  XVisualInfo *	 pxvi		 ;
  int		 type		 ;
  char*		 vistyp[8]	 ;
  int		 supported	 ;
  vistyp[StaticGray]  = "StaticGray: monochrome readonly"		 ;
  vistyp[GrayScale]   = "GrayScale: monochrome r/w"			 ;
  vistyp[StaticColor] = "StaticColor: color undecomposed readonly"	 ;
  vistyp[PseudoColor] = "PseudoColor: color undecomposed r/w"		 ;
  vistyp[TrueColor]   = "TrueColor: color decomposed readonly"		 ;
  vistyp[DirectColor] = "DirectColor: color decomposed r/w"		 ;
  if( (screenno < 0 ) | (screenno > 9) ){
    w("invalid screennumber: %d ; using default", screenno );
    screenno = 0 ;
    }
  screenix = screenno ;  
  if( screenno != 0 ){ screenstring[3] = '0' + screenno ; }
  pCon = XOpenDisplay( screenstring );
  if( pCon == NULL ){ ex("failed to connect to X server"); }
  xfh = ConnectionNumber( pCon );
  if( xfh <= 0 ){ e("got nonpositive connection number %d", xfh ); }
  v("filehandle of connection to X is %d", xfh );
  /* get info about default visual : */
  xvi.visual	 = XDefaultVisual(	 pCon, screenix	 );
  xvi.visualid	 = XVisualIDFromVisual(	 xvi.visual	 );
  pxvi		 = XGetVisualInfo(	 pCon, VisualIDMask, &xvi, &i );
  if( (pxvi == NULL) || (i < 0) || (i > 1) ){ ex("cant find info on default visual."); }
  xdefpaltypdat = *pxvi ;
  v("DefaultVisual: id=0x%X screen=0x%X depth=0x%X=%dbpp class=0x%X=%s\n bitsperrgb=0x%X",
    (uint)pxvi->visualid, (uint)pxvi->screen, pxvi->depth, pxvi->depth,
    pxvi->class, vistyp[pxvi->class], pxvi->bits_per_rgb );
  v("DefaultVisual: redmask=%lX greenmask=%lX bluemask=%lX colormapsize=%d",
    pxvi->red_mask, pxvi->green_mask, pxvi->blue_mask, pxvi->colormap_size );
  /* find bpp : */
  bpp  = xdefpaltypdat.depth ;
  type = xdefpaltypdat.class ;
  supported = 0 ;
  if(  (bpp == 8 ) && (type == PseudoColor) ){ supported = 1 ; }
  if( ((bpp == 16) || (bpp == 24) || (bpp == 32))
    && ((type == DirectColor) || (type == TrueColor)) ){ supported = 1 ; }
  if( supported != 1 ){
    ex("unsupported visual: bpp=%d and visualtype=%d=%s", bpp, type, vistyp[type] );
    }
  /* cant initialize color-lookup-table for 8bpp, because it needs specifying public colors */
  /* if using 16 bpp, check that encoding is 565 : */
  if( bpp == 16 ){
    if(  (xdefpaltypdat.red_mask   != 0xF800)
      || (xdefpaltypdat.green_mask != 0x07E0)
      || (xdefpaltypdat.blue_mask  != 0x001F)
      ){ ex("unsupported pixval encoding for 16 bpp"); }
    }
  /* if pixelsize more than 1, verify that byte-order of server equals that of client : */
  if( ximgpixelsize > 1 ){
    uint16_t	 ui16		 ;
    uint16_t*	 pui16		 ;
    uint8_t*	 pui8		 ;
    char	 cliec, servc	 ;
    servc = (XImageByteOrder(pCon)==MSBFirst)? 'M' : 'L' ;
    ui16 = (uint16_t) 0x0102 ; pui16 = &ui16 ; pui8 = (uint8_t*) pui16 ;
    cliec = (*pui8==1)? 'M' : 'L' ;
    if( cliec != servc ){
      w("ximage byteorder is %c for client but %c for server",cliec,servc); }
    }
  parw  = XDefaultRootWindow( pCon );
  rootw = XDefaultRootWindow( pCon );
  rootwi = DisplayWidth(  pCon, screenno );
  roothi = DisplayHeight( pCon, screenno );
  }

/*----------------- InitXimageGlobals --------------------------------------------------*/

static void InitXimageGlobals( void ){
  XImage*		 pxim	 ;
  XPixmapFormatValues*	 pxpfv	 ;
  int			 cel, i	 ;
  int			 bpp	 ;
  int			 bitpad	 ;
  pxim = XGetImage( pCon, parw, 0, 0, 1, 1, 0xFFFFFFFF, ZPixmap );
  if( pxim         == NULL    ){ ex("NULL pxim"); }
  if( pxim->format != ZPixmap ){ ex("ximage format is not ZPixmap"); }
  bpp = pxim->bits_per_pixel ;
  ximgpixelsize = bpp >> 3 ;
  if( pxim->width != 1 ){
    e("width is not 1 (width=%d bytes_per_line=%d bitmap_pad=%d)",
      pxim->width, pxim->bytes_per_line, pxim->bitmap_pad );
    ximgunitsize  = (pxim->bitmap_pad) >> 3 ;
    }
  else { ximgunitsize = pxim->bytes_per_line ; }
  bitpad = ximgunitsize << 3 ; cel = 0 ;
  pxpfv = XListPixmapFormats( pCon, &cel );
  for( i=0 ; i<cel ; i++ ){
    if( pxpfv[i].depth != bpp ){ continue ; }
    if( pxpfv[i].scanline_pad > bitpad ){ bitpad = pxpfv[i].scanline_pad ; }
    }
  ximgunitsize = bitpad >> 3 ;
  if(   ximgpixelsize == 1 ){ ximgpixsizepow = 0 ; }
  elif( ximgpixelsize == 2 ){ ximgpixsizepow = 1 ; }
  else                      { ximgpixsizepow = 2 ; }
  vvvv("ximgpixelsize = %d  ximgunitsize = %d  ximgpixsizepow = %d",
    ximgpixelsize, ximgunitsize, ximgpixsizepow );
  XFree(pxim);
  }

/*----------------- NewXimage --------------------------------------------------------------*/

/* An ximage can not be larger than rootwindow, or it would cause an XError.		 */
/* An ximage's linesize (in bytes) must be a multiple of ximgunitsize			 */
/*   which is on my box always 32bits, even for smaller bpp.				 */

static XImage* NewXimage( int getwi, int gethi ){
  XImage*	 pxim	 ;	 /* (for values of this struct, see /my/test/x/ximage2.c) */
  if( pCon == NULL ){ ex("NULL pCon"); }
  if( getwi <= 0      || gethi <= 0     ){ ex("requested sizes are not positive"); }
  if( getwi >  rootwi || gethi > roothi ){ ex("requested sizes dont fit on rootwindow"); }
  if( ((getwi * ximgunitsize) & 7) != 0 ){ ex("linewidth is not a whole number of bytes");}
  pxim = XGetImage( pCon, parw, 0, 0, getwi, gethi, 0xFFFFFFFF, ZPixmap );
  if( (pxim->bits_per_pixel>>3) != ximgpixelsize ){
    e("imagepixelsize=%d but am using pixelsize=%d",(pxim->bits_per_pixel>>3),ximgpixelsize );
    }
  if( (pxim->bitmap_pad>>3) != ximgunitsize ){
    e("bitmap_pad is %d but am using unitsize %d", (pxim->bitmap_pad>>3), ximgunitsize );
    }
  return( pxim );
  }

/*----------------- InitX -----------------------------------------------------------------*/

void InitX( void ){
  orgb		 olist[2]	 ;
  mkwindata	 mwd		 ;

  Connect(0);
  olist[0] = owhite ; olist[1] = oblack ; InitColor( olist, 2 );
  mwd.wi = winwi ;
  mwd.hi = winhi ;
  mwd.bg = owhite ;
  mwd.backstore = NotUseful ;
  mwd.bitgrav = ForgetGravity ;
  MakeWindow( &mwd );		 /* window id is put in 'myw' */
				 /* this sets window's eventmask to StructureNotifyMask */
  myw = mwd.win ;
  gcv = SimpleGcval(); GetGC( &gcv, myw ); mygc = gcv.gc ;
  InitXimageGlobals();
  pxim = NewXimage( winwi, winhi );
  }

/*----------------- RenderXimage --------------------------------------------------------*/

void RenderXimage( void ){
  XPutImage( pCon, myw, mygc, pxim, 0, 0, 0, 0, winwi, winhi );
  XFlush(pCon);
  }


/****************** InitColor ************************************************************/

/* when 8 bpp, alloc palette and reorder elements to harmonize.				 */
/* also register contents of palette (in a format of my own), and create reverse table too. */

void InitColor(
  orgb * PublicColors   , /* list of colors to allocate in default map.
			     NULL specifies empty list */
  int	 PublicColorsCel  /* countof elements in PublicColors.			 */
  ){
  Colormap	 xDefpalId	 ; /* identifier				 */
  int		 xDefpalCel	 ; /* countof elements				 */
  XColor	 xDefpal[256]	 ;
  /* tmp variables : */
  XColor *	 pxc		 ;
  XColor	 xc		 ;
  orgb *	 po		 ;
  orgb		 o		 ;
  short		 imp		 ; /* index in my palette. short for looptests.	 */
  short		 ixp		 ; /* index in x's palette			 */
  int		 i, ck, count	 ;
  ulong		 ul		 ;
  if( (bpp != 8) && (bpp != 16) && (bpp != 24) && (bpp != 32) ){
    ex("unsupported bpp %d", bpp ); }
  SwitchFuncpointersForBpp();
  if( bpp != 8 ){ return ; }
  /* get info about default colortable : */
  xDefpalId	 = XGetDefaultpaletteId(  pCon, screenix );
  xDefpalCel	 = XGetDefaultpaletteCel( pCon, screenix );
  if( xDefpalCel != 256 ){
    ex("XGetDefaultpaletteCel returned %d insteadof 256", xDefpalCel ); }
  /* allocate colors in X's default colortable, if desired, and as far as possible : */
  if( PublicColors != NULL ){
    xc.flags = DoRed | DoGreen | DoBlue ; xc.pad = 0 ;
    for( i=0 ; i<PublicColorsCel ; i++ ){ po = &PublicColors[i] ;
      xc.red   = (po->r << 8) | po->r ;
      xc.green = (po->g << 8) | po->g ;
      xc.blue  = (po->b << 8) | po->b ;
      ck = XAllocateSharableColor( pCon, xDefpalId, &xc );
      if( ! ck ){ break ; }			 /* no space left in X's table. that's OK */
      }
    }
  /* create harmonized private colormap, and color-lookup-tables in my format : */
  MakePalette666();
  /* get contents of X's default colortable : */
  for( i=0 ; i<256 ; i++ ){ xDefpal[i].pixel = i ; }
  XGetPalettecolors( pCon, xDefpalId, xDefpal, 256 );
  /* also make this palette available in orgb and ics formats : */
	 /* DOES NOT CREATE ICS PALETTE */
  for( ixp=0, pxc=xDefpal, po=xdefpal ; ixp<256 ; ixp++, pxc++, po++ ){
    po->r = pxc->red >> 8 ;
    po->g = pxc->green >> 8 ;
    po->b = pxc->blue >> 8 ;
    po->o = 0xFF ;
    }
  /* Initialize helper arrays : */
  for( imp=0 ; imp<256 ; imp++ ){ IxpOfImpValid[imp] = 0 ; }
  for( ixp=0 ; ixp<256 ; ixp++ ){ IxpAssigned[ixp] = 0 ; }
  /* Only way to find out which elements are valid and which are not	 */
  /*   is to test by reserving r/w colortable entries :			 */
  count = 0 ;
  for( i=0 ; i<256 && XAllocatePrivateColors( pCon, xDefpalId, &ul, 1 ) ; i++ ){
    if( ul >= 256 ){ e("pixelvalue out of range"); continue ; }
    xdefpal[ul].o = 0 ; count++ ;
    }
  t("there are %d unused entries in X's default colortable", count );
  /* now that i know which ones are not in use, i can free all cells again : */
  for( ixp=0 ; ixp<256 ; ixp++ ){
    if( xdefpal[ixp].o != 0 ){ continue ; }
    ul = (ulong) ixp ;
    XUnallocateColors( pCon, xDefpalId, &ul, 1 );
    }
  Harmonize();
  /* DEBUG: verify that all was done : */
  for( imp=0 ; imp<256 ; imp++ ){
    if( ! IxpOfImpValid[imp] ){ e("imp %d has no ixp", imp ); } }
  for( ixp=0 ; ixp<256 ; ixp++ ){
    if( ! IxpAssigned[ixp]   ){ e("ixp %d has no imp", ixp ); } }
  /* if any entry in my palette is unused, give it colors of it's ixp : */
  for( imp=0 ; imp<256 ; imp++ ){
    if( OrgbOfImp[imp].o ){ continue ; }
    if( ! IxpOfImpValid[imp] ){ e("ixp of imp not valid"); }
    ixp = IxpOfImp[imp] ;
    OrgbOfImp[imp] = xdefpal[ixp] ;
    }
  /* fill OrgbOfIxp[] : */
  for( imp=0 ; imp<256 ; imp++ ){
    ixp = IxpOfImp[imp] ;
    OrgbOfIxp[ixp] = OrgbOfImp[imp] ;
    }
  /* create private colormap : */
  mycolmap = XCreatePalette(
    pCon, XDefaultRootWindow(pCon), xdefpaltypdat.XPaltypDatP, AllocAll);
  for( ixp=0 ; ixp<256 ; ixp++ ){
    o = OrgbOfIxp[ixp] ;
    xc.pixel = ixp ;
    xc.red   = o.r << 8 | o.r ;
    xc.green = o.g << 8 | o.g ;
    xc.blue  = o.b << 8 | o.b ;
    xc.flags = DoRed | DoGreen | DoBlue ;
    XStoreColor( pCon, mycolmap, &xc );
    }
  FillImpOf565();					 /* create fast pixel lookup table */
  }

/****************** Palette 666 **********************************************************/

/* has 6x6x6 = 216 color basic palette, plus 10 greys */

/*----------------- MakePalette666 ------------------------------------------------------*/

static void MakePalette666( void ){
  uchar	 rl, gl, bl, i1, i2, i3, i	 ;
  orgb	 o				 ;
  short	 val				 ;
  o.o = 0xFF ;
  for( rl=0 ; rl<6 ; rl++ ){     o.r = rl * 51 ; i1 = rl * 6	    ;
    for( gl=0 ; gl<6 ; gl++ ){   o.g = gl * 51 ; i2 = (i1 + gl) * 6 ;
      for( bl=0 ; bl<6 ; bl++ ){ o.b = bl * 51 ; i3 = i2 + bl	    ;
								 /* 36 * rl + 6 * gl + bl */
        OrgbOfImp[i3] = o ;
        }
      }
    }
  /* fill in gry part of palette : */
  for( i=216, val=0 ; i<232 ; i++, val+=17 ){ o.r = o.g = o.b = val ; OrgbOfImp[i] = o ; }
  for( i=216        ; i<232 ; i+=3         ){ OrgbOfImp[i].o = 0 ; }  /* already in palette */
  /* set pointer to lookup routine : */
  ImpOfOrgb = ImpOfOrgb666 ;
  }

/*----------------- Speed-up array 'ImpOf565' -------------------------------------------*/

static void FillImpOf565( void ){
  ulong	 i ;
  uchar* p ;
  orgb	 o ;
  for( i=0, p=ImpOf565 ; i<0x10000 ; i++, p++ ){
    o = U16ToOrgb(i);
    *p = ComputeImpOfOrgb666(o);
    }
  }

/*----------------- ImpOfOrgb666 --------------------------------------------------------*/

static uchar ImpOfOrgb666( orgb o ){
  ulong fsf ;
  fsf = OrgbToU16(o);
  return( ImpOf565[fsf] );
  }

/*----------------- ComputeImpOfOrgb666 -------------------------------------------------*/

static uchar ComputeImpOfOrgb666( orgb o ){
  uchar	 ret, rl, gl, bl, min, max, gryix	 ;
  char	 isgry		 ;
  short	 sum		 ;
  int	 grytabix	 ;

  if( o.r < o.g ){ min = o.r ; max = o.g ; } else { min = o.g ; max = o.r ; }
  if( o.b < min ){ min = o.b ; }
  if( o.b > max ){ max = o.b ; }
  if(   max - min >= 37 ){ isgry = -1 ; }
  elif( max - min <= 21 ){ isgry =  1 ; }
  else                   { isgry =  0 ; }
  if( isgry >= 0 ){						 /* could be gry.	 */
    sum = o.r + o.g + o.b ;
    grytabix = (int) (sum * 15.0 / 765.0 + 0.5) ;		 /* max val is 765	 */
    gryix = grytable[ grytabix ] ;
    if( isgry > 0 ){ return( gryix ); }
    }
  rl = (o.r * 6) >> 8 ;			 /* right shift is faster than divide by 256	 */
  gl = (o.g * 6) >> 8 ;
  bl = (o.b * 6) >> 8 ;
  ret  = rl * 36 + gl * 6 + bl ;
  if( (isgry == 0) && (ret != gryix) ){	 /* need see whether color or grey is best	 */
    int		 D	 ;
    orgb	 ot	 ;
    short	 d	 ;
    ot = OrgbOfImp[ret  ]; d= ot.r-o.r ; D =d*d; d= ot.g-o.g ; D+=d*d; d= ot.b-o.b ; D+=d*d ;
    ot = OrgbOfImp[gryix]; d= ot.r-o.r ; D-=d*d; d= ot.g-o.g ; D-=d*d; d= ot.b-o.b ; D-=d*d ;
    if( D >= 0 ){ return( gryix ); }
    }
  return(ret);
  }

/*================= OrgbToPixval v.v. ===================================================*/

/*----------------- U32 -----------------------------------------------------------------*/

static ulong	 OrgbToU32( orgb o ){ return( (o.r << 16) | (o.g << 8) | o.b ); }

static orgb	 U32ToOrgb( ulong u ){
  orgb o ;
  o.r = (u >> 16) & 0xFF ;
  o.g = (u >>  8) & 0xFF ;
  o.b =  u        & 0xFF ;
  return(o);
  }

/*----------------- U16 -----------------------------------------------------------------*/

static ulong	 OrgbToU16( orgb o ){
  return( ((o.r & 0xF8) << 8) | ((o.g & 0xFC) << 3) | (o.b >> 3) ); }

static orgb	 U16ToOrgb( ulong u ){
  orgb	 o	 ;
  uchar	 b	 ; /* byte */
  b = (u & 0xF800) >> 8 ; o.r = b | (b >> 5) ;
  b = (u & 0x07E0) >> 3 ; o.g = b | (b >> 6) ;
  b = (u & 0x001F) << 3 ; o.b = b | (b >> 5) ;
  return(o);
  }

/*----------------- U8 ------------------------------------------------------------------*/

static ulong	 OrgbToU8( orgb	 o ){ return( IxpOfImp[ ImpOfOrgb(o) ] ); }
static orgb	 U8ToOrgb( ulong u ){ return( OrgbOfIxp[u] ); }

/*----------------- SwitchFuncpointersForBpp --------------------------------------------*/

static void SwitchFuncpointersForBpp( void ){
  if( bpp == 32 ){ OrgbToPixval = OrgbToU32 ; PixvalToOrgb = U32ToOrgb ; }
  if( bpp == 24 ){ OrgbToPixval = OrgbToU32 ; PixvalToOrgb = U32ToOrgb ; }
  if( bpp == 16 ){ OrgbToPixval = OrgbToU16 ; PixvalToOrgb = U16ToOrgb ; }
  if( bpp ==  8 ){ OrgbToPixval = OrgbToU8  ; PixvalToOrgb =  U8ToOrgb ; }
  }

/****************** Harmonize ************************************************************/

/*----------------- Coldif --------------------------------------------------------------*/

static int Coldif( orgb o1, orgb o2 ){
  /* computes color difference as square of component differences. Ignores .o .	 */
  int	 D, d	 ;
  d = o1.r - o2.r ; D  = d * d ;
  d = o1.g - o2.g ; D += d * d ;
  d = o1.b - o2.b ; D += d * d ;
  return( D );
  }

/*----------------- ColdifRandom --------------------------------------------------------*/

static int ColdifRandom( orgb o ){
  /* computes average colordifference as square of component differences */
  int	 d	 ;
  uchar	 r, g, b ;
  r = o.r ; g = o.g ; b = o.b ;
  d = SQFF - 0xFF * (r + g + b) + r*r + g*g + b*b ;
  return( d );
  }

/*----------------- ComputeM ------------------------------------------------------------*/

static void ComputeMRow( short imp ){
  short	 ixp	 ;
  orgb	 ooimp	 ;
  ooimp = OrgbOfImp[imp] ;
  for( ixp=0 ; ixp<256 ; ixp++ ){
    if(   IxpAssigned[ixp]	 ){ continue ; }
    elif( ! ooimp.o		 ){ M[imp][ixp] = BADMATCH ; }	 /* reserve for bad matches */
    elif( ! xdefpal[ixp].o	 ){ M[imp][ixp] = ColdifRandom( ooimp );		 }
    else			  { M[imp][ixp] = Coldif( ooimp, xdefpal[ixp] );	 }
    }
  }

static void ComputeM( void ){		 /* compute contents of colordiff array */
  short imp ;
  for( imp=0 ; imp<256 ; imp++ ){
    if( IxpOfImpValid[imp] ){ continue ; }
    ComputeMRow(imp);
    }
  }

/*----------------- Harmonize -----------------------------------------------------------*/

static void Harmonize( void ){
  short	 imp, ixp, ixp1			 ;
  int	 count, m, m1, require, didswap	 ;
  /* this code can be used after some imps have been assigned ixp's ,
     so first check that count of assigned imps equals countof imps with valid ixp : */
  count = 0 ;
  for( ixp=0 ; ixp<256 ; ixp++ ){ if( IxpAssigned[  ixp] ){ count++ ; } }
  for( imp=0 ; imp<256 ; imp++ ){ if( IxpOfImpValid[imp] ){ count-- ; } }
  if( count != 0 ){ ex("unequal countof assigned elements"); }
  ComputeM();						 /* initialize M[][]		 */
  for( imp=0 ; imp<256 ; imp++ ){			 /* initialize IxpOfImp[]	 */
    m1 = VERYMUCH ; ixp1 = INVALID ;
    if( IxpOfImpValid[imp] ){ continue ; }
    /* find best matching ixp : */
    for( ixp=0 ; ixp<256 ; ixp++ ){
      if( IxpAssigned[ixp]	 ){ continue ; }
      m = M[imp][ixp] ; if( m <  m1 ){ m1 = m ; ixp1 = ixp ; }
      }
    if( ixp1 < 0 ){ break ; }
    IxpOfImp[     imp ]	 = ixp1	 ;
    IxpOfImpValid[imp ]	 = 1	 ;
    IxpAssigned[  ixp1]	 = 1	 ;
    }
  /* change matchdeviation of unused entries in my palette to 0, to prepare for bubble : */
  for( imp=0 ; imp<256 ; imp++ ){
    if( OrgbOfImp[imp].o ){ continue ; }
    for( ixp=0 ; ixp<256 ; ixp++ ){ M[imp][ixp] = 0 ; }
    }
  for( require = 0x2000 ; require > 0x800 ; require >>= 1 ){ Bubble( require );		 }
  for( didswap = 2      ; didswap > 1     ;		  ){ didswap = Bubble(0) ;	 }
  }

/*----------------- Bubble --------------------------------------------------------------*/

static int Bubble( int require ){
  short	 lvi , lvo		 ; /* inner and outer loop values (for imp) */
  short	 ixpi, ixpo		 ;
  int	 mii , mio , moi , moo	 ;
  int	 maxold, maxnew		 ;
  int	 minold, minnew		 ;
  int	 didswap		 ;
  /* swap wherever it improves a deviation larger than or equal to $require : */
  didswap = 0 ;
  for( lvo=0 ; lvo<256 ; lvo++ ){
    ixpo = IxpOfImp[lvo] ;
    moo = M[lvo][ixpo] ;
    if( moo < require ){ continue ; }
    for( lvi=0 ; lvi<256 ; lvi++ ){	 if( lvi == lvo    ){ continue ; }
      ixpi = IxpOfImp[lvi] ;
      mii = M[lvi][ixpi] ;
      if( mii > moo ){ maxold = mii ; minold = moo ; } else { maxold = moo ; minold = mii ; }
      mio = M[lvi][ixpo] ;
      moi = M[lvo][ixpi] ;
      if( mio > moi ){ maxnew = mio ; minnew = moi ; } else { maxnew = moi ; minnew = mio ; }
      if( maxnew > maxold ){ continue ; }			 /* not an improvement */
      if( (maxnew == maxold) && (minnew >= minold) ){ continue ; } /* not an improvement */
      IxpOfImp[lvi] = ixpo ; IxpOfImp[lvo] = ixpi ;
      didswap++ ;
      ixpo = IxpOfImp[lvo] ; moo = M[lvo][ixpo] ; if( moo < require ){ break ; }
      }
    }
  return( didswap );
  }

/************************************************************************************************/

/*----------------- MakeWindow -----------------------------------------------------------------*/

static void MakeWindow( mkwindata * pmwd ){
  XSetWindowAttributes	 xswa		 ;
  ulong			 xswamask	 ;
  XEvent		 xe		 ;
  XAnyEvent *		 pxae		 ;
  ulong			 bgpixel	 ;

  /* Connect()   should already have run and set pCon, bpp and defvisinfo.	 */
  /* InitColor() should already have run and set mycolmap.			 */

  if( pCon == NULL ){ ex("pCon is NULL"); }
  v("making window");
  bgpixel  = OrgbToPixval( pmwd->bg );
  parw = DefaultRootWindow( pCon );		 /* specific to this routine */
  xswa.background_pixel	 = bgpixel		 ; xswamask  = CWBackPixel	 ;
  xswa.backing_store	 = pmwd->backstore	 ; xswamask |= CWBackingStore	 ;
  xswa.bit_gravity	 = pmwd->bitgrav	 ; xswamask |= CWBitGravity	 ;
  xswa.event_mask	 = StructureNotifyMask	 ; xswamask |= CWEventMask	 ;
  xswa.colormap		 = mycolmap		 ; xswamask |= CWColormap	 ;
  pmwd->win = XCreateWindow( pCon, parw,
          0, 0, pmwd->wi, pmwd->hi, 0,		 /* borwi 0 because borders dont show up */
          CopyFromParent, CopyFromParent, CopyFromParent, /* depth/class/visual */
          xswamask, &xswa );
  vv("created window");
  XMapWindow( pCon, pmwd->win ); XSync( pCon, 0 );

  /* wait for mapnotify event, or get killed due to expiring alarm : */
  SetTimeout( 3, 0, "did not receive map event", 1, NULL );
  while(1){
    XNextEvent( pCon, &xe ); XSync( pCon, 0 );
    pxae = (XAnyEvent *) &xe ;
    if( pxae->display != pCon	  ){ continue ; }
    if( pxae->window  != pmwd->win ){ continue ; }
    /* does not catch destroy event, as instead of receiving destroy event,	 */
    /*   program gets SIGPIPE, which terminates it.				 */
    if( pxae->type    != MapNotify ){ continue ; }
    v("received mapnotify event");
    if( pxae->send_event ){ t("this mapnotify event was sent by another client !"); }
    break ;
    }
  UnsetTimeout();
  }

/*----------------- SimpleGcval ---------------------------------------------------------*/

/* returns gcval with defaults for most basic elements.	 */
/* doesn't create a GC. doesn't alloc anything.		 */

static gcval SimpleGcval( void ){
  gcval	 ret	 ;
  ulong	 mask	 = 0 ;

  ret.func	 = GXcopy			 ; mask |= GCFU ;
  ret.fg	 = XBlackPixel(pCon,screenix)	 ; mask |= GCFG ;
  ret.bg	 = XWhitePixel(pCon,screenix)	 ; mask |= GCBG ;
  ret.planes	 = 0xFFFF			 ; mask |= GCPL ;
  ret.linewi	 = 0				 ; mask |= GCLW ;
  ret.linestyle	 = LineSolid			 ; mask |= GCLS ;
  ret.capstyle	 = CapButt			 ; mask |= GCCS ;
  ret.joinstyle	 = JoinMiter			 ; mask |= GCJS ;
  ret.fillstyle	 = FillSolid			 ; mask |= GCFS ;
  ret.fillrule	 = EvenOddRule			 ; mask |= GCFR ;
  ret.arcmode	 = ArcPieSlice			 ; mask |= GCAM ;
  ret.graphexp	 = 1				 ; mask |= GCGE ;
  ret.pNext	 = NULL				 ; mask |= GCNE ;
  ret.clips	 = NULL				 ;
  ret.ppifw	 = NULL				 ;
  ret.colt	 = NULL				 ;
  ret.link	 = NULL				 ; mask |= GCLI ;
  ret.cuse	 = 0				 ; mask |= GCCU ;
  mask |= GCMA ; ret.mask = mask ;
  return( ret );
  }

/*----------------- GetGC ---------------------------------------------------------------*/

static void GetGC( gcval * pg, Window w ){
  /* doesnt check for identical gcvals, so every caller gets a new one. */
  pg->gc = XCreateGC( pCon, w, pg->mask, (XGCValues*) pg ); pg->mask |= GCGC ;
  pg->cuse = 1 ; pg->mask |= GCCU ;
  }

/*================= multi-image-buffers =================================================*/
/* these are for accelerated (pre-computed) animations */

/*----------------- StoreXimage ---------------------------------------------------------*/

int StoreXimage( int bufix ){
  int size ;
  if( bufix < 0 || bufix > 100 ){ e("bufix out of range"); return(-1); }
  size = ximgpixelsize * winwi * winhi ;
  if( ximagebufs[bufix] == NULL ){ ximagebufs[bufix] = ckalloc( size ); }
  if( ximagebufs[bufix] == NULL ){ e("not enough memory"); return(-1); }
  memcpy( ximagebufs[bufix], pxim->data, size );
  return(0);
  }

/*----------------- LoadXimage ----------------------------------------------------------*/

int LoadXimage( int bufix ){
  int size ;
  if( bufix < 0 || bufix > 100 ){ e("bufix out of range"); return(-1); }
  size = ximgpixelsize * winwi * winhi ;
  if( ximagebufs[bufix] == NULL ){ e("empty ximagebuffer"); return(-1); }
  memcpy( pxim->data, ximagebufs[bufix], size );
  return(0);
  }

/*----------------- FreeXimagebuf -------------------------------------------------------*/

int FreeXimagebuf( int bufix ){
  if( bufix < 0 || bufix > 100  ){ e("bufix out of range"); return(-1); }
  if( ximagebufs[bufix] == NULL ){ e("already free");       return(-1); }
  ckfree( ximagebufs[bufix] );
  return(0);
  }

/*================= EVENTS ==============================================================*/

/*----------------- PollEvents ----------------------------------------------------------*/

void PollEvents( void ){
  XEvent	 xe	 ;
  XAnyEvent*	 pxae	 ;
  XButtonEvent*	 pxbe	 ;

  while( XPending(pCon) > 0 ){
    XNextEvent( pCon, &xe );
    pxae = (XAnyEvent*) &xe ;
    if( pxae->window != myw ){ continue ; }
    if(  pxae->type == Expose	      /* part of area needs to be redrawn */
      || pxae->type == GraphicsExpose /* part of area is src of screencopy, but not visible */
      ){ RenderXimage(); return ; }
    elif( pxae->type == ButtonPress ){
      pxbe = (XButtonEvent*) pxae ; vvv("press %d", pxbe->button );
      RegisterClickdrag( PRESS, pxbe->button, pxbe->x, pxbe->y, 0, 0 );
      }
    elif( pxae->type == ButtonRelease ){
      pxbe = (XButtonEvent*) pxae ; vvv("release %d", pxbe->button );
      RegisterClickdrag( RELEASE, pxbe->button, pxbe->x, pxbe->y, 0, 0 );
      }
    elif( pxae->type == LeaveNotify ){
      RegisterClickdrag( LEAVE, 0, 0, 0, 0, 0 ); vvv("leave");
      }
    }
  }

/*----------------- SetEventmask --------------------------------------------------------*/

/*
 eventmask is first StructureNotifyMask (to get window mapped event)
 and later (at init) is ExposureMask (to redraw when necessary)
 and to receive click and drags, it must be made to
   include ButtonPressMask and ButtonReleaseMask and LeaveWindowMask
 */

void SetEventmask( char c, ulong newmask ){
  static ulong		 eventmask	 ;
  XSetWindowAttributes	 xswa		 ;

  if(   c == '=' ){
    if( eventmask == newmask ){ return ; }
    eventmask = newmask ;
    }
  elif( c == '+' ){
    if( eventmask == (eventmask | newmask) ){ return ; }
    eventmask |= newmask ;
    }
  else {  /* '-' */
    if( (eventmask & newmask) == 0 ){ return ; }
    eventmask &= ~newmask ;
    }
  xswa.event_mask = eventmask ;
  XChangeWindowAttributes( pCon, myw, CWEventMask, &xswa );
  XSync( pCon, 0 );
  }

/*----------------- InitEvents ----------------------------------------------------------*/

void InitEvents( void ){
  SetEventmask( '=', ExposureMask );
  }
