/* ImoToJpeg.c */

/* compile with :
pushd /my/prive/projects/Canvas/v4/util
cc ImoToJpeg.c -o ImoToJpeg -W -Wall -ljpeg -I/my/prive/projects/Canvas/v4/source -O3
*/

/* STATUS: new version that uses only local includes for canvas
   Compiles (no problems). so probably works.
*/


/*
#include "defines.c"
#include "errordef.c"
#include "ph.c"
*/

#include "defines.h"

#include <stdio.h>
#include <string.h>
#include <jpeglib.h>

#include "myincludesT.h"	 /* filebuf, orgb, imoheader */
#include "myincludes.h"
#include "myincludes.c"

/*
#include "mem/ckalloc.c"
#include "file/FileExists.c"
#include "file/QualifyFilename.c"
#include "file/open.c"
#include "file/filebuf.c"
#include "time/timestamp.c"
#include "imo/imoheader.c"
*/

/*----------------- structs ---------------------------------------------------------------------*/

#define ROWSPERBUFFER 8				 /* using 1 is not slower than using 8 */

typedef struct {
  JSAMPROW	jrowpointers[ROWSPERBUFFER]	 ;
  JSAMPLE*	jsampbuf			 ; /* filled in dynamically cf ih.wi	 */
  int		wi				 ; /* number of valid pixels per row	 */
  int		hi				 ; /* number of valid rows		 */
  } jconvertbuf ;


/*----------------- globals ---------------------------------------------------------------------*/

char*	 infilename	 ;
char*	 outfilename	 ;
int	 needprogress	 ;

Filebuf	 ifb		 ; /* auto-mmaped with inputfile.				 */
jconvertbuf jcb		 ;
FILE *	 ofp		 ; /* for letting libjpeg write to.				 */

int	 ck		 ;
imoheader ih		 ;
orgb*	 pixbuf		 ; /* pointo start of pixeldata in mmap of imofile		 */
orgb*	 opixbuf	 ; /* poinro startofcomments or eof				 */
orgb	 obg		 ;
orgb	 owhite = {0xFF,0xFF,0xFF,0xFF};
orgb	 ogrey  = {0xFF,0x80,0x80,0x80};
orgb	 oblack = {0xFF,0x00,0x00,0x00};

/*----------------- help ------------------------------------------------------------------------*/

char * help[] = {
  "ImoToJpeg : read an imo file, and create a jpeg file containing same image.     ",
  "Syntax: ImoToJpeg [-v] [-p] <infilename> [<outfilename>] [-bw]                  ",
  "  -v causes verbose output for debugging.                                       ",
  "  -p makes it print progress reports                                            ",
  "  if <outfilename is not specified, it is created                               ",
  "    by stripping any '.imo' from <infilename> and appending '.jpeg'.            ",
  "  -b option specifies color to use for transparent pixels,                      ",
  "    default is black, -bw changes that to white, -bg to grey.                   ",
  NULL };

/*----------------- CheckArguments --------------------------------------------------------------*/

void CheckArguments( int argc, char ** argv ){
  int	 i	 ;
  char * arg	 ;

  verbose = 0 ;
  needprogress = 0 ;
  for( i=1 ; i<argc ; i++ ){ arg = argv[i] ;
    if( arg[0] == '-' ){
      if( arg[1] == 'v' ){ verbose++ ; needprogress = 1 ; continue ; }
      if( arg[1] == 'h' ){ phe(); }
      if( arg[1] == 'p' ){ needprogress = 1 ; continue ; }
      if( arg[1] == 'b' ){ if(   arg[2] == 'w' ){ obg = owhite ; }
                           elif( arg[2] == 'b' ){ obg = oblack ; }
                           elif( arg[2] == 'g' ){ obg = ogrey  ; }
                           else { ex("invalid background specifier"); }
                           continue ;
                           }
      ex("invalid option %s", arg );
      }
    if( infilename  == NULL ){ infilename  = QualifyFilename( arg ); continue ; }
    if( outfilename == NULL ){ outfilename = QualifyFilename( arg ); continue ; }
    e("invalid argument %s", arg ); phe();
    }

  if( infilename  == NULL ){ e("no infilename specified"); phe(); }
  infilename = QualifyFilename( infilename );
  if( outfilename == NULL ){
    int len ; char * p ;
    len = strlen( infilename );
    p = (char*) alloca( len + 7 ); if( p==NULL){ex("alloca failed"); }
    if( (len >= 4) && (0 == strcmp( infilename + len - 4, ".imo"))  ){ len -= 4 ; }
    strcpy( p, infilename );
    strcpy( p + len, ".jpeg");
    outfilename = QualifyFilename( p );
    }
  else { outfilename = QualifyFilename( outfilename ); }

  ck = FileExists( infilename, NULL );
  if( ck == 0 ){ ex("specified infile %s does not exist",infilename); }
  if( ck >  1 ){ ex("infile %s is not an ordinary file", infilename); }
  ck = FileExists( outfilename, NULL );
  if( ck >  1 ){ ex("outfile %s exists and is not an ordinary file",outfilename ); }
  if( (ck == 1) && verbose ){ w("overwriting existing outfile %s", outfilename ); }

  ifb = GetMmappedFilebuf( infilename, 1 );
  ofp = fopen( outfilename, "w" );
  if( ofp == NULL ){ esx("failed fopen_w %s", outfilename ); }

  v("infile=%s outfile=%s reportprogress=%d",infilename,outfilename,needprogress);
  }

/*----------------- GetImoheader ----------------------------------------------------------------*/

void GetImoheader( void ){ /* all i/o through global imoheader ih */
  int npix ;
  int filenpix ;
  int filesize ;

  filesize = ifb.ocont - ifb.cont ;
  if( (uint) filesize < sizeof(imoheader) ){ ex("not a valid imo file: too short for header"); }
  ih = *(imoheader*) ifb.buf ;
  pixbuf = (orgb*) (ifb.buf + sizeof(imoheader)) ;
  npix = ih.wi * ih.hi ;
  filenpix = (filesize - sizeof(imoheader))/sizeof(orgb) ;
  if( filenpix < npix ){
    e("file too short: header says wi=%d hi=%d but filesize can contain only %d pixels",
      ih.wi, ih.hi, filenpix );
    ih.hi = filenpix / ih.wi ;
    if( ih.hi <= 0 ){ ex("file not even large enough to contain one line. stopping."); }
    w("using hi=%d", ih.hi );
    }
  opixbuf = pixbuf + (ih.wi * ih.hi ) ;
  }

/*----------------- PrepareConvertbuffer --------------------------------------------------------*/

void PrepareConvertbuffer( void ){
  JSAMPLE *	 jrowpointer	 ;
  int		 i		 ;

  jcb.jsampbuf = ckalloc( ROWSPERBUFFER * ih.wi * 3 );
  for( i=0, jrowpointer=jcb.jsampbuf ; i < ROWSPERBUFFER ; i++ ){
    jcb.jrowpointers[i] = jrowpointer ;
    jrowpointer += ih.wi * 3 ;
    }
  jcb.wi = ih.wi ;
  jcb.hi = 0 ;
  }

/*----------------- FillConvertbuffer -----------------------------------------------------------*/

/* For largish images, this function is called repetitively.					 */
/* Converts one strip worth of input, and places it into convertbuffer.				 */
/* ifb.cont points to next part of olo to read, and cfb.ocont points to where to put it.	 */
/* Non-rectangular images are filled up with black.						 */
/* Input is always read a whole line at a time.							 */
/* Returns 1 if there is more input to convert, 0 at end of file.				 */
/* Image must be rectangular ; fill it up if necessary.						 */

/* nextline is number of next line to give data for */

void FillConvertbuffer( int nextline ){
  orgb		 o		 ;
  orgb*		 src		 ;
  JSAMPLE*	 dst		 ;
  int		 rowsavailable	 ;
  int		 i		 ;
  int		 npixels	 ;

  src = (orgb*) (ifb.buf + sizeof(imoheader) + nextline * ih.wi * sizeof(orgb)) ;
  dst = jcb.jsampbuf ;
v("src=%p dst=%p",src,dst);

  rowsavailable = ((void*)ifb.ocont - (void*) src)/sizeof(orgb) /ih.wi ;
v("rowsavailable=%d sizeof(orgb)=%d ih.wi=%d",rowsavailable, sizeof(orgb), ih.wi );

  if( rowsavailable <= 0 ){ /* something went wrong */
    e("something went wrong");
    return ;			 /* present last rows once more until compressor is satisfied */
    }
  jcb.hi = (ROWSPERBUFFER < rowsavailable)? ROWSPERBUFFER : rowsavailable ;
v("jcb.hi=%d",jcb.hi);

  npixels = jcb.hi * jcb.wi ;
  for( i=0 ; i < npixels ; i++ ){
    o = *(src++) ;

    /*--- modify transparent pixels ---*/

    if(   o.o ==   0 ){ o = obg ; }
    elif( o.o == 255 ){ ; }
    else {
      o.r = (o.r * o.o + obg.r * (255 - o.o)) /255 ;
      o.g = (o.g * o.o + obg.g * (255 - o.o)) /255 ;
      o.b = (o.b * o.o + obg.b * (255 - o.o)) /255 ;
      }

    /*--- end modify transparent pixels ---*/

    *(dst++) = o.r ; *(dst++) = o.g ; *(dst++) = o.b ;
    }
  }

/*----------------- Convert ---------------------------------------------------------------------*/

/* not setting errorhandler for now. maybe later, when things go wrong :). */

void ConvertWrite( void ){

  struct jpeg_compress_struct	 cinfo		 ;
  struct jpeg_error_mgr		 jerr		 ;
  int				 row_stride	 ;

/* step 1 : get compression and  error objects and initialize them */
  cinfo.err = jpeg_std_error( &jerr );
  jpeg_create_compress( &cinfo );
v("did step 1");
/* step 2 : open outputfile */
  /* fpo = fopen( outfilename, "w" ); already done in CheckArguments() */
  jpeg_stdio_dest( &cinfo, ofp );
v("did step 2");
/* step 3 : set compression parameters */
  cinfo.image_width      = ih.wi	 ;
  cinfo.image_height     = ih.hi	 ;
  cinfo.input_components = 3		 ;
  cinfo.in_color_space   = JCS_RGB	 ;
  jpeg_set_defaults( &cinfo );
v("did step 3");
/* step 3.5 : tweak compression parameters */
  /* adjust quality settings (or set any other parameter to a non-default value) : */
  /* jpeg_set_quality( &cinfo, quality, TRUE ); */

/* step 4 : start compression */
  jpeg_start_compress( &cinfo, TRUE );
v("did step 4");
/* step 5 : loop over all lines of file */
  row_stride = ih.wi * 3 ;
  while( cinfo.next_scanline < cinfo.image_height ){
v("doing lines from %d", cinfo.next_scanline );
    FillConvertbuffer( cinfo.next_scanline );		 /* also sets row pointers */
v("calling for write scanlines");
    jpeg_write_scanlines( &cinfo, jcb.jrowpointers, jcb.hi );
    }
v("did step 5");
/* step 6 : finish compression */
  jpeg_finish_compress( &cinfo );
  /* fclose( ofp ); closing outfile is done automatically by program termination */
v("did step 6");
/* step 7 : release JPEG compression object */
  jpeg_destroy_compress( &cinfo );
v("did step 7");
  }

/*----------------- main ------------------------------------------------------------------------*/

int main( int argc, char ** argv ){
  CheckArguments( argc, argv );			v("checked arguments");
  GetImoheader();				v("got imoheader");
  PrepareConvertbuffer();			v("prepared convertbuffer");
  ConvertWrite();				v("converted and wrote");

  /* closing is done by terminating. */
  return(0);
  }

