/* ImoToPng.c */

/* compile with :
pushd /my/prive/projects/Canvas/v4/util
cc ImoToPng.c -oImoToPng -I/my/prive/projects/Canvas/v4/source -lpng -W -Wall
*/

/* STATUS : converted to using local includes. compiled flawlessly. probably works.
*/

/*----------------- include ------------------------------------------------------------*/

#include "defines.h"
/*
#include "defines.c"
#include "errordef.c"
*/
#include <png.h>

#include "myincludesT.h"
#include "myincludes.h"
#include "myincludes.c"
/*
#include "file/QualifyFilename.c"
#include "file/filebuf.c"
#include "imo/orgbT.h"
#include "imo/imoheader.c"
#include "ph.c"
*/

/*----------------- define -------------------------------------------------------------*/

typedef struct { uint8_t r ; uint8_t g ; uint8_t b ; uint8_t o ; } Rgbo ;

/*----------------- globals ------------------------------------------------------------*/

char* help[] = {
  "ImoToPng : read an imo file and create a png file with same imagecontent.",
  "syntax : ImoToPng ( -v)* $imofilename ($pngfilename)?                    ",
  "this is version 1 : offset and scale are not preserved.                  ",
  NULL } ;

png_structp	 ps		 ;
png_infop	 psi		 ;
char*		 imofilename	 ;
char*		 pngfilename	 ;


/*----------------- ParseArguments -----------------------------------------------------*/

void ParseArguments( int argc, char** argv ){
  int	 i	 ;
  char*	 arg	 ;

  verbose = 0 ; imofilename = pngfilename = NULL ;
  for( i=1 ; i < argc ; i++ ){ arg = argv[i] ;
    if( arg[0] == '-' ){
      if( arg[1] == 'v' ){ verbose++ ; continue ; }
      if( arg[1] == 'h' ){ phe(); }
      ex("invalid option %s", arg );
      }
    if(   imofilename == NULL ){ imofilename = arg ; }
    elif( pngfilename == NULL ){ pngfilename = arg ; }
    else { e("only two filenames allowed"); phe(); }
    }

  if( imofilename == NULL ){ e("no imofilename specified"); phe(); }

  imofilename = QualifyFilename( imofilename );
  if( pngfilename == NULL ){
    int len = strlen( imofilename );
    pngfilename = (char*) ckalloc( len + 5 );
    memcpy( pngfilename, imofilename, len + 1 );
    if(  0 == strncmp( pngfilename + len - 4, ".imo", 4 )
      || 0 == strncmp( pngfilename + len - 4, ".IMO", 4 ) ){
      memcpy( pngfilename + len - 4, ".png", 4 );
      }
    else { memcpy( pngfilename + len, ".png", 4 ); }
    }
  else { pngfilename = QualifyFilename( pngfilename ); }

  v("imofilename=%s pngfilename=%s verbose=%d", imofilename, pngfilename, verbose );
  }

/*----------------- Cleanup ------------------------------------------------------------*/

void Cleanup( void ){
  png_destroy_write_struct( &ps, &psi );
  }

/*----------------- OrgbRowToRgbo ------------------------------------------------------*/

/* need to write custom transformation function : */

void OrgbRowToRgbo( orgb* po, int wi, Rgbo* pr ){
  /* pr is pointer to first mini-array */
  int i ;
  if( po == NULL || pr == NULL || wi <= 0 ){ ex("invalid arg"); }
  for( i=0 ; i < wi ; i++, po++, pr++ ){
    pr->r = po->r ; pr->g = po->g ; pr->b = po->b ; pr->o = po->o ;
    }
  }

/*----------------- Convert ------------------------------------------------------------*/

/* libpng can not change dataformat 'on the fly' when encoding to png,		 */
/*   so i must convert dataformat from orgb to uchar rgbo[4] beforehand.	 */
/* But i do not want to change datasource (which is mmapped),			 */
/*   so i must use progressive writing, and convert per scanline.		 */
/* Therefore i must use low-level writing.					 */

void Convert( void ){
  imoheader*	 pih	 ;
  orgb*		 ro	 ;
  int		 wi, hi	 ;
  int		 i	 ;
  orgb*		 po	 ;
  Rgbo*		 pr	 ;

  /* get a filebuffer for imofile : */
  Filebuf fb ;
  fb = GetMmappedFilebuf( imofilename, 0 );
  pih = fb.buf ;
  ro = (orgb*) (fb.buf + sizeof(imoheader)) ;
  if( 0 != strncmp( pih->sig, "imo0", 4 ) ){ ex("invalid imo sig %s", pih->sig ); }
  if( pih->byteorder != 0x01020304 ){ ex("incorrect byteorder %X", pih->byteorder ); }
  wi = pih->wi ;
  hi = pih->hi ;
  /* offsets and scales are ignored in this version */

  /* initialize i/o to outfile : */
  FILE* ofp ;
  if( NULL == (ofp= fopen( pngfilename, "wb"))){ esx("nok open pngfile"); }

  /* get png structs : */
  if( NULL== (ps= png_create_write_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL ))){
    ex("nok create writestruct"); }
  if( NULL== (psi= png_create_info_struct(ps))){
    png_destroy_write_struct( &ps, (png_infopp) NULL );
    ex("nok create infostruct");
    }
  /* setup error handling return address */
  if( setjmp( png_jmpbuf(ps) ) ){ Cleanup(); ex("bailed"); }
  /* let libpng know about outfile : */
  png_init_io( ps, ofp );

  /* fill png_info struct with info about image : */
  png_set_IHDR( ps, psi, wi, hi, 8 /*bpp*/, PNG_COLOR_TYPE_RGB_ALPHA,
    PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT );

  /* here can later add comment texts for offsets and scale */

  png_write_info( ps, psi );

  pr = (Rgbo*) alloca( wi * 4 * sizeof(uchar) );
  for( i=0, po=ro ; i < hi ; i++, po+=wi ){
    OrgbRowToRgbo( po, wi, pr );
    png_write_row( ps, (png_bytep) pr );
    }

  /* finish write : */
  png_write_end( ps, psi );

  Cleanup();
  DiscardFilebuf( &fb );				 /* superfluous */
  if( 0 != fclose( ofp ) ){ es("nok fclose outfile"); }	 /* superfluous */
  }

/*----------------- main ---------------------------------------------------------------*/

int main( int argc, char** argv ){
  ParseArguments( argc, argv );
  Convert();
  return(0);
  }

