/* JpegToImo.c */

/* compile with :
install libjpg62-dev
pushd /my/prive/projects/Canvas/v4/util
cc JpegToImo.c -o JpegToImo -W -Wall -ljpeg -I/my/prive/projects/Canvas/v4/source -O3
*/

/* STATUS: converted to use local includes. compiles. probably works.
   TODO: remove commented-out timing statements
   TODO: remove debugging statements.
*/

/*
#include "defines.c"
*/
#include "defines.h"

#include <stdio.h>
#include <string.h>
#include <jpeglib.h>

#include "myincludesT.h"
#include "myincludes.h"
#include "myincludes.c"
/*
#include "errordef.c"
#include "ph.c"
#include "mem/ckalloc.c"
#include "file/QualifyFilename.c"
#include "file/FileExists.c"
#include "file/open.c"
#include "time/timestamp.c"
#include "imo/imoheader.c"
#include "imo/orgbT.h"
#include "file/ckwrite.c"
*/

/*----------------- globals ---------------------------------------------------------------------*/

char *	 infilename	 ;
char *	 outfilename	 ;
int	 needprogress	 ;
int	 infilesize	 ;
FILE *	 ifp		 ;
int	 ofh		 ;
int	 ck		 ;
int	 JpegStatedWi	 ;
int	 JpegStatedHi	 ;
int	 LinesWritten	 ;

/*----------------- help ------------------------------------------------------------------------*/

char * help[] = {
  "JpegToImo : read a JPEG file, decode it's image,                            ",
  "            and create an imo file containing same image                    ",
  "Syntax: JpegToImo [-v] [-p] <infilename> [<outfilename>]                    ",
  "  -v causes verbose output for debugging.                                   ",
  "  -p makes it print progress reports                                        ",
  "  if <outfilename is not specified, outfilename to use is constructed by    ",
  "    stripping any '.jpg' or '.jpeg' from <infilename> and appending '.imo'. ",
  NULL };

/*----------------- CheckArguments ---------------------------------------------------------*/

void CheckArguments( int argc, char ** argv ){
  int	 i	 ;
  char * arg	 ;

  verbose = 0 ; needprogress = 0 ;
  for( i=1 ; i<argc ; i++ ){ arg = argv[i] ;
    if( arg[0] == '-' ){
      if( arg[1] == 'v' ){ verbose++ ; needprogress = 1 ; continue ; }
      if( arg[1] == 'h' ){ phe(); }
      if( arg[1] == 'p' ){ needprogress = 1 ; continue ; }
      ex("invalid option %s", arg );
      }
    if( infilename == NULL ){ infilename = QualifyFilename( arg ); continue ; }
    if( outfilename == NULL ){ outfilename = QualifyFilename( arg ); continue ; }
    e("invalid argument %s", arg ); phe();
    }
  if( infilename == NULL ){ e("no infilename specified"); phe(); }
  if( outfilename == NULL ){
    int len ; char * p ;
    len = strlen( infilename );
    p = (char*) alloca( len + 6 ); if( p==NULL){ex("alloca failed"); }
    if( len >= 5 ){
      if( 0 == strcmp( infilename + len - 4, ".jpg")  ){ len -= 4 ; }
      if( 0 == strcmp( infilename + len - 5, ".jpeg") ){ len -= 5 ; }
      }
    strcpy( p, infilename );
    strcpy( p + len, ".imo");
    outfilename = QualifyFilename( p );
    }
  ck = FileExists( infilename, &infilesize );
  if( ck == 0 ){ ex("specified infile %s does not exist",infilename); }
  if( ck >  1 ){ ex("infile %s is not an ordinary file", infilename); }
  ck = FileExists( outfilename, NULL );
  if( ck >  1 ){ ex("outfile %s exists and is not an ordinary file",outfilename ); }
  /* silently overwrites any existing outfile of that name */

  v("infile=%s outfile=%s reportprogress=%d",infilename,outfilename,needprogress);
  }

/*----------------- ImoCheckHeader ---------------------------------------------------------*/

 /* In a jpeg file, number of lines is allowed to be stated incorrectly in it's header,	 */
 /*   but in a bolo file this is not allowed,						 */
 /*   so at end of conversion check that number of lines written is as jpeg header stated, */
 /*   and correct it if it is not.							 */

 /* ofh is assumed to already have been closed. */

void ImoCheckHeader( void ){
  if( JpegStatedHi == LinesWritten ){ return ; }
  v("JpegStatedHi = %d , LinesWritten = %d ; correcting.", JpegStatedHi, LinesWritten );
  ofh = open_rw( outfilename, 1 );
  WriteImoheader( ofh, JpegStatedWi, LinesWritten, 0, 0, 1.0, 1.0 );
  close( ofh );
  }

/*----------------- OutputScanline ---------------------------------------------------------*/

void OutputScanline3( JSAMPLE * jbuf, int wi ){
  static uint*	 obuf	 = NULL	 ;
  static int	 bufwi		 ;
  JSAMPLE *	 pj		 ;
  uint*		 po		 ;
  int		 i		 ;
  orgb		 o		 ;
  int		 outputsize	 ; /* in bytes */
  static int	 bufsize	 ; /* in bytes */

  if( wi < 0 ){ ex("wi negative"); }
  if( obuf == NULL ){
    bufsize = wi * sizeof(uint32_t) ;
v("bufsize = %d", bufsize );
    obuf = (uint*) ckalloc( bufsize );
    bufwi = wi ;
    WriteImoheader( ofh, wi, JpegStatedHi, 0, 0, 1.0, 1.0 );
    LinesWritten = 0 ;
v("did initialize");
    }
  if( wi != bufwi ){ ex("wi != bufwi ; wi=%d bufwi=%d",wi,bufwi); }

  pj = jbuf ; po = obuf ;
  for( i = wi ; i > 0 ; i--, pj+=3, po++ ){
    o.o = 0xFF   ;
    o.r = pj[ 0] ;
    o.g = pj[ 1] ;
    o.b = pj[ 2] ;
    *po = *((uint*) &o) ;
    }

  outputsize = bufsize ;
  ckwrite( ofh, obuf, outputsize );
v("wrote %d byte\n", outputsize );
  LinesWritten++ ;
  }

/*----------------- ConvertWrite -----------------------------------------------------------*/

void ConvertWrite( void ){
  int				 wi, hi		 ;
  int				 ccomp		 ;
  JSAMPLE *			 jbuf		 ;
  struct jpeg_decompress_struct	 cinfo		 ;
  struct jpeg_error_mgr		 jerr		 ;
  int				 line		 ;
/*  int				 ms		 ;*/ /* tmp. milliseconds */
/*  int				 refdiffsecs	 ;*/
/*  int				 oldrefdiffsecs	 ;*/

  /* 1 : setup errorhandling : */
  cinfo.err = jpeg_std_error( &jerr );

  /* 2 : initialize values in datastruct : */
  jpeg_create_decompress( &cinfo );

  /* 3 : specify datasource : */
  jpeg_stdio_src( &cinfo, ifp );

  /* 4 : read metadata from file (and set decompression parameters to default) : */
  jpeg_read_header( &cinfo, TRUE );

  cinfo.dct_method = JDCT_ISLOW ;
  /* cinfo.dct_method = JDCT_FLOAT ; is 6% slower */

  /* optional : modify decompression parameters : */
  /* optional : peek image width/height (only usefull when using lib's memoryallocator) : */
  /* optional : do something with metadata : */
  if( jpeg_has_multiple_scans( &cinfo ) ){ v("Has multiple scans"); }
  else                                   { v("Does not have multiple scans."); }

  /* 5 : initialize decompression (and retrieve metadata of image, eg width/height) : */
  jpeg_start_decompress( &cinfo );

  /* optional : do something with metadata of image : */
  wi = cinfo.output_width ;
  hi = cinfo.output_height ;
  ccomp = cinfo.output_components ;
  v("imagefile says that fileimage has %d lines of %d pixels, each having %d color components",
     cinfo.image_height, cinfo.image_width, cinfo.out_color_components );
  v("resultfile will have %d lines of %d pixels, each having %d colorcomponents",hi,wi,ccomp);
  v("recommended output buffer height is %d", cinfo.rec_outbuf_height );
  if( hi    <  0 ){ e("height is negative"); }			 /* is allowed */
  if( wi    <= 0 ){ ex("widht is non-positive"); }
  if( ccomp >  4 ){ ex("file has more than 4 color components. not supported."); }
  if( ccomp <  1 ){ ex("file has %d color components", ccomp ); }
  if( ccomp != 3 ){ ex("only 3 color components per pixel supported for now"); }

  JpegStatedWi = wi ; /* global */
  JpegStatedHi = hi ; /* global */

  jbuf = (JSAMPLE *) ckalloc( ccomp * wi * sizeof(JSAMPLE) + 8 ); /* small safety margin */

  /* 6 : request output scanlines until whole image has been decompressed : */
  /* MUST keep using values from struct, because they may change ;	 */
  /*   a jpeg is allowed to declare wrong number of lines.		 */

/*  UpdateTime(); SetReftime(); oldrefdiffsecs = 0 ;*/
  for( line=0 ; cinfo.output_scanline < cinfo.output_height ; line++ ){
    jpeg_read_scanlines( &cinfo, (JSAMPARRAY) &jbuf, 1 );
    OutputScanline3( jbuf, wi );
v("did output scanline. LinesWritten=%d. cinfo.output_scanline=%d. cinfo.output_height=%d.\n",
  LinesWritten, cinfo.output_scanline, cinfo.output_height );
/*    if( needprogress && ((line & 0xFF) == 0) ){*/
/*      SetRefdifftime(); refdiffsecs = RefdiffSecs();*/
/*      if( refdiffsecs >= 5 + oldrefdiffsecs ){*/
/*        ms = RefdiffMsecs() + 1 ;*/
/*        t("did %4d of %d lines in %6d milliseconds. (average %d kilopixels/second)",*/
/*          line, hi, ms, line * wi / ms );*/
/*        oldrefdiffsecs = refdiffsecs ;*/
/*        }*/
/*      }*/
    }
/*  if( needprogress ){*/
/*    SetRefdifftime(); ms = RefdiffMsecs() + 1 ;*/
/*    t("did %4d of %d lines in %6d milliseconds. (average %d kilopixels/second)",*/
/*      line, hi, ms, line * wi / ms );*/ ;
/*    }*/

  /* 7 : release memory of image object : */
  jpeg_finish_decompress( &cinfo );

  if( cinfo.err->num_warnings != 0 ){ e("there are %ld warnings", cinfo.err->num_warnings ); }
  else				    { v("there are 0 warnings"); }

  /* 9 : cleanup : */
  jpeg_destroy_decompress( &cinfo );

  }


/*----------------- main -------------------------------------------------------------------*/

int main( int argc, char ** argv ){

  CheckArguments( argc, argv );

  ifp = fopen( infilename, "r" );
  if( ifp == NULL ){ esx("failed fopen_r %s", infilename ); }
  ofh = open_w( outfilename, 1 );

  ConvertWrite();

  ck = fclose( ifp ); if( ck != 0 ){ es("error closing infile"); }
  ck =  close( ofh ); if( ck != 0 ){ es("error closing outfile"); }

  ImoCheckHeader();

  return(0);
  }

